/* ---------------------------------------------------------------- */
//
// Cache.m
//
// Created by Christoph Birk on 2017-06-01
// Copyright (c) 2017 Carnegie Observatories. All rights reserved.
//
// 2018-11-01  move iVARs into implementation
//
/* ---------------------------------------------------------------- */

#define DEBUG           0

/* ---------------------------------------------------------------- */

#import "Cache.h"

#if (DEBUG > 0)
#import "Logger.h"
extern Logger *main_logger;
#endif

#include <malloc/malloc.h>             // malloc_size()
#include <objc/runtime.h>              // class_getInstanceSize()

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation CacheObject

+ (CacheObject*)withArray:(NSArray*)a
{
  CacheObject *obj = [[CacheObject alloc] initWithArray:a];
  return [obj autorelease];
}

/* --- */

- (CacheObject*)initWithArray:(NSArray*)a
{
  if (!(self = [super init])) return nil;

  _array = [[NSArray alloc] initWithArray:a];
  _accessTime = [[NSDate date] timeIntervalSince1970];
  _size = class_getInstanceSize([self class]) + (malloc_size([a firstObject])*[a count]);
  
  return self;
}

/* --- */

- (void)dealloc
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",__func__,self);
#endif
  // assert([_array retainCount] == 1); // someone else might still own it
 
  [_array release];
  [super dealloc];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation Cache
{
  NSMutableDictionary *dict; 
  size_t capacity,size;
  double lifeTime;
}

/* ---------------------------------------------------------------- */

- (Cache*)initWithCapacity:(u_int)kb forTime:(double)life
{
  if (!(self = [super init])) return nil;
  
  dict = [[NSMutableDictionary alloc] initWithCapacity:32];
  capacity = 1024*kb;                  // [kBytes]
  lifeTime = life;                     // [seconds]
  _name = [[self description] copy];
  
  return self;
}

/* ---------------------------------------------------------------- */

- (NSArray*)loadKey:(NSString*)key context:(void*)context
{
  NSArray *array=nil;
#if (DEBUG > 0)
  NSString *string;
#endif
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%p)\n",__func__,[key UTF8String],context);
#endif

  @synchronized(self) {
    CacheObject *obj = [dict objectForKey:key];
    if (obj) {                         // cache hit
#if (DEBUG > 0)
      string = [NSString stringWithFormat:@"%@: hit     %@",_name,key];
      [main_logger append:[string UTF8String]];
#endif
      obj.accessTime = [[NSDate date] timeIntervalSince1970];
      array = [obj.array retain];      // retain now, autorelease at return
    } else {
#if (DEBUG > 0)
      string = [NSString stringWithFormat:@"%@: miss    %@",_name,key];
      [main_logger append:[string UTF8String]];
#endif
    }
  }

  [self flush];

  if (!array) {                        // cache miss
    if (_delegate) {                   // call delegate to load data
      array = [_delegate cacheLoadKey:key context:context];
    } else {
      array = [NSArray arrayWithContentsOfFile:key];
    }
    if (array) {                       // loaded data
#if (DEBUG > 0)
      string = [NSString stringWithFormat:@"%@: load    %@",_name,key];
      [main_logger append:[string UTF8String]];
#endif
      CacheObject *obj = [CacheObject withArray:array];
      if (obj) {                       // addCacheObject to dictionary
        @synchronized(self) {
          [dict setObject:obj forKey:key];
          size += obj.size;            // update cache size
        }
      }
      [array retain];                  // retain now, autorelease at return
    }
#if (DEBUG > 0)
    string = [NSString stringWithFormat:@"%@: size=   %lu [kB]",_name,size/1024];
    [main_logger append:[string UTF8String]];
#endif
  }

  return [array autorelease];
}

/* ---------------------------------------------------------------- */

- (void)flush
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",__func__,self);
#endif
#if (DEBUG > 0)
  size_t oldSize = size;
#endif

  @synchronized(self) {                  // remove older entries
    if (dict.count > 0) {
      double now = [[NSDate date] timeIntervalSince1970];
      NSMutableArray *array = [NSMutableArray arrayWithCapacity:dict.count];
      [dict enumerateKeysAndObjectsUsingBlock:^(NSString* key,CacheObject* obj,BOOL *stop) {
        double age =  now - obj.accessTime;
        if (age > lifeTime) {
          size -= obj.size;
          [array addObject:key];         // mark for removal
        }
      }];
      for (NSString *key in array) {     // remove marked objects
#if (DEBUG > 0)
        NSString *string = [NSString stringWithFormat:@"%@: remove  %@",_name,key];
        [main_logger append:[string UTF8String]];
#endif
        [dict removeObjectForKey:key];
      }
    } // endif(count)
  } // end-synchronized

  @synchronized(self) {
    while (size > capacity) {
      if (![self expunge]) break;        // remove oldest entry
    }
  } // end-synchronized
  
#if (DEBUG > 0)
  if (size != oldSize) { NSString *string;
    string = [NSString stringWithFormat:@"%@: size=   %lu [kB]",_name,size/1024];
    [main_logger append:[string UTF8String]];
  }
#endif

}

/* ---------------------------------------------------------------- */

- (BOOL)expunge
{
  BOOL r=FALSE;
  double now = [[NSDate date] timeIntervalSince1970];

  @synchronized(self) {                  // find oldest object
    __block double maxAge=0;
    __block NSString *oldestKey=nil;
    __block CacheObject *oldestObj=nil;
    [dict enumerateKeysAndObjectsUsingBlock:^(NSString* key,CacheObject* obj,BOOL *stop) {
      double age =  now - obj.accessTime;
      if (age > maxAge) {
        maxAge = age; oldestKey = key; oldestObj = obj;
      }
    }];
    if (oldestKey) {
      assert(oldestObj);
      size -= oldestObj.size;
#if (DEBUG > 0)
      NSString *string = [NSString stringWithFormat:@"%@: expunge %@",_name,oldestKey];
      [main_logger append:[string UTF8String]];
#endif
      [dict removeObjectForKey:oldestKey];
      r = TRUE;
    }
  }
  
  return r;
}

/* ---------------------------------------------------------------- */

- (void)clear
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",__func__,self);
#endif

#if (DEBUG > 0)
  if (size) {
    NSString *string = [NSString stringWithFormat:@"%@: cleared",_name];
    [main_logger append:[string UTF8String]];
  }
#endif

  @synchronized(self) {
    [dict removeAllObjects];
    size = 0;
  }
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",__func__,self);
#endif

  [dict removeAllObjects];
  [dict release];
  [_name release];
  [super dealloc];
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
