/* ---------------------------------------------------------------- */
//
// Sound.m
//
// Created by Christoph Birk on 2013-04-09
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG           0              // debug level
#endif

/* ---------------------------------------------------------------- */

#import "Sound.h"

/* ---------------------------------------------------------------- */

static NSLock         *mutex=nil;      // class variables
static NSDate         *last_time;
static NSString       *last_name;
static NSMutableArray *playing=nil;

/* ---------------------------------------------------------------- */

@implementation Sound

/* ---------------------------------------------------------------- */

+ (void)initialize
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p:\n",__func__,self);
#endif
  if (self == [Sound class]) {        // avoid call by sub-class
    mutex   = [[NSLock alloc] init];
    playing = [[NSMutableArray alloc] initWithCapacity:2];
  }
}

/* ---------------------------------------------------------------- */

+ (void)say:(NSString*)text
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,text);
#endif

  [self say:text withVoice:nil];
}

/* --- */

+ (void)say:(NSString*)text withVoice:(NSString*)voice
{
  NSString *vstr=nil;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%s)",__func__,[text UTF8String],[voice UTF8String]);
#endif

  if (voice) vstr = [NSString stringWithFormat:@"com.apple.speech.synthesis.voice.%@",voice];
  
  NSSpeechSynthesizer *speech = [[NSSpeechSynthesizer alloc] initWithVoice:vstr];
  if (speech) {
    [mutex lock];                      // MT-safe
    [playing addObject:speech];
    [mutex unlock];
    [speech setDelegate:(Sound*)self];
    [speech startSpeakingString:text];
  }
}

/* ---------------------------------------------------------------- */

+ (BOOL)play:(const char*)sound block:(double)timeout
{
  BOOL r=TRUE;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,sound);
#endif
  
  NSString *name = [NSString stringWithCString:sound encoding:NSASCIIStringEncoding];

  [mutex lock];                        // MT-safe
  if (last_name && last_time) {        // don't play the same sound twice
    if ([last_name isEqualToString:name]) {
      if ([last_time timeIntervalSinceNow] > -timeout) {
        r = FALSE;
      }
    }
  }
  if (r) {
    if (last_name) [last_name release];
    last_name = [name retain];
    if (last_time) [last_time release];
    last_time = [[NSDate date] retain];
  }
  [mutex unlock];

  if (r) r = [self play:sound];
  
  return r;
}

/* ---------------------------------------------------------------- */

+ (BOOL)play:(const char*)sound
{
  return [self play:sound volume:1.0f];
}

/* ---------------------------------------------------------------- */

+ (BOOL)play:(const char*)sound volume:(float)volume
{
  BOOL r=TRUE;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,sound);
#endif

  if (!sound) return FALSE;
  
  NSString *name = [NSString stringWithCString:sound encoding:NSASCIIStringEncoding];

  if (r) {
    if (strchr(sound,'\'') || strchr(sound,'\"')) { // speak
      [self say:name]; 
    } else {                                        // sound
      NSSound *snd = [[NSSound alloc] initWithContentsOfFile:name byReference:YES];
      if (snd) {
        [mutex lock];                  // MT-safe
        [playing addObject:snd];
        [mutex unlock];
        [snd setVolume:volume];
        [snd setDelegate:(Sound*)self];
        [snd play];
      }
    }
  }
  
  return r;
}

/* ---------------------------------------------------------------- */

+ (int)isPlaying
{
  int r;

  [mutex lock];                        // MT-safe
  r = (int)[playing count];
  [mutex unlock];

  return r;
}

/* ---------------------------------------------------------------- */

+ (void)sound:(NSSound*)snd didFinishPlaying:(BOOL)finished
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d)\n",PREFUN,snd,finished);
#endif
  [mutex lock];                        // MT-safe
  [playing removeObject:snd];
  [mutex unlock];
  [snd release];
}

/* ---------------------------------------------------------------- */
/* NSSpeechSynthesizerDelegate */

+ (void)speechSynthesizer:(NSSpeechSynthesizer*)syn didFinishSpeaking:(BOOL)finished
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d)\n",PREFUN,syn,finished);
#endif
  [mutex lock];                        // MT-safe
  [playing removeObject:syn];
  [mutex unlock];  
  [syn release];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
