/* ---------------------------------------------------------------- */
//
// StarCatalog.m
//
// Created by Christoph Birk on 2016-06-07
// Copyright (c) 2016 Christoph Birk. All rights reserved.
//
// 2017-07-26  added GSCplus catalog [CCB]
// 2017-07-27  _N, +nameFromIndex, -starTag
// 2018-03-13  fix ibounds overflow when delta~90
// 2018-06-05  Gaia
// 2018-11-01  move iVARs into implementation
//
/* ---------------------------------------------------------------- */

#define DEBUG             0
#define TIME_TEST         0

/* ---------------------------------------------------------------- */

#import  "StarCatalog.h"
#import  "Yutils.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

static const char sp0[]=" OBAFGKM";        // colors

/* ---------------------------------------------------------------- */

@implementation CatStar
{
  char spc_str[8];
}

/* --- */

+ (instancetype)starWithNum:(u_int)num
{
  CatStar *star = [[CatStar alloc] initWithNum:num];
  
  return [star autorelease];
}

- (instancetype)initWithNum:(u_int)n
{
  self = [super init];

  if (self) _num = n;

  return self;
}

/* --- */

- (char*)spc
{
  return (char*)spc_str;
}

/* --- */

- (BOOL)isEqual:(CatStar*)s
{
  return [self isEqual:s prec:0];
}

/* --- */

- (BOOL)isEqual:(CatStar*)s prec:(double)p
{
  double a,d;
  
  precess(s.alpha,s.delta,s.equinox,&a,&d,_equinox);
  if (fabs(_alpha-a)/3600.0 > p/15.0) return FALSE;
  if (fabs(_delta-d)/3600.0 > p) return FALSE;
  if (fabs(_mag-s.mag) > 0.005) return FALSE;
  
  return TRUE;
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

@implementation StarList

#ifndef NDEBUG
- (instancetype)init
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  assert(0);
  
  self = [super init]; if (!self) return nil;

  return self;
}
#endif

/* --- */

- (instancetype)initWithMaglim:(float)mlim
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  self = [super init]; if (!self) return nil;

  _array = [[NSMutableArray alloc] initWithCapacity:128];
  _maglim = mlim;
  
  return self;
}

/* --- */

- (void)dealloc
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  [_array release];
  [super dealloc];
}

/* --- */

- (CatStar*)getClosestA:(double)a0 D:(double)d0 E:(double)e0
{
  double alpha,delta,dista,distd,dist,distmin=9.9e99;
  CatStar *star=nil;
  
  precess(a0,d0,e0,&alpha,&delta,_equinox);
  const double cosd15 = 15.0*3600.0*cos(_delta/RADS);
 
  for (CatStar *s in _array) {         // find closest star
    distd = 3600.0*(delta - s.delta);
    dista =    fabs(alpha - s.alpha);
    if (dista > 12.0) dista = 24.0 - dista;
    dista *= cosd15;
    if ((dist=dista*dista+distd*distd) < distmin) {
      distmin = dist;
      star = s;
    }
  }
  
  return star;
}

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

@implementation StarCatalog
{
  int      adim,ddim;
  int      ia1,ia2,id1,id2;
  double   epoch;
  NSMutableString *path;
  NSLock   *mutex;
  Cache    *catCache;
}

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (int)indexFromName:(NSString*)name;
{
  NSRange range;
 
  range = [name rangeOfString:@"SAO" options:NSCaseInsensitiveSearch];
  if (range.location != NSNotFound) {
    range = [name rangeOfString:@"7" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_DIMM;
    return SCAT_SAO;
  }
  range = [name rangeOfString:@"DIMM" options:NSCaseInsensitiveSearch];
  if (range.location != NSNotFound) return SCAT_DIMM;
  
  range = [name rangeOfString:@"Tycho" options:NSCaseInsensitiveSearch];
  if (range.location != NSNotFound) return SCAT_TYCHO;
  
  range = [name rangeOfString:@"GSC" options:NSCaseInsensitiveSearch];
  if (range.location != NSNotFound) {
    range = [name rangeOfString:@"plus" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_GSCPLUS;
    return SCAT_GSC;
  }
  
  range = [name rangeOfString:@"USNO" options:NSCaseInsensitiveSearch];
  if (range.location != NSNotFound) {
    range = [name rangeOfString:@"13" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_USNO13;
    range = [name rangeOfString:@"14" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_USNO14;
    range = [name rangeOfString:@"15" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_USNO15;
    range = [name rangeOfString:@"16" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_USNO16;
    return SCAT_USNO;
  }

  range = [name rangeOfString:@"MASS" options:NSCaseInsensitiveSearch];
  if (range.location != NSNotFound) return SCAT_MASS;

  range = [name rangeOfString:@"Gaia" options:NSCaseInsensitiveSearch];
  if (range.location != NSNotFound) {
    range = [name rangeOfString:@"13" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_GAIA13;
    range = [name rangeOfString:@"15" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_GAIA15;
    range = [name rangeOfString:@"17" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_GAIA17;
    range = [name rangeOfString:@"19" options:NSCaseInsensitiveSearch];
    if (range.location != NSNotFound) return SCAT_GAIA19;
    // return SCAT_GAIA;
  }
  
  return -1;
}

/* --- */

+ (NSString*)nameFromIndex:(int)i
{
  NSString *string;
  
  switch (i) {
    case SCAT_SAO:     string = @"SAO"; break;
    case SCAT_DIMM:    string = @"DIMM"; break;
    case SCAT_TYCHO:   string = @"Tycho"; break;
    case SCAT_GSC:     string = @"GSC"; break;
    case SCAT_GSCPLUS: string = @"GSCplus"; break;
    case SCAT_USNO:    string = @"USNO"; break;
    case SCAT_USNO13:  string = @"USNO_13"; break;
    case SCAT_USNO14:  string = @"USNO_14"; break;
    case SCAT_USNO15:  string = @"USNO_15"; break;
    case SCAT_USNO16:  string = @"USNO_16"; break;
    case SCAT_MASS:    string = @"2MASS"; break;
    case SCAT_GAIA19:  string = @"Gaia19"; break;
    case SCAT_GAIA17:  string = @"Gaia17"; break;
    case SCAT_GAIA15:  string = @"Gaia15"; break;
    case SCAT_GAIA13:  string = @"Gaia13"; break;
    case -1:           string = @"none"; break;
    default:           string = @"unknown"; break;
  }
  
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (instancetype)init
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  self = [super init]; if (!self) return nil;

  mutex = [[NSLock alloc] init];
  path = [[NSMutableString alloc] initWithCapacity:128];
  
  epoch = get_epoch(0);
  
  catCache = [[Cache alloc] initWithCapacity:50000 forTime:36];
  catCache.delegate = self;
  catCache.name = [NSString stringWithFormat:@"StarCatalog:%p",self];

  _catalog = -1;
 
  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,_catalog);
#endif
  
  [catCache release];
  [mutex release];
  
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (BOOL)loadCatalog:(int)c path:(NSString*)folder
{
  BOOL     r=TRUE;
  int      a=0,d=0;
  float    m=0;
  double   e=0;
  NSString *base,*p=nil;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,c,folder);
#endif
  
  [mutex lock];

  if (folder) base = folder;
  else        base = @SCAT_BASE_PATH;

  switch (c) {
  case SCAT_DIMM:                      // C-DIMM
    p = [NSString stringWithFormat:@"%@/DIMM.StarCatalog",base];
    a =  12; d =  12; m =  7.0f; e = 2000.0;
    break;
  case SCAT_SAO:                       // SAO
    p = [NSString stringWithFormat:@"%@/SAO.StarCatalog",base];
    a =  24; d =  18; m = 12.0f; e = 2000.0;
    break;
  case SCAT_USNO13:                    // USNO-A2 (cutoff at mag=13)
    p = [NSString stringWithFormat:@"%@/USNO_13.StarCatalog",base];
    a =  48; d =  36; m = 13.0f; e = 2000.0;
    break;
  case SCAT_USNO14:                    // USNO-A2 (cutoff at mag=14)
    p = [NSString stringWithFormat:@"%@/USNO_14.StarCatalog",base];
    a =  48; d =  36; m = 14.0f; e = 2000.0;
    break;
  case SCAT_TYCHO:                     // Tycho2
    p = [NSString stringWithFormat:@"%@/Tycho.StarCatalog",base];
    a =  48; d =  36; m = 15.2f; e = 2000.0;
    break;
  case SCAT_USNO15:                    // USNO-A2 (cutoff at mag=15)
    p = [NSString stringWithFormat:@"%@/USNO_15.StarCatalog",base];
    a =  96; d =  72; m = 15.0f; e = 2000.0;
    break;
  case SCAT_USNO16:                    // USNO-A2 (cutoff at mag=16)
    p = [NSString stringWithFormat:@"%@/USNO_16.StarCatalog",base];
    a =  96; d =  72; m = 16.0f; e = 2000.0;
    break;
  case SCAT_GSC:                       // GSC
    p = [NSString stringWithFormat:@"%@/GSC.StarCatalog",base];
    a =  96; d =  72; m = 16.0f; e = 2000.0;
    break;
  case SCAT_GSCPLUS:                   // GSCplus
    p = [NSString stringWithFormat:@"%@/GSCplus.StarCatalog",base];
    a =  96; d =  72; m = 16.0f; e = 2000.0;
    break;
  case SCAT_USNO:                      // USNO-A2
    p = [NSString stringWithFormat:@"%@/USNO.StarCatalog",base];
    a = 192; d = 180; m = 19.0f; e = 2000.0;
    break;
  case SCAT_MASS:                      // 2-MASS
    p = [NSString stringWithFormat:@"%@/MASS.StarCatalog",base];
    a = 192; d = 180; m = 20.0f; e = 2000.0;
    break;
  case SCAT_GAIA19:                    // Gaia (m=19)
    p = [NSString stringWithFormat:@"%@/Gaia19.StarCatalog",base];
    a = 240; d = 180; m = 19.0f; e = 2000.0;
    break;
  case SCAT_GAIA17:                    // Gaia (m=17)
    p = [NSString stringWithFormat:@"%@/Gaia17.StarCatalog",base];
    a =  96; d =  72; m = 17.0f; e = 2000.0;
    break;
  case SCAT_GAIA15:                    // Gaia (m=15)
    p = [NSString stringWithFormat:@"%@/Gaia15.StarCatalog",base];
    a =  48; d =  36; m = 15.0f; e = 2000.0;
    break;
  case SCAT_GAIA13:                    // Gaia (m=13)
    p = [NSString stringWithFormat:@"%@/Gaia13.StarCatalog",base];
    a =  24; d =  18; m = 13.0f; e = 2000.0;
    break;
	default:                             // unknown catalog
		r = FALSE;
    break;
  }
  if (r) {
#if (DEBUG > 0)
    fprintf(stderr,"path=%s\n",[p UTF8String]);
#endif
    r = YpathReadable([p UTF8String]);
  }
  if (r) {
    [catCache clear];
    [path setString:p];
    adim = a; ddim = d; _mag = m; _equinox = e;
    _catalog = c;
  }
  epoch = get_epoch(0);
  
  [mutex unlock];

  return r;
}

/* --- */

- (BOOL)loadCatalog:(NSString*)name
{
  int c = [StarCatalog indexFromName:name];
  if (c < 0) return FALSE;
  
  return [self loadCatalog:c path:nil];
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)iboundsA:(double)a0 D:(double)d0 W:(double)rx H:(double)ry
{
	switch (_catalog) {
  case SCAT_DIMM:                      // index-boundaries for C-DIMM
    id1 = (int)floor((90.0+d0-ry/3600.0)/15.0); // delta b-width = 15
    id2 = (int)floor((90.0+d0+ry/3600.0)/15.0);
		break;
  case SCAT_SAO:                       // index-boundaries for SAO
  case SCAT_GAIA13:
    id1 = (int)floor((90.0+d0-ry/3600.0)/10.0); // delta b-width = 10
    id2 = (int)floor((90.0+d0+ry/3600.0)/10.0);
		break;
	case SCAT_TYCHO:                     // index-boundaries for Tycho2
  case SCAT_USNO13:
  case SCAT_USNO14:
  case SCAT_GAIA15:
    id1 = (int)floor((90.0+d0-ry/3600.0)/5.0);  // delta b-width = 5
    id2 = (int)floor((90.0+d0+ry/3600.0)/5.0);
		break;
	case SCAT_GSC: case SCAT_GSCPLUS:    // index-boundaries for GSC
  case SCAT_USNO15: case SCAT_USNO16:
  case SCAT_GAIA17:
    id1 = (int)floor((90.0+d0-ry/3600.0)/2.5);  // delta b-width = 2.5
    id2 = (int)floor((90.0+d0+ry/3600.0)/2.5);
		break;
  case SCAT_USNO:                      // index boundaries for UNSO-A
  case SCAT_MASS:                      // index boundaries for 2MASS
  case SCAT_GAIA19:
    id1 = (int)floor((90.0+d0-ry/3600.0)/1.0);  // delta b-width = 1.0
    id2 = (int)floor((90.0+d0+ry/3600.0)/1.0);
		break;
  default:
    assert(0);
    id1 = id2 = 0;
    break;
	}
  id1 = imax(0,imin(ddim-1,id1));
  id2 = imax(0,imin(ddim-1,id2));
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d): id1=%d, id2=%d\n",PREFUN,_catalog,id1,id2);
#endif

  while (fabs(d0) > 89.9) { d0 *= 0.999; } // avoid overflow
  double cosd = 15.0*3600.0*cos(d0/RADS);
	switch(_catalog) {                   // alpha indices
  case SCAT_DIMM:                      // C-DIMM catalog
    ia1 = (int)floor(0.5*(a0-rx/cosd));
    ia2 = (int)floor(0.5*(a0+rx/cosd));
    break;
  case SCAT_SAO:                       // SAO catalog
  case SCAT_GAIA13:
    ia1 = (int)floor(1.0*(a0-rx/cosd));
    ia2 = (int)floor(1.0*(a0+rx/cosd));
    break;
  case SCAT_TYCHO:                     // Tycho2
  case SCAT_USNO13: case SCAT_USNO14:
  case SCAT_GAIA15:
    ia1 = (int)floor(2.0*(a0-rx/cosd));
    ia2 = (int)floor(2.0*(a0+rx/cosd));
    break;
  case SCAT_GSC: case SCAT_GSCPLUS:    // GSC
  case SCAT_USNO15: case SCAT_USNO16:
  case SCAT_GAIA17:
    ia1 = (int)floor(4.0*(a0-rx/cosd));
    ia2 = (int)floor(4.0*(a0+rx/cosd));
    break;
  case SCAT_USNO:                      // USNO-A2 catalog
  case SCAT_MASS:                      // 2MASS catalog
    ia1 = (int)floor(8.0*(a0-rx/cosd));
    ia2 = (int)floor(8.0*(a0+rx/cosd));
    break;
  case SCAT_GAIA19:
    ia1 = (int)floor(10.0*(a0-rx/cosd));
    ia2 = (int)floor(10.0*(a0+rx/cosd));
    break;
  default:
    assert(0);
    ia1 = ia2 = 0;
    break;
  }
  // Note: modulo() is applied in getList: method
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d): ia1=%d, ia2=%d\n",PREFUN,_catalog,ia1,ia2);
#endif
}

/* ---------------------------------------------------------------- */

- (CatStar*)nextSAO:(FILE*)fp 
{
  const u_int RECORD_SIZE=11;
  u_char buf[RECORD_SIZE];
  double x;
  
  assert((_catalog == SCAT_SAO) || (_catalog == SCAT_DIMM));
  size_t r = fread(buf,RECORD_SIZE,1,fp);      // read record
  if (!r) return nil;
 
  u_int num = (u_int)buf[0]*65536 + (u_int)buf[1]*256 + (u_int)buf[2];
  CatStar *star = [CatStar starWithNum:num];
  
  x = (double)buf[3]*65536.0 + (double)buf[4]*256.0 +(double)buf[5];
  star.alpha = (x/(3600.0*15.0*10.0));
  x = (double)buf[6]*65536.0 + (double)buf[7]*256.0 +(double)buf[8];
  star.delta = (x/(3600.0*10.0)) - 90.0;
  star.equinox = _equinox;
  
  // IDEA proper motion @ current epoch (not yet in SAO.StarCatalog)
  
  star.mag = (float)buf[9]/10.0f;

  star.spc[0] = sp0[(buf[10] >> 4)];            // higher 4 bit
  star.spc[1] = '0' + (buf[10] & (u_char)0x0f); // lower 4 bit
  star.spc[2] = '\0';
 
  return star;
}

/* ---------------------------------------------------------------- */

- (CatStar*)nextTycho:(FILE*)fp
{
  const u_int RECORD_SIZE=12;
  u_char buf[RECORD_SIZE];
  double x;
 
  assert(_catalog == SCAT_TYCHO);
  size_t r = fread(buf,RECORD_SIZE,1,fp); // read record
  if (!r) return nil;
 
  u_int num = (u_int)buf[0]*65536*256 + (u_int)buf[1]*256*256 +
              (u_int)buf[2]*256 + (u_int)buf[3];
  CatStar *star = [CatStar starWithNum:num];
  
  x = (double)buf[4]*65536.0 + (double)buf[5]*256.0 + (double)buf[6];
  star.alpha = (x/(3600.0*15.0*10.0));
  x = (double)buf[7]*65536.0 + (double)buf[8]*256.0 + (double)buf[9];
  star.delta = (x/(3600.0*10.0)) - 90.0;
  star.equinox = _equinox;

  // IDEA proper motion @ current epoch (not yet in Tycho.StarCatalog)
  
  float b  = (float)buf[10]/10.0f;
  star.mag = (float)buf[11]/10.0f;

  sprintf(star.spc,"C%.1f",b-star.mag);
 
  return star;
}

/* ---------------------------------------------------------------- */

- (CatStar*)nextGSC:(FILE*)fp
{
  const u_int RECORD_SIZE=8;
  u_char buf[RECORD_SIZE];
  double x;
 
  assert(_catalog == SCAT_GSC);
  size_t r = fread(buf,RECORD_SIZE,1,fp); // read record
  if (!r) return nil;
 
  CatStar *star = [[CatStar alloc] init];
  
  x = (double)buf[0]*65536.0 + (double)buf[1]*256.0 + (double)buf[2];
  star.alpha = (x/(3600.0*15.0*10.0));
  x = (double)buf[3]*65536.0 + (double)buf[4]*256.0 + (double)buf[5];
  star.delta = (x/(3600.0*10.0)) - 90.0;
  star.equinox = _equinox;

  star.mag = (float)buf[6]/10.0f;

  star.spc[0] = sp0[buf[7] >> 4];                 // upper 4 bits
  if ((buf[7] & 0x0f) == 0x0a) star.spc[1] = ' '; // lower 4 bits
  else                         star.spc[1] = (buf[7] & 0x0f) + '0';
  star.spc[2] = '\0';
 
  return [star autorelease];
}

/* ---------------------------------------------------------------- */

- (CatStar*)nextUSNO:(FILE*)fp
{
  const u_int RECORD_SIZE=8;
  u_char buf[RECORD_SIZE];
  double x;

  assert((_catalog == SCAT_USNO)   || (_catalog == SCAT_USNO13) ||
         (_catalog == SCAT_USNO14) || (_catalog == SCAT_USNO15) ||
         (_catalog == SCAT_USNO16) || (_catalog == SCAT_GSCPLUS));
  
  size_t r = fread(buf,RECORD_SIZE,1,fp); // read record
  if (!r) return nil;
 
  CatStar *star = [[CatStar alloc] init];
  
  x = (double)buf[0]*65536.0 + (double)buf[1]*256.0 + (double)buf[2];
  star.alpha = (x/(3600.0*15.0*10.0));
  x = (double)buf[3]*65536.0 + (double)buf[4]*256.0 + (double)buf[5];
  star.delta = (x/(3600.0*10.0)) - 90.0;
  star.equinox = _equinox;
 
  double red  = (double)buf[6]/10.0;
  double blue = (double)buf[7]/10.0;
  if ((red > 0.0) && (blue > 0.0)) {
    star.mag = (float)((red+blue)/2.0);
    sprintf(star.spc,"C%.1f",blue-red);
  } else if (red > 0.0) {
    star.mag    = (float)red;
    star.spc[0] = '\0';
  } else if (blue > 0.0) {
    star.mag    = (float)blue;
    star.spc[0] = '\0';
  } else {
    assert(0);
  }
  return [star autorelease];
}

/* ---------------------------------------------------------------- */

- (CatStar*)nextMASS:(FILE*)fp
{
  const u_int RECORD_SIZE=10;
  u_char buf[RECORD_SIZE];
  double x;
 
  assert(_catalog == SCAT_MASS);
  size_t r = fread(buf,RECORD_SIZE,1,fp); // read record
  if (!r) return nil;
 
  CatStar *star = [[CatStar alloc] init];

  x = (double)buf[0]*65536.0 + (double)buf[1]*256.0 + (double)buf[2];
  star.alpha = (x/(3600.0*15.0*10.0));
  x = (double)buf[3]*65536.0 + (double)buf[4]*256.0 + (double)buf[5];
  star.delta = (x/(3600.0*10.0)) - 90.0;
  star.equinox = _equinox;
 
  star.mag_j = (float)buf[6]/10.0f;
  star.mag_h = (float)buf[7]/10.0f;
  star.mag_k = (float)buf[8]/10.0f;
  star.mag   = (star.mag_j+star.mag_h+star.mag_k)/3.0f;
  if (buf[9]) { 
    sprintf(star.spc,"C%.2f",(double)buf[9]/50.0 - 2.5);
  } else {
    strcpy (star.spc,"");
  }
 
  return [star autorelease];
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (CatStar*)nextGaia:(FILE*)fp
{
  const u_int RECORD_SIZE=10;
  const double gEpoch=2015.5;
  u_char buf[RECORD_SIZE];
  double x;
 
  assert((_catalog == SCAT_GAIA13) ||
         (_catalog == SCAT_GAIA15) || (_catalog == SCAT_GAIA17));
  size_t r = fread(buf,RECORD_SIZE,1,fp);
  if (!r) return nil;
 
  CatStar *star = [[CatStar alloc] init];

  x = (double)buf[0]*65536.0 + (double)buf[1]*256.0 + (double)buf[2];
  star.alpha = (x/(3600.0*15.0*10.0));
  x = (double)buf[3]*65536.0 + (double)buf[4]*256.0 + (double)buf[5];
  star.delta = (x/(3600.0*10.0)) - 90.0;
  star.equinox = _equinox;

  star.mag_k = (float)buf[6]/10.0f;         // R
  star.mag_h = (float)buf[7]/10.0f - 2.0f;  // G
  star.mag_j = (float)buf[8]/10.0f;         // B
  star.mag   = star.mag_h;
  if ((star.mag_k > 0) && (star.mag_j > 0)) {
    sprintf(star.spc,"C%.1f",(star.mag_j-star.mag_k));
  } else
  if (star.mag_k > 0) {
    sprintf(star.spc,"C%.1f",(star.mag_h-star.mag_k)*2.0f);
  } else
  if (star.mag_j > 0) {
    sprintf(star.spc,"C%.1f",(star.mag_j-star.mag_h)*2.0f);
  } else {
    star.spc[0] = '\0';
  }

  assert(RECORD_SIZE == 10);
  u_char flags = buf[9];
  if (flags & 0x01) strcat(star.spc,"v");  // variable star

  assert(epoch > 2000);
  if (flags & 0x02) {                  // have PMa
    (void)fread(buf,1,1,fp);
    x = 0.050 * (double)buf[0];        // [mas/yr]
    assert(x > 0);
    if (flags & 0x04) x = -x;
    star.alpha += (x/(15.0*3600.0*cos(star.delta/RADS))) * (epoch-gEpoch);
  }
  if (flags & 0x08) {                  // have PMd
    (void)fread(buf,1,1,fp);
    x = 0.050 * (double)buf[0];        // [mas/yr]
    assert(x > 0);
    if (flags & 0x10) x = -x;
    star.delta += (x/3600.0) * (epoch-gEpoch);
  }
 
  return [star autorelease];
}

/* --- */

- (CatStar*)nextGaia19:(FILE*)fp alpha:(double)a0 delta:(double)d0
{
  const u_int RECORD_SIZE=8;
  const double gEpoch=2015.5;
  u_char buf[RECORD_SIZE];
  double x;
 
  assert((_catalog == SCAT_GAIA19));
  size_t r = fread(buf,RECORD_SIZE,1,fp);
  if (!r) return nil;
 
  CatStar *star = [[CatStar alloc] init];

  x = (double)buf[0]*256.0 + (double)buf[1];
  assert(adim >= 240);
  star.alpha = a0 + (x/(3600.0*15.0*10.0));
  x = (double)buf[2]*256.0 + (double)buf[3];
  star.delta = d0 + (x/(3600.0*10.0));
  star.equinox = _equinox;

  star.mag_k = (float)buf[4]/10.0f;         // R
  star.mag_h = (float)buf[5]/10.0f - 2.0f;  // G
  star.mag_j = (float)buf[6]/10.0f;         // B
  star.mag   = star.mag_h;
  if ((star.mag_k > 0) && (star.mag_j > 0)) {
    sprintf(star.spc,"C%.1f",(star.mag_j-star.mag_k));
  } else
  if (star.mag_k > 0) {
    sprintf(star.spc,"C%.1f",(star.mag_h-star.mag_k)*2.0f);
  } else
  if (star.mag_j > 0) {
    sprintf(star.spc,"C%.1f",(star.mag_j-star.mag_h)*2.0f);
  } else {
    star.spc[0] = '\0';
  }

  assert(RECORD_SIZE == 8);
  u_char flags = buf[7];
  if (flags & 0x01) strcat(star.spc,"v");  // variable star

  assert(epoch > 2000);
  if (flags & 0x02) {                  // have PMa
    (void)fread(buf,1,1,fp);
    x = 0.050 * (double)buf[0];        // [mas/yr]
    assert(x > 0);
    if (flags & 0x04) x = -x;
    star.alpha += (x/(15.0*3600.0*cos(star.delta/RADS))) * (epoch-gEpoch);
  }
  if (flags & 0x08) {                  // have PMd
    (void)fread(buf,1,1,fp);
    x = 0.050 * (double)buf[0];        // [mas/yr]
    assert(x > 0);
    if (flags & 0x10) x = -x;
    star.delta += (x/3600.0) * (epoch-gEpoch);
  }
 
  return [star autorelease];
}

/* ---------------------------------------------------------------- */
#pragma mark "Delegate"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (NSArray*)cacheLoadKey:(NSString*)key context:(void*)context
{
  NSMutableArray *array=nil;
  CatStar *star=nil;
#if (DEBUG > 1)
  fprintf(stderr,"%s: path=%s\n",PREFUN,[key UTF8String]);
#endif

  FILE *fp = fopen([key UTF8String],"r");
  if (fp) { double a0=0,d0=0;
    if (_catalog == SCAT_GAIA19) { char file[512],*p;
      strcpy(file,[key UTF8String]);  // calculate coordinate offset
      p = strstr(file,".cat"); *p = '\0'; p = strrchr(file,'/');
      a0 = 24.0*atof(p+1)/(double)adim;
      *p = '\0'; p = strrchr(file,'/');
      d0 = -90.0 + 180.0*atof(p+1)/(double)ddim;
    }
    array = [NSMutableArray arrayWithCapacity:2048];
    do {
      switch (_catalog) {
        case SCAT_SAO: case SCAT_DIMM:
          star = [self nextSAO:fp];
          break;
        case SCAT_TYCHO:
          star = [self nextTycho:fp];
          break;
        case SCAT_GSC:
          star = [self nextGSC:fp];
          break;
        case SCAT_GSCPLUS: case SCAT_USNO:
        case SCAT_USNO13: case SCAT_USNO14: case SCAT_USNO15: case SCAT_USNO16:
          star = [self nextUSNO:fp];
          break;
        case SCAT_MASS:
          star = [self nextMASS:fp];
          break;
        case SCAT_GAIA13: case SCAT_GAIA15: case SCAT_GAIA17:
          star = [self nextGaia:fp];
          break;
        case SCAT_GAIA19:
          star = [self nextGaia19:fp alpha:a0 delta:d0];
          break;
        default:
          star = nil;
          break;
      }
      if (!star) break;
      assert(star.equinox != 0);
      [array addObject:star];
    } while (star);
    fclose(fp);
  } // endif(fp)
    
  return array;
}

/* ---------------------------------------------------------------- */

- (StarList*)getList:(double)a0 D:(double)d0 E:(double)e0 W:(float)rx H:(float)ry M:(float)mlim
{
  double   dista,cosd15,alpha,delta;
  NSString *key=nil;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%f,%f,%.1f,%.1f,%.1f,%.2f)\n",PREFUN,a0,d0,e0,rx,ry,mlim);
#endif
  
  if (_catalog < 0) return nil;
#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif
  assert(_equinox);
  assert(_mag);
  assert(mutex);
  [mutex lock];
  
  epoch = get_epoch(0);
  cosd15 = 15.0*3600.0*cos(d0/RADS);   // local 'cos(delta)'
  precess(a0,d0,e0,&alpha,&delta,_equinox); // precess to catalog equinox

  StarList *list = [[StarList alloc] initWithMaglim:mlim];
  
  [self iboundsA:alpha D:delta W:rx H:ry];

  for (int j=id1; j<=id2; j++) {       // delta-bands
    for (int i=ia1; i<=ia2; i++) {     // hour-bands
      int ii = (int)modulo(i,0,adim);
      switch (_catalog) {
      case SCAT_USNO:
      case SCAT_MASS:
        key = [NSString stringWithFormat:@"%@/%03d/%05d.cat",path,j,125*ii];
        break;
      case SCAT_GAIA19: case SCAT_GAIA17: case SCAT_GAIA15: case SCAT_GAIA13:
        key = [NSString stringWithFormat:@"%@/%03d/%03d.cat",path,j,ii];
        break;
      default:
        key = [NSString stringWithFormat:@"%@/%02d/%02d.cat",path,j,ii];
        break;
      }
      if (!key) continue;
      NSArray *a = [catCache loadKey:key context:NULL];
      for (CatStar *star in a) {
        if (fabs(delta-star.delta)*3600.0 > ry) continue; // distd
        dista = fabs(alpha-star.alpha);                   // dista
        if (dista > 12.0) dista = 24.0-dista;
        if (dista*cosd15 > rx) continue;  // out of FOV
        if (star.mag > mlim) continue;    // star too faint
        [list.array addObject:star];
      }  // endfor(n)
    }  // endfor(RA-bands)
  }  // endfor(delta-bands)
#if (DEBUG > 1)
  fprintf(stderr,"%s(): nlist=%lu\n",PREFUN,[list.array count]);
#endif

  list.alpha = alpha; list.delta = delta; list.equinox = _equinox;

  [mutex unlock];
#if (TIME_TEST > 0)
  printf("%s: %.2f [ms]\n",PREFUN,YmsSince(date));
#endif
  return [list autorelease];
}

/* ---------------------------------------------------------------- */

- (NSString*)starName:(u_int)number
{
  NSString *name=nil;
 
  switch (_catalog) {
  case SCAT_SAO:                        // SAO
  case SCAT_DIMM:
    name = [NSString stringWithFormat:@"SAO %u",number];
    break;
  case SCAT_TYCHO:                      // Tycho2
    name = [NSString stringWithFormat:@"%04u-%05u-%u",
            (number >> 18) & 0x03fff,
            (number >> 3) & 0x7fff, number & 0x07];
    break;
  default:
    assert(number == 0);
    name = @"";
    break;
  }

  return name;
}

/* ---------------------------------------------------------------- */

- (NSString*)starTag:(CatStar*)star
{
  NSString *tag=nil;

  switch (_catalog) {
  case SCAT_GAIA13: case SCAT_GAIA15: case SCAT_GAIA17: case SCAT_GAIA19:
    if ((star.mag_k > 0) && (star.mag_j > 0)) {
      tag = [NSString stringWithFormat:@"R=%.1f,G=%.1f,B=%.1f (%s)",
               star.mag_k,star.mag_h,star.mag_j,star.spc];
    } else
    if (star.mag_k > 0) {
      tag = [NSString stringWithFormat:@"R=%.1f,G=%.1f (%s)",
               star.mag_k,star.mag_h,star.spc];
    } else
    if (star.mag_j > 0) {
      tag = [NSString stringWithFormat:@"G=%.1f,B=%.1f (%s)",
               star.mag_h,star.mag_j,star.spc];
    } else {
      tag = [NSString stringWithFormat:@"G=%.1f (%s)",star.mag_h,star.spc];
    }
    break;
  case SCAT_MASS:
    tag = [NSString stringWithFormat:@"J=%.1f,H=%.1f,Ks=%.1f",
             star.mag_j,star.mag_h,star.mag_k];
    break;
  case SCAT_SAO: case SCAT_DIMM: case SCAT_TYCHO:
    tag = [NSString stringWithFormat:@"%@,%s,%.1f",
             [self starName:star.num],star.spc,star.mag];
    break;
  default:
    tag = [NSString stringWithFormat:@"%s,%.1f",star.spc,star.mag];
    break;
  }
  return tag;
}

/* ---------------------------------------------------------------- */

- (NSString*)catName
{
  return [[self class] nameFromIndex:_catalog];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
