/* ---------------------------------------------------------------- */
//
// SwopeIO.m
// Swope
//
// Created by Christoph Birk on 2013-08-14
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG           1              // debug level

#define RESPONSE_SIZE   512

#undef  USE_TCSSIM                     // SwopeTCS simulator workarounds 

/* ---------------------------------------------------------------- */

#import  "SwopeIO.h"
#import  "Logger.h"
#import  "Yutils.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

static Logger *telio_logger=nil;

/* ---------------------------------------------------------------- */

@implementation SwopeIO

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (void)initialize
{
  extern BOOL use_shared;
  
  if (self != [SwopeIO class]) return; // prevent sub-classes
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  telio_logger = [[Logger alloc] initWithBase:"telio" useShared:use_shared];
#ifdef NDEBUG
  telio_logger.mode = LOGGER_OPEN;    // keep open, no flush
#else
  telio_logger.mode = LOGGER_FLUSH;   // keep open and flush
#endif
}

/* ---------------------------------------------------------------- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
  case E_swope_error:   string = @"SwopeTCS: error"; break;
  case E_swope_notok:   string = @"SwopeTCS: command not accepted"; break;
  case E_swope_slewing: string = @"SwopeTCS: slew timed out"; break;
  case E_swope_focus:   string = @"SwopeTCS: focus motion timeout"; break;
  default:
    if (!string) string = [TCPIP_Connection errorString:err];
    break;
  }
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)initWithHost:(const char*)hstr port:(int)pint online:(BOOL)on
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d,%d)\n",PREFUN,hstr,pint,on);
#endif

  self = [super initWithHost:hstr port:pint online:on];
  if (!self) return nil;

  tcs = on;
  terminator = '\0';                   // no terminator

  return self;
}

/* --- */

- (void)dealloc
{
  [_tcsError release];
  [super dealloc];
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)getAlpha:(double*)alp delta:(double*)del equinox:(double*)equ
{
#if (DEBUG > 0) 
  fprintf(stderr,"%s(%p,%p,%.4f)\n",PREFUN,alp,del,*equ);
#endif

  int err = [self open]; if (err) return err;
  
  if (tcs == TCS_RPI) {            
    NSDictionary *d = [self JSONgetStatus];
    if (d) { NSNumber *n;
      n = [d objectForKey:@"RA"];
      double ra = [n doubleValue]/15.0;
      n = [d objectForKey:@"Dec"];
      double dc = [n doubleValue];
      precess(ra,dc,get_epoch(0),alp,del,*equ);
    } else err = E_swope_error;
  }
  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

#if 0 // unused
static int is_ok(NSDictionary* d)      // b0484
{
  NSNumber *n = [d objectForKey:@"ok"];
  return [n intValue];
}

/* --- */

static NSString* has_error(NSDictionary* d)
{
  NSString *str=nil;
  
  if (!is_ok(d)) {                     // ok:false
     str = [d objectForKey:@"Error"];
     if (!str) str = [SwopeIO errorString:E_swope_error];
  }
  return str;
}

/* --- */

- (int)get_error:(NSDictionary*)d
{
  self.tcsError = has_error(d);        // check key 'ok'
  
  return (_tcsError) ? E_swope_error : 0;
}
#endif

/* ---------------------------------------------------------------- */

- (int)mstat:(double)timeout           // wait for mount to stop slewing
{
  NSDictionary *d;
  NSNumber *n;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%.1f)\n",PREFUN,timeout);
#endif
  
  int err = [self open]; if (err) return err;
  
  if (tcs == TCS_RPI) {
    NSDate *date = [NSDate date];
    err = E_swope_slewing;
    do {
      [NSThread sleepForTimeInterval:2.0];
      d = [self JSONgetStatus]; if (!d) break;
      n = [d objectForKey:@"Slewing"];
      if ([n intValue] == 0) { err=0; break; }
#ifdef USE_TCSSIM                      // SwopeTCS simulator b0486
      [self stop];
#endif
    } while (YsecSince(date) < timeout);
  }
  [self close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)offsetRA:(double)ofra dec:(double)ofdc delta:(double)delta
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%.1f,%.1f,%f)\n",PREFUN,ofra,ofdc,delta);
#endif

  int err = [self open]; if (err) return err;

  double timeout = 60.0;              // MarcoF: 30 seconds for < 0.3 deg
  if (tcs == TCS_RPI) { NSMutableDictionary *d;
    d = [NSMutableDictionary dictionaryWithObject:@"OFFSET" forKey:@"Command"];
    [d setObject:@(ofra) forKey:@"HA"];
    [d setObject:@(ofdc) forKey:@"Dec"];
    err = [self JSONcommand:d];
    if (!err) {                       // ok:true
      err = [self mstat:timeout];     // wait for 'Slewing:false'
    }
  }
  [self close];    
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)moveRA:(double)ra dec:(double)dc equ:(double)eq
{
  NSMutableDictionary *d;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%.6f,%.6f,%.3f)\n",PREFUN,ra,dc,eq);
#endif

  self.tcsError = nil;
  
  int err = [self open]; if (err) return err;
  
  d = [NSMutableDictionary dictionaryWithObject:@"NEXTOBJ" forKey:@"Command"];
  [d setObject:@(15.0*ra) forKey:@"RA"];
  [d setObject:@(dc) forKey:@"Dec"];
  [d setObject:@(eq) forKey:@"Equinox"];
  // NSLog(@"%@",d);
  if (tcs == TCS_RPI) {             // new Swope TCS
    err = [self JSONcommand:d];
    if (!err) {
#ifdef USE_TCSSIM
            [self JSONcommand:@{ @"Command" : @"SLEW" }]; // always ok:false
#else
      err = [self JSONcommand:@{ @"Command" : @"SLEW" }];
#endif
    }
    if (!err) {
      double timeout=600.0;         // MarcoF: 0.3 deg/sec
      err = [self mstat:timeout];
    }
  } else Ysleep(2000);              // offline
  [self close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getAirmass:(float*)am
{
  int  err=0;

  if (tcs == TCS_RPI) { NSNumber *n=nil;
    NSDictionary *d = [self JSONgetStatus];
    if (d) n = [d objectForKey:@"Airmass"];
    if (n) *am = [n doubleValue];
    else   err = E_tcpip_error;
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getFocus:(int*)fo
{
  int err=0;

  if (tcs == TCS_RPI) { NSNumber *n=nil;
#if 0 // there are two ways to get the current focuser position
    NSDictionary *d = [self JSONgetStatus];
    if (d) n = [d objectForKey:@"FocusPos"];
#else // less overhead
    NSMutableDictionary *d;            // b0488
    d = [NSMutableDictionary dictionaryWithObject:@"FOCUSER" forKey:@"Command"];
    NSDictionary *r = [self JSONsend:d];
    if (r) n = [r objectForKey:@"FocusPos"];
#endif
    if (n) *fo = [n intValue];
    else   err = E_tcpip_error;
  }
  return err;
}

/* --- */

- (int)setFocus:(int)fo
{
  int err=0;
  
  if (tcs == TCS_RPI) { NSMutableDictionary *d;
    d = [NSMutableDictionary dictionaryWithObject:@"FOCUSER" forKey:@"Command"];
    [d setObject:@(fo) forKey:@"SET"];
    err = [self JSONcommand:d];
  }
  return err;
}

/* --- */

- (BOOL)focusMoving
{
  BOOL r=NO;
  
  if (tcs == TCS_RPI) { NSNumber *n=nil;
    NSDictionary *d = [self JSONgetStatus];
    if (d) n = [d objectForKey:@"FocusMoving"];
    if (n) r = ([n intValue]) ? YES : NO;
  }
  return r;
}

/* --- */

- (int)focusStop
{
  int err=0;
  
  if (tcs == TCS_RPI) { NSMutableDictionary *d;
    d = [NSMutableDictionary dictionaryWithObject:@"FOCUSER" forKey:@"Command"];
    [d setObject:@"STOP" forKey:@"SET"];
    err = [self JSONcommand:d];
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getST:(int*)st                  // [s]
{
  int  err=0;
  
  if (tcs == TCS_RPI) { NSNumber *n=nil;
    NSDictionary *d = [self JSONgetStatus];
    if (d) n = [d objectForKey:@"LST"]; // [deg]
    if (n) *st = (int)(24.0*3600.0*[n doubleValue]/360.0);
    else   err = E_tcpip_error;
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getUT:(int*)ut                  // [s]
{
  int err=0;
  
  if (tcs == TCS_RPI) { NSNumber *n=nil;
    NSDictionary *d = [self JSONgetStatus];
    if (d) n = [d objectForKey:@"UT"];   // [hrs]
    if (n) *ut = (int)(3600.0*[n doubleValue]);
    else   err = E_tcpip_error;
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)set_guiding:(BOOL)status  // Note: does not work currently
{
  int err=0;

  if (tcs == TCS_RPI) { NSMutableDictionary *d; // b0485
    d = [NSMutableDictionary dictionaryWithObject:@"GUIDE" forKey:@"Command"];
    [d setObject:@(status) forKey:@"ENABLE"];
#ifdef USE_TCSSIM
          [self JSONcommand:d];
#else
    err = [self JSONcommand:d];
#endif
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)stop
{
  int err=0;
#if 0 // TESTING -- Simulator returns reasonable answers
  [self JSONsend:@{ @"Command" : @"PING" }];  // ok:true
  [self JSONsend:@{ @"Command" : @"LIMITS" }]; // HardLimits:0
  [self JSONsend:@{ @"Command" : @"LOCALCONTROL" }]; // LocalControl:false
  [self JSONsend:@{ @"Command" : @"READ24V" }]; // SysADC:1024,SysVolt:24.0
  [self JSONsend:@{ @"Command" : @"ENCSTAT" }]; // {Enc_Dec1:0 ... }
#endif
  
  if (tcs == TCS_RPI) {
    err = [self JSONcommand:@{ @"Command" : @"SLEWSTOP" }];
  }
  
  return err;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

- (NSDictionary*)JSONgetStatus
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert(tcs == TCS_RPI);

  NSDictionary *r = [self JSONsend:@{ @"Command" : @"STATUS" }];
  
  return r;
}

/* ---------------------------------------------------------------- */

- (NSDictionary*)JSONsend:(NSDictionary*)d
{
  NSDictionary *r=nil;
  char buffer[1024];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,d);
#endif
  assert(online);

  // NSLog(@"--> %@",d);
  NSData *json = [NSJSONSerialization dataWithJSONObject:d options:0 error:nil];
  if (json.length < sizeof(buffer)) { // append to logfile
    memcpy(buffer,json.bytes,json.length); buffer[json.length] = '\0';
    [telio_logger append:buffer]; // printf("--> %s\n",buffer);
  }
  int err = [self open]; if (err) return nil;

  err = TCPIP_Send(sock,json.bytes,json.length);
  if (!err) {
    size_t s = TCPIP_Recv(sock,buffer,sizeof(buffer),2000);
    if (s < sizeof(buffer)) {        // append to logfile
      buffer[s] = '\0'; [telio_logger append:buffer]; // printf("<-- %s\n",buffer);
    }
    NSData *t = [NSData dataWithBytes:buffer length:s];
    r = [NSJSONSerialization JSONObjectWithData:t options:0 error:nil];
  }
  [self close];
  // NSLog(@"<-- %@",r);

  return r;
}

/* ---------------------------------------------------------------- */

- (int)JSONcommand:(NSDictionary*)d
{
  int err=0;
  
  NSDictionary *r = [self JSONsend:d];
  // NSLog(@"%@",r);
  if (!r) return E_tcpip_open;
  
  NSNumber *n = [r objectForKey:@"ok"];
  if ([n intValue]) {                  // ok:true
    self.tcsError = nil;
  } else {
    self.tcsError = [r objectForKey:@"Error"]; // use string if provided
    if (!_tcsError) self.tcsError = [SwopeIO errorString:E_swope_error];
    err = E_swope_error;
  }
  return err;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
