/* ---------------------------------------------------------------- */
//
// AppDelegate.m
// Henrietta
//
// Created by Christoph Birk on 2022-06-05
// Copyright © 2022 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG              1            // debug level

/* ---------------------------------------------------------------- */

#import "AppDelegate.h"
#import "Yutils.h"
#import "Notice.h"
#import "Sound.h"
#import "MTwister.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

#define FRAME_MESSAGES     "MessagePanel"

#define DBE_EMAIL          "dbe_email_address"
#define DEF_EMAIL          "birk"  // ?Nick
#define DBE_ENGMODE        "dbe_engmode"
#define DBE_DISPERSER      "dbe_disperser"
#define DBE_HEATER_WARNING "dbe_heater_warning" // ?Nick
#define DBE_THIN_TIME      "dbe_thinning_time"

#define DBE_ARCHON_START   "dbe_archon_startup"

#define DBE_LAKE_START     "dbe_lakeshore_startup"

#define DBE_GALIL_START    "dbe_galil_startup"
#define DBE_TCS_START      "dbe_tcs_startup"

#define DBE_TCPIP          "dbe_tcpip_server"

static const char *engWarning="Engineering mode required (Preferences/Settings)";

/* ---------------------------------------------------------------- */

static int verify_path(const char* path,BOOL interactive)
{
  BOOL k=YES;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,path,interactive);
#endif

  if (*path == '\0') return -1;        // empty path
  if (YpathWritable(path)) return 0;   // path exists

  if (interactive) {                   // modal dialog
    k = YalertQuestion(path,"directory does not exist. Create?",0,"OK","cancel");
  }
  if (!k) return -1;
  
  k = YpathCreate(path,YES);
  if (k) {                             // path created
    k  = YpathWritable(path);          // but not writable
    if (!k) YfileRemove(path);         // remove it
  }

  if (!k && interactive) { char buf[512];
    sprintf(buf,"failed to set data path\r%s",path);
    [Notice openNotice:buf title:PREFUN time:0 level:1];
  }
  
  return (k) ? 0 : 1;
}

/* ---------------------------------------------------------------- */

@interface AppDelegate ()

@property (weak) IBOutlet NSWindow *window;
@end

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation AppDelegate
{
  TCPIP_Server *tcpip_server;
  u_int counter;
}

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
  default:
    if (!string) string = [CameraController errorString:err];
    assert(string); // CameraController never returns 'nil'
    break;
  }
  if (!string) {                       // fall-back to excuse
    string = [NSString stringWithFormat:@"%@ (%d)",[self makeExcuse],err];
  }

  return string;
}

/* --- */

+ (NSString*)makeExcuse
{
  static NSArray* arrays=nil;
  
  @synchronized(self) {
    if (!arrays) {
      NSString *file = [[NSBundle mainBundle] pathForResource:@"excuses" ofType:@"xml"];
      arrays = [[NSArray arrayWithContentsOfFile:file] retain];
    }
  }
  assert(arrays);
  if (!arrays) return @"no excuse";
  assert([arrays count] == 4);
  if ([arrays count] != 4) return @"invalid excuse";

  NSMutableString *excuse = [NSMutableString stringWithCapacity:64];
  
  for (int j=0; j<4; j++) {
    NSArray *a = [arrays objectAtIndex:j];
    NSString *word = [a objectAtIndex:[MTwister intValue:(int)[a count]]];
    [excuse appendString:word];
    if ((j == 2) && ([MTwister doubleValue:1.0] > 0.2)) break;
    [excuse appendString:@" "];
  }
  return excuse;
}

/* --- */

+ (void)goodbye:(NSString*)text error:(int)err
{
  NSString *string = [NSString stringWithFormat:@"%@, err=%d",text,err];
  NSString *info = [self errorString:err];
  NSImage *image = [[NSImage alloc] initWithContentsOfFile:[[NSBundle mainBundle]
                                    pathForResource:@"macOSbomb" ofType:@"png"]];
  dispatch_async(MAIN_QUEUE,^(void) {
    NSImageView *view = [[NSImageView alloc]
                          initWithFrame:NSMakeRect(0,0,image.size.width,image.size.height)];
    [view setImage:image];
    NSAlert *alert = [[NSAlert alloc] init];
    if (text) [alert setMessageText:string];
    if (info) [alert setInformativeText:info];
    [alert addButtonWithTitle:@"ok"];
    [alert setAccessoryView:view];
    [view release];
    [alert runModal];
    [alert release];
    [NSApp terminate:self];
  });
  [image release];
}

/* ---------------------------------------------------------------- */
#pragma mark "Application Delegate"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
  int online,start;
  char buf[128];
  NSString *str;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,aNotification);
#endif

#if defined(SIM_ONLY) && defined(NDEBUG)
  [Notice openNotice:"Warning: Simulator-ONLY version" title:P_TITLE];
#endif
  
  YbundleInfo(&p_exec,&p_build,&p_version,NULL,&r_path,&p_ident);
  // printf("%s\n",r_path);

  // defaults write <app domain name> NSAppSleepDisabled -bool YES
  if (!YappNapDisabled()) {
    [Notice openNotice:"Warning: AppNap enabled\rPlease restart the GUI"
            title:P_TITLE time:0 level:1];
  }

  NSArray *users = [NSArray arrayWithObjects:@"birk",@"pfs",nil];
  NSLog(@"userName= %@",NSUserName());
  if ([users indexOfObject:NSUserName()] == NSNotFound) {
    use_shared = YprefSetupShared();
  } else {
    use_shared = NO;
  }
  (void)YappSupportCreate(&s_path,use_shared);
  
  main_logger = [[Logger alloc] initWithBase:nil useShared:use_shared];
  // main_logger.showUT = TRUE;      // ?Nick wants local time
#ifdef NDEBUG
  main_logger.mode = LOGGER_OPEN;    // keep open
#else
  main_logger.mode = LOGGER_FLUSH;   // keep open, auto-flush
#endif
  str = [NSString stringWithFormat:@"user= %@ , sharedSetup= %d , UT= %s",
           NSUserName(),use_shared,get_ut_timestr(buf,cor_time(0))];
  [main_logger append:[str UTF8String]];

  [edit_observer setStringValue:YprefGetString(@DBE_OBSERVER,@DEF_OBSERVER,NO)];

  online = YprefGetInt(DBE_ARCHON_ONLINE,DEF_STATE,use_shared);
  [popup_archon selectItemAtIndex:online];
  [edit_archon setStringValue:YprefGetString(@DBE_ARCHON_HOST,@DEF_ARCHON_HOST,use_shared)];
  start = YprefGetInt(DBE_ARCHON_START,1,use_shared);
  [chk_archon setState:(start) ? NSOnState : NSOffState];

  online = YprefGetInt(DBE_LAKE_ONLINE,DEF_STATE,use_shared);
  [popup_lake selectItemAtIndex:online];
  str = [TempController defaultHost];
  [edit_lake setStringValue:YprefGetString(@DBE_LAKE_HOST,str,use_shared)];
  start = YprefGetInt(DBE_LAKE_START,1,use_shared);
  [chk_lake setState:(start) ? NSOnState : NSOffState];

  online = YprefGetInt(DBE_GALIL_ONLINE,DEF_STATE,use_shared);
  [popup_galil selectItemAtIndex:online];
  [edit_galil setStringValue:YprefGetString(@DBE_GALIL_HOST,@DEF_GALIL_HOST,use_shared)];

  online = YprefGetInt(DBE_TCS_ONLINE,DEF_STATE,use_shared); 
  [popup_tcs selectItemAtIndex:online];
  [edit_tcs setStringValue:YprefGetString(@DBE_TCS_HOST,@DEF_TCS_HOST,use_shared)];

  [edit_email setStringValue:YprefGetString(@DBE_EMAIL,@DEF_EMAIL,use_shared)];

  online = YprefGetInt(DBE_TCPIP,0,use_shared); // TODO default=1 later
  [chk_tcpip setState:(online ? NSControlStateValueOn : NSControlStateValueOff)];

#ifndef SIM_ONLY
  [chk_disperser setHidden:YES];
#endif
#ifdef NDEBUG
  [edit_thinning setHidden:YES];
  [label_thinning setHidden:YES];
#endif

 // [self action_startup:nil]; // TESTING only

}

/* ---------------------------------------------------------------- */

- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication*)appl
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,appl);
#endif

  return YES;
}

/* ---------------------------------------------------------------- */

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication*)appl
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,appl);
#endif

  return NSTerminateNow;
}

/* ---------------------------------------------------------------- */

- (void)applicationWillTerminate:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,note);
#endif

  [main_logger append:PREFUN];
  [main_logger close];

  if ([panel_log isVisible]) {
    [panel_log saveFrameUsingName:@FRAME_MESSAGES];
  }
}

/* ---------------------------------------------------------------- */

- (void)fireTimer:(NSTimer*)theTimer
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,theTimer);
#endif
  assert([NSThread isMainThread]);

#ifdef NDEBUG
  [main_logger sync];
#endif
}

/* ---------------------------------------------------------------- */
#pragma mark "Actions (Menu)"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_startup:(id)sender
{
  int start;
  NSString *string;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  [panel_log setFrameUsingName:@FRAME_MESSAGES]; 
  [main_logger setTable:messageTable];
  [panel_log orderFront:self];
  [main_logger message:PREFUN];

  YprefPutString(@DBE_OBSERVER,[edit_observer stringValue],NO);
  
  string = [edit_archon stringValue];
  if ([string compare:@""] == NSOrderedSame) string = @DEF_ARCHON_HOST;
  YprefPutString(@DBE_ARCHON_HOST,string,use_shared);
  YprefPutInt(DBE_ARCHON_ONLINE,(int)[popup_archon indexOfSelectedItem],use_shared);
  start = ([chk_archon state] == NSOnState) ? 1 : 0;
  YprefPutInt(DBE_ARCHON_START,start,use_shared);
  if (start) {
    _camController = [[CameraController alloc] initWithWindowNibName:@"CameraController"];
  }

  _heaterWarning = YprefGetDouble(DBE_HEATER_WARNING,0.3,use_shared);
  _thinTime = YprefGetInt(DBE_THIN_TIME,1200,use_shared); // 20 [min]

  string = [edit_lake stringValue];
  if ([string compare:@""] == NSOrderedSame) string = [TempController defaultHost];
  YprefPutString(@DBE_LAKE_HOST,string,use_shared);
  YprefPutInt(DBE_LAKE_ONLINE,(int)[popup_lake indexOfSelectedItem],use_shared);
  start = ([chk_lake state] == NSOnState) ? 1 : 0;
  YprefPutInt(DBE_LAKE_START,start,use_shared);
  if (start) {
    _tempController = [[TempController alloc] initWithWindowNib:@"TempController"];
    YmenuEnableTitles("Temperature",YES);
  }
  
  string = [edit_galil stringValue];
  if ([string compare:@""] == NSOrderedSame) string = @DEF_GALIL_HOST;
  YprefPutString(@DBE_GALIL_HOST,string,use_shared);
  YprefPutInt(DBE_GALIL_ONLINE,(int)[popup_galil indexOfSelectedItem],use_shared);
  
  int online = (int)[popup_tcs indexOfSelectedItem];
  YprefPutInt(DBE_TCS_ONLINE,online,use_shared);
  string = [edit_tcs stringValue];
  if ([string compare:@""] == NSOrderedSame) string = @DEF_TCS_HOST;
  YprefPutString(@DBE_TCS_HOST,string,use_shared);

#ifdef NDEBUG
  _engMode = YprefGetInt(DBE_ENGMODE,0,NO);
#else
  _engMode = YprefGetInt(DBE_ENGMODE,0,NO);
  _disperser = YprefGetInt(DBE_DISPERSER,0,NO);
#endif

  if (_camController) {                // create galil hardhat
    assert(!_galilHH);
    _galilHH = [[GalilHardhat alloc] initWithWindowNibName:@"galilhardhat"];
    if ([popup_galil indexOfSelectedItem]) { // Galil online && eng-mode
      if (_engMode) [_galilHH.window orderFront:self];
    }
  }

  int run_server = ([chk_tcpip state] == NSControlStateValueOn) ? 1 : 0;
  YprefPutInt(DBE_TCPIP,run_server,use_shared);
  if (run_server) {
    int port = 50001+PROJECT_ID*100;   // 52801
    tcpip_server = [[TCPIP_Server alloc] initWithPort:port delegate:self];
#ifdef NDEBUG
    tcpip_server.throttle = 0.5f;      // max. 2 connections per second
#endif
    dispatch_async(GLOBAL_QUEUE,^(void) { char buf[256];
      int err = [tcpip_server start];
      if (err) {
        sprintf(buf,"%s: TCP/IP server failed, err=%d",P_TITLE,err);
        [main_logger append:buf];
        YnotifyNotice(buf,[[[tcpip_server class] errorString:err] UTF8String],0,2);
        // YprefPutInt(DBE_AUTOSTART,0,NO);
        // dispatch_sync(MAIN_QUEUE,^(void) { [NSApp terminate:self]; });
      } else {
        sprintf(buf,"TCPIP_Server running on port %d",port);
        [main_logger message:buf file:YES];  // Note: no message window
      }
    });
  } // endif(run_server)

  [NSTimer scheduledTimerWithTimeInterval:30.0
           target:self selector:@selector(fireTimer:)
           userInfo:nil repeats:YES];


  [_window orderOut:self];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_prefs:(id)sender
{
  int   i;
  char  title[64],buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  
  Ystr2chr([sender title],title,sizeof(title));
  
  if (strstr(title,"Preferences") || strstr(title,"Settings")) {   // open panel
    [edit_snd_expdone setStringValue:YprefGetString(@DBE_SND_EXPDONE,@DEF_SND_EXPDONE,NO)];
    [chk_eng setState:((_engMode) ? NSControlStateValueOn : NSControlStateValueOff)];
    [chk_disperser setState:((_disperser) ? NSControlStateValueOn : NSControlStateValueOff)];
    i = YprefGetInt(DBE_SATLEV,DEF_SATLEV,use_shared);
    sprintf(buf,"%d",i); YeditSet(edit_satlev,buf);
    sprintf(buf,"%d",_thinTime/60);  YeditSet(edit_thinning,buf);
    sprintf(buf,"%.1f",_heaterWarning); YeditSet(edit_heater,buf);
    [popup_refPixel selectItemAtIndex:_camController.refPixelMode]; 
    [chk_showRaw setState:((_camController.showRaw) ? NSControlStateValueOn : NSControlStateValueOff)];
    [chk_saveRaw setState:((_camController.saveRaw) ? NSControlStateValueOn : NSControlStateValueOff)];
    [chk_showFrames setState:((_camController.showFrames) ? NSControlStateValueOn : NSControlStateValueOff)];
    [chk_saveFrames setState:((_camController.saveFrames) ? NSControlStateValueOn : NSControlStateValueOff)];
    [chk_showCurrent setState:((_camController.showCurrent) ? NSControlStateValueOn : NSControlStateValueOff)];
    [panel_prefs makeKeyAndOrderFront:self];
  } else
  if (strstr(title,"Apply")) {           // apply Preferences
    YprefPutString(@DBE_SND_EXPDONE,[edit_snd_expdone stringValue],NO);
    _engMode = ([chk_eng state]==NSControlStateValueOn) ? 1 : 0;
    YprefPutInt(DBE_ENGMODE,_engMode,NO);
    _disperser = ([chk_disperser state]==NSControlStateValueOn) ? 1 : 0;
    YprefPutInt(DBE_DISPERSER,_disperser,NO);
    YprefPutInt(DBE_SATLEV,[edit_satlev intValue],use_shared);
    _thinTime = 60*imax(1,[edit_thinning intValue]);
    YprefPutInt(DBE_THIN_TIME,_thinTime,use_shared);
    _heaterWarning = [edit_heater floatValue];
    YprefPutDouble(DBE_HEATER_WARNING,_heaterWarning,use_shared);
    YprefPutString(@DBE_EMAIL,[edit_email stringValue],use_shared);
    YprefPutInt(DBE_REFPIXMODE,(int)[popup_refPixel indexOfSelectedItem],NO);
    YprefPutInt(DBE_SHOWRAW,(chk_showRaw.state==NSControlStateValueOn) ? 1 : 0,NO);
    YprefPutInt(DBE_SAVERAW,(chk_saveRaw.state==NSControlStateValueOn) ? 1 : 0,NO);
    YprefPutInt(DBE_SHOWFRAMES,(chk_showFrames.state==NSControlStateValueOn) ? 1 : 0,NO);
    YprefPutInt(DBE_SAVEFRAMES,(chk_saveFrames.state==NSControlStateValueOn) ? 1 : 0,NO);
    YprefPutInt(DBE_SHOWCURRENT,(chk_showCurrent.state==NSControlStateValueOn) ? 1 : 0,NO);
    [panel_prefs orderOut:self];
    [_camController reloadPreferences];
  } else
  if (strstr(title,"Browse")) {          // sound file
    NSURL *url = [NSURL fileURLWithPath:NSHomeDirectory() isDirectory:YES];
    NSOpenPanel *panel = [[NSOpenPanel openPanel] retain];
    [panel setDirectoryURL:url];
    [panel beginSheetModalForWindow:panel_prefs
           completionHandler:^(NSInteger returnCode)
      {
        if (returnCode == NSModalResponseOK) {
          NSArray* urls = [panel URLs];
          if ([urls count] > 0) {
            if (sender == but_snd_expdone) {
              [edit_snd_expdone setStringValue:[[urls objectAtIndex:0] path]];
            }
          }
          [panel release];
        }
      } // end-block
    ];
  } else
  if (strstr(title,"Test")) { char buf[128],tstr[32];  // send test email
    YprefPutString(@DBE_EMAIL,[edit_email stringValue],use_shared);
    sprintf(buf,"This is just a test (UT=%s)",get_ut_datestr(tstr,0));
    [self emailText:buf subject:"Test of the Henrietta Alert System"];
#if 0 // TEST sharing service
    NSSharingService *service = [NSSharingService sharingServiceNamed:NSSharingServiceNameComposeEmail];
    NSLog(@"%p",service);
    NSArray *items = [NSArray arrayWithObjects:@"This is a test message",nil];
    NSLog(@"%d",[service canPerformWithItems:items]);
    service.recipients = [NSArray arrayWithObjects:@"birk@carnegiescience.edu",nil];
    service.subject = @"This is a test subject";
    // service.delegate = self;
    [service performWithItems:items];
#endif
#if 0 // TEST mailto:
    NSString *string = [NSString stringWithFormat:@"mailto:%@?Subject=%@&body=%@",
      @"birk@yahoo.com",@"This is a test subject",@"This is a test message"];
    string = [string stringByReplacingOccurrencesOfString:@" " withString:@"%20"];
    NSLog(@"%@",string);
    [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:string]];
#endif
  } else {
    assert(0);
  }
}


/* ---------------------------------------------------------------- */

- (IBAction)action_file:(id)sender
{
  char title[64];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  
  Ystr2chr([sender title],title,sizeof(title));
  
  if (strstr(title,"Logfile")) {       // show Logfile(s)
#if (MAC_OS_X_VERSION_MIN_REQUIRED > MAC_OS_X_VERSION_10_13)
    [[NSWorkspace sharedWorkspace] openURL:[NSURL fileURLWithPath:main_logger.logfile]];
#else
    [[NSWorkspace sharedWorkspace] openFile:main_logger.logfile];
#endif
  } else
  if (strstr(title,"Data")) {          // data folders
    NSArray *a = [self dataPaths];
    for (NSString *p in a) {
      if ([p compare:@""] != NSOrderedSame) {
        [[NSWorkspace sharedWorkspace] openURL:[NSURL fileURLWithPath:p isDirectory:YES]];
      }
    }
  } else
  if (strstr(title,"AppSupport")) {    // Application Support
    NSURL *url = [NSURL fileURLWithPath:@(s_path) isDirectory:YES];
    [[NSWorkspace sharedWorkspace] openURL:url];
  } else
  if (strstr(title,"Load")) {          // load FITS file/sky
    BOOL sky = strstr(title,"FITS") ? NO : YES;
    if (_camController.loop_doing && !sky) {
      [Notice openNotice:"cannot load FITS files during exposure loop" title:p_exec];
    } else {
      NSArray *a = [self dataPaths];
      NSURL *url = [NSURL fileURLWithPath:a[0] isDirectory:YES];
      NSOpenPanel *panel = [[NSOpenPanel openPanel] retain];
      [panel setAllowedFileTypes:@[@"fits"]];
      // [panel setAllowsMultipleSelection:YES]; -- NO
      [panel setDirectoryURL:url];
      [panel beginSheetModalForWindow:[_camController window]
             completionHandler:^(NSInteger returnCode)
      {
        if (returnCode == NSModalResponseOK) {
          [_camController loadFITS:[[panel URLs] firstObject] asSky:sky];
        }
        [panel release];
      }];
    }
  } else {
    assert(0);
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_datapath:(id)sender
{
  char title[64];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,[[sender title] UTF8String]);
#endif
  assert(N_FOLDERS == 2);
  assert([NSThread isMainThread]);

  Ystr2chr([sender title],title,sizeof(title));

  if (strstr(title,"DataPath")) {      // open datapath window
    NSArray *a = [self dataPaths];
    assert([a count] == N_FOLDERS);
    assert(N_FOLDERS == 2);
    [edit_path1 setStringValue:a[0]];
    [edit_path2 setStringValue:a[1]];
    [panel_datapath makeKeyAndOrderFront:self];
  } else
  if (!strcmp(title,"Default")) {      // default datapath1
    time_t ut; struct tm res1,res2; char dir[64],buf[128]; NSString *path;
    ut = cor_time(0) - 9*3600; gmtime_r(&ut,&res1); // today
    ut = ut + 24*3600;         gmtime_r(&ut,&res2); // tomorrow
    strftime(dir,sizeof(dir),"ut%Y%m%d",&res1);
    strftime(buf,sizeof(buf),"_%d",&res2);
    strcat(dir,buf);
#if defined(SIM_ONLY) || !defined(NDEBUG)
    path = [NSString stringWithFormat:@"%@/Data/Swope/Henrietta/%s",NSHomeDirectory(),dir];
#else
    path = [NSString stringWithFormat:@"/Volumes/DATA_SWOPE/Henrietta/%s",dir];
#endif
    [edit_path1 setStringValue:path];
  } else                               // browse datapath2
  if (!strcmp(title,"Browse")) {
    NSOpenPanel* panel = [[NSOpenPanel openPanel] retain];
    [panel setCanChooseFiles:NO];
    [panel setCanChooseDirectories:YES];
    [panel setCanCreateDirectories:YES];
    // default [panel setAllowsMultipleSelection:NO];
    [panel beginSheetModalForWindow:panel_datapath
           completionHandler:^(NSInteger returnCode)
    {
      if (returnCode == NSModalResponseOK) {
        assert([NSThread isMainThread]);
        [edit_path2 setStringValue:[[panel directoryURL] path]];
      }
      [panel release];
    }];
  } else
  if (!strcmp(title,"Apply")) { int n=0; char path1[512],path2[512];
    YeditGet(edit_path1,path1,sizeof(path1));
    cut_spaces(path1);
    if ((*path1 == '\0') || (verify_path(path1,YES) == 0)) {
      if (*path1) n++;
      YprefPutChar(DBE_DATAPATH1,path1,NO);
    }
    YeditGet(edit_path2,path2,sizeof(path2));
    cut_spaces(path2);
    if (!strcmp(path2,path1)) *path2='\0';
    if ((*path2 == '\0') || (verify_path(path2,YES) == 0)) {
      if (*path2) n++;
      YprefPutChar(DBE_DATAPATH2,path2,NO);
    }
    if (n == 0) [Notice openNotice:"no data path" title:P_TITLE time:0 level:1];
    [panel_datapath orderOut:self];    // hide window
    [_camController reloadDatapath];
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_camera:(id)sender
{
  char title[64];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  Ystr2chr([sender title],title,sizeof(title));

  if (strstr(title,"Look")) {          // QuickLook tool
    [_camController showQltool];
  } else
  if (strstr(title,"Archon")) {        // show Archon hardhat
    if (!self.engMode) {               // require eng.mode ?Nick
      [Notice openNotice:engWarning title:P_TITLE time:15];
    } else {
      [_camController.panel_sta makeKeyAndOrderFront:self];
    }
  } else
  if (strstr(title,"Galil")) {
    if (!self.engMode) {               // require eng.mode ?Nick
      [Notice openNotice:engWarning title:P_TITLE time:15];
    } else {
      if (!_galilHH) {
        _galilHH = [[GalilHardhat alloc] initWithWindowNibName:@"galilhardhat"];
      }
      [_galilHH.window orderFront:self];
    }
  } else
  if (strstr(title,"Number")) {        // enable edit file number
    [_camController action_file:self];
  } else {
    assert(0);
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_temperature:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,sender,[[sender title] UTF8String]);
#endif

  [_tempController action_menu:sender];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_help:(id)sender
{
  char title[64];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  Ystr2chr([sender title],title,sizeof(title));

  if (strstr(title,"Web")) {
    // TODO Ywww("http://users.obs.carnegiescience.edu/crane/pfs");
  } else {
    Ywww("http://instrumentation.obs.carnegiescience.edu/Software/Henrietta/");
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Utils"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (NSArray*)dataPaths               
{
  assert(N_FOLDERS == 2);
  
  NSString *path1 = YprefGetString(@DBE_DATAPATH1,NSHomeDirectory(),NO);
  if (!YpathWritable([path1 UTF8String])) path1 = NSHomeDirectory();
  assert(path1);
  
  NSString *path2 = YprefGetString(@DBE_DATAPATH2,@"",NO);
  if ([path2 compare:@""] != NSOrderedSame) {
    if (!YpathWritable([path2 UTF8String])) path2 = @"";
  }
  assert(path2);

  return @[path1,path2];
}

/* ---------------------------------------------------------------- */

- (void)emailFile:(NSString*)file to:(NSString*)addr subject:(NSString*)subj
{
#ifdef NDEBUG // -- requires 'postfix' configured -- tested 2018-02-07 [CCB]
  NSString *cmd = [NSString stringWithFormat:@"mail -s \'%@\' %@ < %@ &",subj,addr,file];
  system([cmd UTF8String]);
#else // requires a valid email address setup with the current account
  NSString *script;  FILE *fp;
  @synchronized(self) {
    script = [NSString stringWithFormat:@"/tmp/%s-mail-%02x.scpt",p_ident,(++counter & 0x3f)];
  }
  if ((fp=fopen([script UTF8String],"w")) == NULL) {
    NSLog(@"failed to open %@",script);
    return;
  }
  fprintf(fp,"set recipientName to \"%s\"\n",P_TITLE);
  fprintf(fp,"set recipientAddress to \"%s\"\n",[addr UTF8String]);
  fprintf(fp,"set theSubject to \"%s\"\n",[subj UTF8String]);
  fprintf(fp,"set txt to (read POSIX file \"%s\")\n",[file UTF8String]);
  fprintf(fp,"tell application \"Mail\"\n");
  fprintf(fp,"\tset theMessage to make new outgoing message with properties {subject:theSubject, content:txt, visible:true}\n");
  fprintf(fp,"\ttell theMessage\n");
  fprintf(fp,"\t\tmake new to recipient with properties {name:recipientName, address:recipientAddress}\n");
  fprintf(fp,"\t\tsend\n");
  fprintf(fp,"\tend tell\n");
  fprintf(fp,"end tell\n");
  fclose(fp);
  YfilePermissions([script UTF8String],6,6,6);
  NSString *cmd = [NSString stringWithFormat:@"osascript %@",script];
  system([cmd UTF8String]);
#endif
}

/* --- */

- (void)emailFile:(NSString*)file subject:(NSString*)subj
{
  NSString *ad0 = YprefGetString(@DBE_EMAIL,@DEF_EMAIL,use_shared);
  NSString *adr = [ad0 stringByReplacingOccurrencesOfString:@" " withString:@","];
  NSArray  *ary = [adr componentsSeparatedByString:@","];
  NSString *str = [NSString stringWithFormat:@"email '%@' to '%@'",subj,adr];
  [main_logger append:[str UTF8String]];
  
  dispatch_async(GLOBAL_QUEUE,^(void) {
    for (NSString *adr in ary) {
      adr = [adr stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
      if ([adr compare:@""] == NSOrderedSame) continue;
      NSRange range = [adr rangeOfString:@"@"];
      if (range.location == NSNotFound) adr = [NSString stringWithFormat:@"%@@carnegiescience.edu",adr];
      [self emailFile:file to:adr subject:subj];
    }
  });
}

/* --- */

- (void)emailText:(const char*)text subject:(const char*)subj
{
  NSString *file;
  FILE     *fp;

  @synchronized(self) {
    file = [NSString stringWithFormat:@"/tmp/%s-mail-%02x.txt",p_ident,(++counter & 0x3f)];
  }
  if ((fp=fopen([file UTF8String],"w")) == NULL) {
    NSLog(@"failed to open %@\n",file);
  } else {
    fprintf(fp,"%s\n",text);
    fclose(fp);
    YfilePermissions([file UTF8String],6,6,6);
    [self emailFile:file subject:@(subj)];
  }
}

/* --- */

- (void)emailText:(const char*)text
{
  [self emailText:text subject:"Henrietta Alert"]; 
}

/* ---------------------------------------------------------------- */

// level=0: prepend "Info", open dialog-box, write to logfile
//      =1: prepend "Notice"  + email to 'fsguru'
//      =2: prepend "Alert"   + email to 'email_address' list
//      >2: prepend "WARNING" + email to 'email_address' list

- (void)alert:(const char*)text level:(int)level dt:(int)dt // TODO unused
{
  char pretext[16];

  switch (level) {
    case 0:  strcpy(pretext,"Info"); break;
    case 1:  strcpy(pretext,"Notice"); break;
    case 2:  strcpy(pretext,"Alert"); break;
    default: strcpy(pretext,"WARNING"); break;
  }
  NSString *string = [NSString stringWithFormat:@"%s: %s",pretext,text];
  [self message:string level:1 time:dt];
  
  string = @(text);
  if (level > 0) {
    dispatch_async(GLOBAL_QUEUE,^(void) {
      if (level == 1) {
        [self emailText:[string UTF8String] subject:"Henrietta Notice"];
      } else {
        [self emailText:[string UTF8String] subject:"Henrietta Alert"];
      }
    });
  }
  [Sound say:string]; 
}

/* ---------------------------------------------------------------- */

- (void)message:(NSString*)text level:(int)lev time:(int)dt
{
  [main_logger message:[text UTF8String] level:lev time:dt]; 
}

/* --- */

- (void)message:(NSString*)text level:(int)lev
{
  [self message:text level:lev time:0];
}

/* ---------------------------------------------------------------- */
#pragma mark "TCPIP_Server delegate"
/* ---------------------------------------------------------------- */

- (char*)tcpip_handler:(const char*)command from:(NSString*)ipNumber socket:(int)msgsock
{
  char cmd[256]="",par1[256]="",par2[256]="",par3[256]="";
  NSString *str1=nil;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%s,%d)\n",PREFUN,command,[ipNumber UTF8String],msgsock);
#endif
  assert(![NSThread isMainThread]);

  NSString *string = @(command);
  NSString *line = [string stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];

  char *answer = (char*)calloc(512,sizeof(char));
  if (strlen(command) > 128) return answer;
  
  string = [NSString stringWithFormat:@"%@-> %@",ipNumber,line];
  [main_logger append:[string UTF8String]];
  sscanf([line UTF8String],"%s %s %s %s",cmd,par1,par2,par3);
  if (*par1) str1 = @(par1);
  
  if (!strcmp(cmd,"version")) {        // version check -----------
    sprintf(answer,"%s (%s)",p_version,p_build);
  } else
  if (!strcmp(cmd,"exptime")) {        // exposure time NEW b0097
    [_camController action_exptime:str1];
    sprintf(answer,"%s",_camController.tcpResponse.UTF8String);
  } else
  if (!strcmp(cmd,"start")) {          // start exposure NEW b0096
    dispatch_sync(MAIN_QUEUE,^(void) {
      if (str1) [_camController action_nloops:str1]; // NEW b0098
      [_camController action_start:nil];
    });
    sprintf(answer,"%s",_camController.tcpResponse.UTF8String);
  } else
  if (!strcmp(cmd,"status")) {         // instrument status NEW b0096
    sprintf(answer,"%d %d",_camController.loop_doing,[_camController motorMoving]);
  } else {                             // unknown command
    sprintf(answer,"unknown command '%s'",cmd);
    Ysleep(1000);                      // throttle
  }
  assert(answer);

  string = [NSString stringWithFormat:@"%@<- %s",ipNumber,answer];
  [main_logger append:[string UTF8String]];
  strcat(answer,"\n");                 // terminating <LF>
#ifdef NDEBUG
  Ysleep(150);                         // throttle commands
#endif

  return answer;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
