/* ---------------------------------------------------------------- */
//
// ArchonController.m
// PFS
//
// Created by Christoph Birk on 2015-07-30
// Copyright (c) 2015 Christoph Birk. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG           1
#define TIME_TEST       1
#define UNIT_TEST       0              // 1:ArchonTimer

/* ---------------------------------------------------------------- */

#import  "ArchonController.h"
#import  "DataSimulator.h"
#import  "AppDelegate.h"
#import  "Notice.h"
#import  "Logger.h"
#import  "MTwister.h"
#import  "Yutils.h"
#ifdef LOG_PSU
#import  "GraphController.h"
#endif

#include "utils.h"

/* ---------------------------------------------------------------- */

#define STA_NTAPS       N_CHANNELS
#define STA_NPIXELS     (64)          // 4CHAN:(8*)
#define STA_NLINES      (STA_NTAPS*STA_NPIXELS)

#define DBE_BACKVERSION "dbe_backplane_version"

#define DEF_ACF         "henrietta.acf"

/* ---------------------------------------------------------------- */

#if 0 // unused
static NSTextField* editCreate(NSView* view,int x,int y,int w)
{
  assert([NSThread isMainThread]);
  
  NSTextField *edit = [[NSTextField alloc] initWithFrame:NSMakeRect(x,y,w,22)];
  [view addSubview:edit];

  return [edit autorelease];
}
#endif

/* --- */

static NSTextField* outputCreate(NSView* view,int x,int y,int w)
{
  assert([NSThread isMainThread]);
  
  NSTextField *edit = [[NSTextField alloc] initWithFrame:NSMakeRect(x,y,w,22)];
  [edit setEditable:NO];
  [edit setBezeled:NO];
  [edit setBordered:YES];
  [edit setDrawsBackground:NO];
  [view addSubview:edit];

  return [edit autorelease];
}

/* --- */

static NSTextField* labelCreate(NSView* view,int x,int y,int w,const char* text)
{
  assert([NSThread isMainThread]);
  
  NSTextField *edit = [[NSTextField alloc] initWithFrame:NSMakeRect(x,y,w,17)];
  [edit setSelectable:NO];
  [edit setBordered:NO];
  [edit setDrawsBackground:NO];
  [edit setAlignment:NSTextAlignmentLeft];
  [edit setStringValue:@(text)];
  [view addSubview:edit];
  
  return [edit autorelease];
}

/* ---------------------------------------------------------------- */

@interface ArchonController ()        // private extensions
{
}
@property (readwrite,atomic) BOOL configured,aborted;
@end

/* ---------------------------------------------------------------- */

@implementation ArchonController
{
  ArchonIO *archon;
  int      last_frame,fetch_buffer;
  time_t   next_back_warning,add_back_warning;
  float    back_warn_temp;
  NSImage  *icon_burning,*icon_caution;
  int      power_warning_sent;
  float    lvlc_v[LVLC_N],lvhc_v[LVHC_N]; // nominal voltages b0090 -- not used anywhere yet
  DataSimulator *dataSim;
#ifdef LOG_PSU
  GraphController *curGraph;
#endif
  NSMutableArray *configuration;
}

static const char *exposuresPar="Exposures";
static const char *resetsPar="Resets";
static const char *dummiesPar="Dummies";     // b0092
static const char *mainResetPar="MainReset";
static const char *startClockingPar="StartClocking";

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
  case E_sta_error: string = @"Archon: generic error"; break;
  case E_sta_power: string = @"Archon: not powered"; break;
  case E_sta_config: string = @"Archon: loading .acf file failed"; break;
  case E_sta_parameter: string = @"Archon: unknown parameter"; break;
  default:
    if (!string) string = [ArchonIO errorString:err];
    break;
  }

  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)init
{
  const int port=4242;
  char host[128],buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  if (!(self = [super init])) return nil;
  
  YprefGetChar(DBE_ARCHON_HOST,DEF_ARCHON_HOST,host,sizeof(host),use_shared);
#ifdef SIM_ONLY
  _online = 0;
#else
  _online = YprefGetInt(DBE_ARCHON_ONLINE,1,use_shared);
#endif
  sprintf(buf,"archon_host= %s:%d (%d)",host,port,_online); [main_logger append:buf];
  archon = [[ArchonIO alloc] initWithHost:host port:port online:_online];
  
  configuration = [[NSMutableArray alloc] initWithCapacity:1024];

  _configured = FALSE;
  _ntaps = STA_NTAPS;                  // array geomentry
  _dimx  = STA_NTAPS * STA_NPIXELS;
  _dimy  = STA_NLINES;
  
  _interval = 2;
  back_warn_temp = (float)YprefGetDouble("dbe_backplane_warn_temp",45.0,use_shared);
  power_warning_sent = -1;
  
  resp_n = 1024;
  resp_head = 0;
  respArray = [[NSMutableArray alloc] initWithCapacity:resp_n];

  NSBundle *bundle = [NSBundle mainBundle];
  NSString *file = [bundle pathForResource:@"BurningIcon" ofType:@"icns"];
  icon_burning = [[NSImage alloc] initWithContentsOfFile:file];
  file = [bundle pathForResource:@"AlertCautionIcon" ofType:@"icns"];
  icon_caution = [[NSImage alloc] initWithContentsOfFile:file];
  
#ifdef LOG_PSU
  const double VAL=0.65;
  const GraphColor myColors[7] = {
    { 0.0, 0.0, 0.0 } ,   // black       2.5
    { VAL, 0.0, 0.0 } ,   // red         5.0
    { 0.0, 0.0, VAL } ,   // blue        +6
    { 0.0, VAL, 0.0 } ,   // green       -6
    { 0.0, VAL, VAL } ,   // turquoise   +17
    { VAL, VAL, VAL } ,   // grey        -17
    { VAL, 0.0, VAL } ,   // purple      +35
  };
  curGraph = [[GraphController alloc] initWithWindowNib:@"Graph"
                context:0 title:"Archon Currents" unit:"A" N:7 colors:myColors];
#endif
  
  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
  [_acf_file release];
  [_acf_version release];

  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (void)awakeFromNib
{
  int i;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert([NSThread isMainThread]);
  assert(panel);
  
  [responseTable setDataSource:self]; 

  assert(STA_NPIXELS*STA_NTAPS == STA_NLINES); // square array
  _dataPixels = STA_NPIXELS;        // 64
  _dataLines = STA_NLINES;          // 32*64=2048

  dispatch_async(GLOBAL_QUEUE,^(void) {
    int err = [archon open];        // keep open
    if (!err) {
      if (!err) err = [archon getSystem];
      if (!err) err = [archon getStatus];
      if (!err) [self uploadConfiguration:nil]; // v0069
    }
    // [archon close];              // NOTE: keep open
    if (err) {
      [AppDelegate goodbye:@"connecting to Archon failed" error:err];
    }
  });

  for (i=0; i<LVLC_N; i++) {        // setup bias panel b0083
    label_lvlc[i] =  labelCreate([panel_biases contentView], 18,25+463-i*25,90,"");
    edit_lvlcP[i] = outputCreate([panel_biases contentView],118,25+463-i*25,50);
    edit_lvlcV[i] = outputCreate([panel_biases contentView],176,25+463-i*25,50);
    edit_lvlcI[i] = outputCreate([panel_biases contentView],234,25+463-i*25,50);
  }
  for (i=0; i<LVHC_N; i++) {
    label_lvhc[i] =  labelCreate([panel_biases contentView], 18,145-i*25,90,"");
    edit_lvhcP[i] = outputCreate([panel_biases contentView],118,145-i*25,50);
    edit_lvhcV[i] = outputCreate([panel_biases contentView],176,145-i*25,50);
    edit_lvhcI[i] = outputCreate([panel_biases contentView],234,145-i*25,50);
  }
  
#ifdef LOG_PSU
  [curGraph.window orderFront:self];
#endif
  assert([NSThread isMainThread]);
  [NSTimer scheduledTimerWithTimeInterval:2.0 target:self
           selector:@selector(fireTimer:) userInfo:nil repeats:YES];

#if (UNIT_TEST == 1)                    // test Archon timer drift vs internal clock
  dispatch_async(GLOBAL_QUEUE,^(void) { // RESULT: -0.2 sec/hr
    [archon open];
    NSDate *date1 = [NSDate date];
    [archon getTimer];
    double t1 = archon.uptime;
    for (int i=0; i<900; i++) {
      Ysleep(2000);
      NSDate *date2 = [NSDate date];
      int err = [archon getTimer];
      if (!err) {
        double dsta = archon.uptime - t1;
        double dcpu = [date2 timeIntervalSinceDate:date1];
        printf("%s(): %.3f %.3f (%.3f sec/hr)\n",PREFUN,dsta,dcpu,3600.0*(dsta-dcpu)/dcpu);
      }
    }
    [archon close];
  });
#endif
}

/* --- */

- (NSString*)uploadConfiguration:(NSString*)file // v0069
{
  int  i=0;
  char path[512]="";
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,file.UTF8String);
#endif

  if (file) {                           // has parameter
    strcpy(path,file.UTF8String);
    i = [self loadConfigurationFromFile:path];
    if (!i) {
      NSString *s = [NSString stringWithFormat:@"loading %s failed",path];
      [main_logger message:s.UTF8String level:1];
    }
  }
  if (i == 0) {                         // failure to parse
    strcpy(path,DEF_ACF);               // fall back to default
    i = [self loadConfigurationFromFile:path];
  }
  if (i == 0) {                          // failed again
    NSString *text = [NSString stringWithFormat:@"invalid file %s",path];
    [AppDelegate goodbye:text error:E_sta_config];
    return nil;
  }
  
  // Archon/H2RG initialization sequence
  // 1) upload .acf file
  // 2) APPLYALL -- turns off power and runs VCPU
  // 3) POWERON -- turn on biases
  // 4) send "MainReset=1" to Archon -- toggles MAINRESETB
  // 5) APPLYDIO0A  -- runs VCPU -- initializes H2RG registers
  // 6) send "StartClocking=1" to Archon -- starts clocking H2RG
  int err = [archon uploadConfiguration:configuration];  // APPLYALL
  if (!err) Ysleep(20);
  if (!err) err = [archon command:"POWERON"];    // power on H2RG
  if (!err) Ysleep(200);
  if (!err) err = [archon setParameterFast:mainResetPar to:1]; // reset H2RG
  if (!err) Ysleep(20);
  if (!err) err = [archon command:"APPLYDIO0A"]; // run VCPU -- init H2RG registers
  if (!err) err = [archon setParameterFast:startClockingPar to:1];

  if (!err) [self statusCheck:YES];

  if (err) {
    [AppDelegate goodbye:@"Archon startup failed" error:err];
  } else { NSString *str;
    str = [NSString stringWithFormat:@"Archon Backplane %@",archon.back_version];
    [main_logger message:[str UTF8String] file:YES];
    NSString *version = YprefGetString(@DBE_BACKVERSION,@"",use_shared);
    if ([archon.back_version compare:version] != NSOrderedSame) {
      str = [NSString stringWithFormat:@"unknown backplane version '%@'",archon.back_version];
      [Notice openNotice:[str UTF8String] title:P_TITLE];
      YprefPutString(@DBE_BACKVERSION,archon.back_version,use_shared);
    }
    self.configured = TRUE;
    YeditSet(edit_acf,path);
  }

  return @(path);
}

/* ---------------------------------------------------------------- */

- (void)fireTimer:(NSTimer*)theTimer
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,theTimer);
#endif

  if (!self.configured) return;
  if (self.fetching) return;

  BOOL bias_visible = [panel_biases isVisible]; 
#ifndef LOG_PSU
  if (!bias_visible) return;
#endif

  dispatch_async(GLOBAL_QUEUE,^(void) {
    int err = [archon getStatus]; if (err) return;
#ifdef LOG_PSU
    time_t ut = cor_time(0);
    [curGraph addData:0 time:ut value:archon.p2v5_i redraw:NO];
    [curGraph addData:1 time:ut value:archon.p5v_i redraw:NO];
    [curGraph addData:2 time:ut value:archon.p6v_i redraw:NO];
    [curGraph addData:3 time:ut value:archon.n6v_i redraw:NO];
    [curGraph addData:4 time:ut value:archon.p17v_i redraw:NO];
    [curGraph addData:5 time:ut value:archon.n17v_i redraw:NO];
    [curGraph addData:6 time:ut value:archon.p35v_i redraw:YES];
#endif
    if (bias_visible) {
      dispatch_async(MAIN_QUEUE,^(void) {    // b0083
        for (int i=0; i<LVLC_N; i++) {
          [edit_lvlcV[i] setStringValue:[NSString stringWithFormat:@"%6.3f",archon.lvlc_v[i]]];
          [edit_lvlcI[i] setStringValue:[NSString stringWithFormat:@"%6.3f",archon.lvlc_i[i]]];
        }
        for (int i=0; i<LVHC_N; i++) {
          [edit_lvhcV[i] setStringValue:[NSString stringWithFormat:@"%6.3f",archon.lvhc_v[i]]]; // b0090
          [edit_lvhcI[i] setStringValue:[NSString stringWithFormat:@"%6.1f",archon.lvhc_i[i]]]; 
        }
      });
    }
  });
}

/* ---------------------------------------------------------------- */

- (int)loadConfigurationFromFile:(const char*)file
{
  int   i;
  float f;
  char  buf[512],key[128],*p;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,file);
#endif

  [configuration removeAllObjects];
  
  if (file[0] == '/')  strcpy(buf,file);  // absolute path
  else                sprintf(buf,"%s/%s",r_path,file);
  FILE *fp = fopen(buf,"r");
  if (fp) { BOOL flag=FALSE;
    while (fgets(buf,sizeof(buf),fp) != NULL) {
      if (flag) {
        if (strlen(buf) <= 4) break;      // end of [CONFIG] section
        // printf("%s",buf);
        p = buf;
        while (*p) {   // re-format from strange file format
          if      (*p == '\n') *p = '\0'; // remove <LF>
          else if (*p == '\r') *p = '\0'; // remove <CR> (DOS artifact)
          else if (*p == '"')  *p = ' ';  // remove quotes
          else if (*p == '\\') *p = '/';  // replace backslash
          p++;
        }
        [configuration addObject:@(buf)];
        if ((p=strstr(buf,"VERSION "))) {
          [main_logger message:p];
          p = strchr(p,' ');
          if (p) self.acf_version = @(p+1);  // NEW b0095
        }
        if (!strncmp(buf,"MOD11/LV",8)) { // get nominal bias voltages b0082
          if (sscanf(buf,"MOD11/LVLC_V%d=%f",&i,&f)) {
            if (i <= LVLC_N) {            // low-voltage, low-current
              sprintf(key,"%6.3f",f); YeditSet(edit_lvlcP[i-1],key);
              lvlc_v[i-1] = f;
            }
          }
          if (sscanf(buf,"MOD11/LVHC_V%d=%f",&i,&f)) {
            if (i <= LVHC_N) {            // low-voltage, high-current
              sprintf(key,"%6.3f",f); YeditSet(edit_lvhcP[i-1],key);
              lvhc_v[i-1] = f;
            }
          }                               // setup bias labels
          if (sscanf(buf,"MOD11/LVLC_LABEL%d=%s",&i,key)) {
            if (i <= LVLC_N) YeditSet(label_lvlc[i-1],key);
          }
          if (sscanf(buf,"MOD11/LVHC_LABEL%d=%s",&i,key)) {
            if (i <= LVHC_N) YeditSet(label_lvhc[i-1],key);
          }
        }
      } else {
        if (strstr(buf,"[CONFIG]")) flag = TRUE; // start of [CONFIG] section
      }
    }
    fclose(fp);
    self.acf_file = @(file);
  }
  sprintf(buf,"loading '%s': %lu lines",file,[configuration count]);
  [main_logger message:buf];
  
  return (int)[configuration count];
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)getParameter:(const char*)par
{
  int      i,r=-1;
  NSString *name;
  
  NSString *para = @(par);
  for (i=0; i<[configuration count]; i++) {
    NSString *line = [configuration objectAtIndex:i];
    NSArray *array = [line componentsSeparatedByString:@"="];
    assert([array count] >= 2);
    if ([array count] < 3) continue;  // CONSTANTxx, PARAMETERxx
    NSRange range = [[array objectAtIndex:0] rangeOfString:@"PARAMETER"];
    if (range.location == NSNotFound) continue;
    name = [[array objectAtIndex:1] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
    if ([name compare:para] == NSOrderedSame) {
      r = [[array objectAtIndex:2] intValue];
      break;
    }
  }
#ifndef NDEBUG
  if (i == [configuration count]) fprintf(stderr,"%s(): unknown '%s'\n",PREFUN,par);
#endif
  
  return r;
}

/* --- */

- (int)setParameter:(const char*)par to:(int)value
{
  int      i,err=E_sta_parameter;
  NSString *parn,*name;
#if (DEBUG > 1) 
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,par,value);
#endif

  NSString *para = @(par);
  for (i=0; i<[configuration count]; i++) {
    NSString *string = [configuration objectAtIndex:i];
    NSArray *array = [string componentsSeparatedByString:@"="];
    assert([array count] >= 2);
    if ([array count] < 3) continue;  // PARAMETERxx (CONSTANTxx)
    parn = [[array objectAtIndex:0] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
    NSRange range = [parn rangeOfString:@"PARAMETER"];
    if (range.location == NSNotFound) continue;
    name = [[array objectAtIndex:1] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
    if ([name compare:para] == NSOrderedSame) {
      NSString *line = [NSString stringWithFormat:@"%@=%@=%d",parn,name,value];
      [configuration replaceObjectAtIndex:i withObject:line];
      err = [archon open];
      if (!err) { NSString *cmd;
        cmd = [NSString stringWithFormat:@"WCONFIG%04X%@",i,line];
        [main_logger append:[cmd UTF8String]];
        if (!err) err = [archon command:[cmd UTF8String]];
        cmd = [NSString stringWithFormat:@"LOADPARAM %@",name];
        [main_logger append:[cmd UTF8String]];
        if (!err) err = [archon command:[cmd UTF8String]];
        [archon close];
      }
      break;
    }
  }
#ifndef NDEBUG
  if (i == [configuration count]) fprintf(stderr,"%s(): unknown '%s'\n",PREFUN,par);
#endif
  
  return err;
}

/* ---------------------------------------------------------------- */

#if 0 // ununsed with Henrietta
- (int)getConfig:(const char*)config
{
  int      i,r=-1;
  NSString *name;

  NSString *conf = @(config);
  for (i=0; i<[configuration count]; i++) {
    NSString *string = [configuration objectAtIndex:i];
    NSArray *array = [string componentsSeparatedByString:@"="];
    assert([array count] >= 2);
    if ([array count] != 2) continue;
    name = [[array objectAtIndex:0] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
    if ([name compare:conf] == NSOrderedSame) {
      r = [[array objectAtIndex:1] intValue];
      break;
    }
  }
#ifndef NDEBUG
  if (i == [configuration count]) fprintf(stderr,"%s(): unknown '%s'\n",PREFUN,config);
#endif

  return r;
}
#endif

/* --- */

#if 0 // unused with Henrietta
- (int)setConfig:(const char*)config to:(int)value 
{
  int      i,err=E_sta_parameter;
  NSString *name;

  NSString *conf = @(config);
  for (i=0; i<[configuration count]; i++) {
    NSString *string = [configuration objectAtIndex:i];
    NSArray *array = [string componentsSeparatedByString:@"="];
    assert([array count] >= 2);
    if ([array count] != 2) continue;
    name = [[array objectAtIndex:0] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
    if ([name compare:conf] == NSOrderedSame) {
      NSString *line = [NSString stringWithFormat:@"%@=%d",name,value];
      [configuration replaceObjectAtIndex:i withObject:line];
      NSString *cmd = [NSString stringWithFormat:@"WCONFIG%04X%@",i,line];
      err = [archon command:[cmd UTF8String]];
      break;
    }
  }
#ifndef NDEBUG
  if (i == [configuration count]) fprintf(stderr,"%s(): unknown '%s'\n",PREFUN,config);
#endif

  return err;
}
#endif

/* ---------------------------------------------------------------- */

- (void)showWarning:(int)err from:(const char*)fun 
{
  NSString *string = [NSString stringWithFormat:@"%s failed, err=%d\r%@",
    fun,err,[[self class] errorString:err]];
  [Notice openNotice:[string UTF8String] title:P_TITLE time:60 level:1];
  [main_logger append:[string UTF8String]];
}

/* --- */

- (void)showError:(int)err from:(const char*)fun
{
  NSString *string = [NSString stringWithFormat:@"%s failed, err=%d\r%@",
    fun,err,[[self class] errorString:err]];
  [Notice openNotice:[string UTF8String] title:P_TITLE time:0 level:2];
  [main_logger append:[string UTF8String]];
}

/* ---------------------------------------------------------------- */

- (void)setupSimulator:(float)ps
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%f)\n",PREFUN,ps);
#endif
  assert(!dataSim);

  dataSim = [[DataSimulator alloc] initWithPixscale:ps];
#if 1 // requires complicated REF_PIXEL scheme
  float w = ps * STA_NPIXELS;          // width per tap
  for (int i=0; i<STA_NTAPS; i++) {
    float x = w*(i-(STA_NTAPS-1)/2.0);
    int lr = (i==0) ? YES : NO;
    int rr = (i==(STA_NTAPS-1)) ? YES : NO;
    [dataSim addArrayW:STA_NPIXELS H:STA_NLINES X:-x Y:0 L:lr R:rr];
  }
#else // simplified Simulator
  [dataSim addArrayW:STA_NLINES H:STA_NLINES X:0 Y:0 L:YES R:YES];
#endif
}

/* ---------------------------------------------------------------- */

- (void)setA:(double)a D:(double)d E:(double)e S:(int)s
{
  [dataSim setAlpha:a delta:d equinox:e];
  dataSim.disperser = s;  
}

/* ---------------------------------------------------------------- */
#pragma mark "Actions"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_status:(id)sender
{
  int err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  assert(sender);

  err = [self statusCheck:YES];

  if (err) [self showWarning:err from:PREFUN];
  else     [self addMessage:[archon.response UTF8String] reload:YES];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_frames:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  assert(sender);

  int err = [archon updateFrames];

  if (err) {
    [self showWarning:err from:PREFUN];
  } else { char buf[128];
    for (int i=0; i<ARCHON_NBUFS; i++) {
      ArchonFrame *frame = &archon.frames[i];
      sprintf(buf,"%d: %4d %5d %5d %5d %c%c%c",
              1+i,frame->frame,
              frame->width,frame->height,frame->lines,
              (frame->complete)  ? 'C' : ' ',
              (archon.wbuf==1+i) ? 'W' : ' ',
              (archon.rbuf==1+i) ? 'R' : ' ');
  
      [self addMessage:buf reload:YES];
    }
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_recon:(id)sender    // re-opens the Archon connection
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  int err = [archon reopen:2];         // Note: blocks the mainThread
  if (err) {
    [self showError:err from:"Archon re-connect"];
    return;
  }
  NSArray *array = [archon getLog];
  if (array) {
    [self addMessageArray:array];
  } else {
    if (sender) [self showWarning:E_tcpip_open from:PREFUN];
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_poweroff:(id)sender  // b0086
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  BOOL k = YalertQuestion("Do you know what you are doing?","",1,"no","Yes");
  if (k) return;                        // abort power-down
  
  [archon command:"HOLDTIMING"];
  [archon command:"POWEROFF"];
#if 1 // 0 -- allows to monitor the biases (needs mainThread)
  YalertPanel("Please turn off the Archon controller now",
              "Stopped clocking and turned off the bias power",
              1,"OK",NULL,NULL) ;
  
  YnotifyNotice("Did you turn off the Archon ?","... did you?",300,1);

  [NSApp terminate:self];
#endif
}

/* ---------------------------------------------------------------- */

- (IBAction)action_biases:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  [panel_biases makeKeyAndOrderFront:self];

}

/* ---------------------------------------------------------------- */

- (IBAction)action_terminal:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  int err = [archon open];
  if (!err) { char cmd[256],res[4096]="";
    err = [archon request:YeditGet(sender,cmd,sizeof(cmd)) response:res];
    if (err) {
      NSArray *array = [archon getLog];
      if (array) [self addMessageArray:array];
    } else {
      [self addMessage:res reload:YES];
    }
    [archon close];
  }
  
  if (sender && err) [self showWarning:err from:PREFUN];
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_acf:(id)sender      // b0069
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  if ([[sender title] compare:@"Upload"] == NSOrderedSame) { // Upload button
    [self uploadConfiguration:[edit_acf stringValue]];
  } else { NSURL *url;                 // Browse button
    if (self.acf_file.UTF8String[0] == '/') {  // absolute path
      url = [[NSURL fileURLWithPath:self.acf_file isDirectory:NO] URLByDeletingLastPathComponent];
    } else {
      url = [NSURL fileURLWithPath:NSHomeDirectory() isDirectory:YES];
    }
    NSOpenPanel *op = [[NSOpenPanel openPanel] retain];
    [op setAllowedFileTypes:[NSArray arrayWithObject:@"acf"]];
    [op setAllowsMultipleSelection:NO];
    [op setDirectoryURL:url];
    [op beginSheetModalForWindow:panel completionHandler:^(NSInteger returnCode)
      {
        if (returnCode == NSModalResponseOK) {
          [edit_acf setStringValue:[[op URL] path]];
        }
        [op release];
      }
    ];
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Exposure"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)startExposure:(int)nframes
{
  return [self startExposure:nframes dummies:0];
}

/* --- */

- (int)startExposure:(int)rframes dummies:(int)dframes
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%d): exptime=%.1f\n",PREFUN,rframes,dframes,_exptime);
#endif
  
#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif

  int err = [archon open]; if (err) return err;

  _aborted = FALSE;

  if (!err) {
    if (_online) [self lockNextFrame:0];  // Simulator: synchronize last_frame
    err = [archon setParameterFast:dummiesPar to:dframes];
    if (!err) {
      if (rframes > 0) {                  // other modes
        err = [archon setParameterFast:exposuresPar to:rframes];
      } else {                            // Reset
        err = [archon setParameterFast:resetsPar to:-rframes];
      }
    }
  }
  _start_int = walltime();

  [archon close];

#if (TIME_TEST > 0)
  NSString *str = [NSString stringWithFormat:@"%s: runtime=%.3f",PREFUN,YsecSince(date)];
  [main_logger append:[str UTF8String]];
#endif
  
  if (!_online) {                      // Simulator TODO disperser
    dataSim.shutter = self.shutter;
    dataSim.objects = self.objects;
    [dataSim startExposure];
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)abortExposure
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  int err = [archon open]; if (err) return err;

  err = [self setParameter:exposuresPar to:0];
  if (!err) err = [self setParameter:resetsPar to:0];
  self.aborted = TRUE;
  
  [archon close];
  
  return err;
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)printBufferState:(double)dt
{
  int err;
  char buf[128],buffer[1024];
  NSDate *date = [NSDate date];
  do {
    if (dt) Ysleep(150); // todo 350 ?
    err = [archon updateFrames]; if (err) continue;
    strcpy(buffer,"");
    for (int i=0; i<ARCHON_NBUFS; i++) {
      ArchonFrame *frame = &archon.frames[i];
      sprintf(buf,"%d: %4d %5d %5d %5d %c%c%c",
              1+i,frame->frame,
              frame->width,frame->height,frame->lines,
              (frame->complete)  ? 'C' : ' ',
              (archon.wbuf==1+i) ? 'W' : ' ',
              (archon.rbuf==1+i) ? 'R' : ' ');
      strcat(buffer,buf); strcat(buffer,"  ");
    }
    printf("%s\n",buffer); //xxx
  } while (YsecSince(date) < dt);
}

/* --- */

- (int)lockNextFrame:(double)timeout
{
  int b=-1,pframe=last_frame;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.1f): lf=%d\n",PREFUN,timeout,last_frame);
#endif
  assert(_online);

  NSDate *date = [NSDate date];
  do {
#ifndef NDEBUG
    [self printBufferState:0];
#else
    int err = [archon updateFrames];
    if (err) break;
#endif
    for (int i=0; i<ARCHON_NBUFS; i++) {
      ArchonFrame *frame = &archon.frames[i];
      if ((frame->frame > last_frame) && frame->complete) {
        last_frame = frame->frame;
        b = fetch_buffer = i+1;
      }
    }
    if (b > 0) { char cmd[32];
      if (timeout > 0) {
        if (last_frame-pframe > 1) { // skipped a frame -- we are hosed
          b = -1; // todo write to logfile TEST@SBS
        } else {
          sprintf(cmd,"LOCK%d",fetch_buffer); [archon command:cmd];
        }
      }
      break;
    }
    if (timeout > 0) Ysleep(350);
  } while (YsecSince(date) < timeout);

  return b;
}

/* ---------------------------------------------------------------- */

- (NSData*)fetchFrame:(float)et
{
  NSData *data=nil;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.3f)\n",PREFUN,et);
#endif

  if (!_online) {                      // Simulator
    fetch_buffer++; if (fetch_buffer > ARCHON_NBUFS) fetch_buffer = 1;
    [dataSim updateDetector:et];       // add photons and readout
    data = [dataSim fetchFrame];
    return data;
  }

  int err = [archon open]; if (err) return nil;

  int b = [self lockNextFrame:2*H2RG_FRAME_TIME]; if (b < 0) return nil; 
  ArchonFrame *frame = &archon.frames[fetch_buffer-1]; // todo use 'b' ?
  printf("b=%d (%d)\n",b,frame->frame); //xxx
  assert(frame->complete);

  u_int nbytes = _dataLines * _dataPixels * STA_NTAPS * sizeof(u_short);
  assert(frame->width == _dataPixels*STA_NTAPS);
  _fetching = TRUE; Ysleep(20);        // delay
  u_short *dbuf = [archon fetchBytes:nbytes fromBuffer:fetch_buffer offset:0];
  _fetching = FALSE;
  if (dbuf) {
    data = [[NSData alloc] initWithBytesNoCopy:dbuf length:nbytes];
  }
  [archon command:"LOCK0"];
  [archon close];
  
  if (!data) { char buf[128];              // no data error
    sprintf(buf,"%s: err=noData",PREFUN); [main_logger append:buf];
  }
  if (data) printf("transfer done (%d,%d)\n",fetch_buffer,last_frame); //xxx

  return data;
}

/* ---------------------------------------------------------------- */
#pragma mark "Housekeeping"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)statusCheck:(BOOL)reset         // resetInterval
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d): interval=%ld\n",PREFUN,reset,_interval);
#endif

  if (self.fetching) return 0; // todo just wait a bit
  while (archon.systemChange) { Ysleep(50); } 
  _last_check = cor_time(0);
    
  int err = [archon getStatus];
  
  if (!err) {
    _array_power = (archon.power == 4) ? YES : NO;
    if (archon.power != 4) {
      if (archon.power != power_warning_sent) {
        sprintf(buf,"Array power changed (%d)",archon.power);
        [main_logger message:buf level:1 time:0];
        power_warning_sent = archon.power;
      }
    } else power_warning_sent = -1; 
  }

  if (!err) {
    sprintf(buf,"SupplyV= %.3f %.3f %.3f %.3f %.3f %.3f %.3f",
            archon.p2v5_v,archon.p5v_v,archon.p6v_v,archon.n6v_v,archon.p17v_v,
            archon.n17v_v,archon.p35v_v);
    [main_logger append:buf];
#ifdef LOG_PSU
    sprintf(buf,"SupplyI= %.3f %.3f %.3f %.3f %.3f %.3f %.3f",
            archon.p2v5_i,archon.p5v_i,archon.p6v_i,archon.n6v_i,archon.p17v_i,
            archon.n17v_i,archon.p35v_i);
    [main_logger append:buf];
#endif
    sprintf(buf,"ModTemp= %.1f %.1f %.1f %.1f %.1f %.1f",
            archon.temp_mod1,archon.temp_mod2,archon.temp_mod3,
            archon.temp_mod6,archon.temp_mod7,archon.temp_mod11);
    [main_logger append:buf];
#if 0 // todo log biases ?Jason 
    for (int i=0; i<LVLC_N; i++) {
      printf("%6.3f %6.3f %6.3f\n",lvlc_v[i],archon.lvlc_v[i],archon.lvlc_i[i]);
    }
    for (int i=0; i<LVHC_N; i++) {
      printf("%6.3f %6.3f %6.3f\n",lvhc_v[i],archon.lvhc_v[i],archon.lvhc_i[i]);
    }
    printf("---\n");
#endif
  }
  if (!err) {            // error dialog displayed by caller
    if (archon.temp_backplane > back_warn_temp) {
      if (cor_time(0) >= next_back_warning) {
        sprintf(buf,"Archon backplane temperature high (%.1f)",archon.temp_backplane);
        add_back_warning += 60+add_back_warning/2;
        next_back_warning = cor_time(0)+add_back_warning;
        [main_logger message:buf level:1 time:(int)add_back_warning];
      }
    } else {
      add_back_warning = next_back_warning = 0;
    }
    sprintf(buf,"BackplaneTemp: %.1f",archon.temp_backplane);
    [main_logger append:buf];
  }

  if (!err) {
    NSArray *array = [archon getLog];
    for (NSString *line in array) {
      NSString *string = [NSString stringWithFormat:@"ArchonLog: %@",line];
      [main_logger message:[string UTF8String] file:YES];
    }
  }
  
  dispatch_async(MAIN_QUEUE,^(void) { char buf[128];
    sprintf(buf,"%6.2f",my_round(archon.p2v5_v,2)); YeditSet(edit_p2v5,buf);
    sprintf(buf,"%6.2f",my_round(archon.p5v_v,2));  YeditSet(edit_p5v,buf);
    sprintf(buf,"%6.2f",my_round(archon.p6v_v,2));  YeditSet(edit_p6v,buf);
    sprintf(buf,"%6.2f",my_round(archon.n6v_v,2));  YeditSet(edit_n6v,buf);
    sprintf(buf,"%6.2f",my_round(archon.p17v_v,2)); YeditSet(edit_p17v,buf);
    sprintf(buf,"%6.2f",my_round(archon.n17v_v,2)); YeditSet(edit_n17v,buf);
    sprintf(buf,"%6.2f",my_round(archon.p35v_v,2)); YeditSet(edit_p35v,buf);
    sprintf(buf,"%.1f",my_round(archon.temp_backplane,1)); YeditSet(edit_tmbp,buf);
    sprintf(buf,"%.1f",my_round(archon.temp_mod1,1));  YeditSet(edit_tm01,buf);
    sprintf(buf,"%.1f",my_round(archon.temp_mod2,1));  YeditSet(edit_tm02,buf);
    sprintf(buf,"%.1f",my_round(archon.temp_mod3,1));  YeditSet(edit_tm03,buf);
    sprintf(buf,"%.1f",my_round(archon.temp_mod6,1));  YeditSet(edit_tm06,buf);
    sprintf(buf,"%.1f",my_round(archon.temp_mod7,1));  YeditSet(edit_tm07,buf);
    sprintf(buf,"%.1f",my_round(archon.temp_mod11,1)); YeditSet(edit_tm11,buf);
  });

#ifdef NDEBUG
  _interval = (reset) ? 5 : lmin(300,_interval+1+_interval/15);
#else
  _interval = (reset) ? 2 : lmin(30, _interval+1+_interval/15);
#endif
  

  return err;
}

/* ---------------------------------------------------------------- */

- (void)addMessageArray:(NSArray*)array
{
  for (NSString* string in array) {
    [self addMessage:[string UTF8String] reload:NO];
  }
  [self addMessage:NULL reload:YES];
}

/* ---------------------------------------------------------------- */

- (void)addMessage:(const char*)text reload:(BOOL)update
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,text,update);
#endif

  if (text) {
    NSString *string = @(text);
    @synchronized(respArray) {
      if (resp_head < [respArray count]) {
        [respArray replaceObjectAtIndex:resp_head withObject:string];
      } else {
        [respArray addObject:string];
      }
      assert([respArray count] <= resp_n);
      resp_head = (resp_head+1) % resp_n;
    }
  }
  if (update) {
    [responseTable performSelectorOnMainThread:@selector(reloadData)
                   withObject:nil waitUntilDone:NO];
  }
}


/* ---------------------------------------------------------------- */
#pragma mark "NSTableView dataSource"
/* ---------------------------------------------------------------- */

- (NSInteger)numberOfRowsInTableView:(NSTableView*)table
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p): %lu\n",PREFUN,table,[respArray count]);
#endif
  return [respArray count];
}

/* ---------------------------------------------------------------- */

- (id)tableView:(NSTableView*)table
      objectValueForTableColumn:(NSTableColumn*)column
      row:(NSInteger)row
{
  NSString *string=nil;
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p,%p,%ld)\n",PREFUN,table,column,row);
#endif
  assert(row < [respArray count]);
  if (row >= [respArray count]) return nil;
  
  @synchronized(respArray) {
    long i = resp_head-row-1; if (i < 0) i += resp_n;
    string = [respArray objectAtIndex:i];
  }

  return string;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
