/* ---------------------------------------------------------------- */
//
// ArchonIO.m
// PFS
//
// Created by Christoph Birk on 2015-07-29
// Copyright (c) 2015 Christoph Birk. All rights reserved.
//
// ?Greg: controller does not respond to invalid/malformed commands
//        BITPIX is wrong (16==signed short)
// ?Greg: archongui does not remember last IP ... very annoying
//        too many tabs: timing script and parameters should be on same tab
// ?Greg: request increment-and-compare-to-X (allows changing exptime) TODO Jason says new version has t
//
/* ---------------------------------------------------------------- */

#define DEBUG           1
#define TIME_TEST       1
#define UNIT_TEST       0           // 1:ArchonLog

#define RESPONSE_SIZE   8192

/* ---------------------------------------------------------------- */

#import "ArchonIO.h"

#import "Yutils.h"
#import "Logger.h"
#import "MTwister.h"

/* ---------------------------------------------------------------- */

static Logger *archon_logger=nil;

extern Logger *main_logger;

/* ---------------------------------------------------------------- */

static NSString* get_value(const char* status,const char* key)
{
  NSString *string=@"";
  char *ptr,*buffer,*token=NULL;
  
  ptr = buffer = (char*)malloc(1+strlen(status));
  strcpy(buffer,status);
  size_t len = strlen(key);
  while ((token=strsep(&buffer," ")) != NULL) {
    if (!strncmp(token,key,len)) break;
  }
  if (token) {
    char *p = strchr(token,'=');
    if (p) string = @((char*)(p+1));
  }
  free((void*)ptr);
  
  return string; // 'NSString:componentsSeparatedByString' is 50 times slower
}

/* --- */

static float get_float(const char* string,const char* key)
{
  return [get_value(string,key) floatValue];
}

/* --- */

static u_long get_hex(const char* string,const char* key)
{
  char   buf[128];
  u_long t=0;

  Ystr2chr(get_value(string,key),buf,sizeof(buf));
  sscanf(buf,"%lx",&t);
  
  return t;
}

/* --- */

static long get_long(const char* string,const char* key)
{
  return [get_value(string,key) integerValue];
}

/* ---------------------------------------------------------------- */

@interface ArchonIO ()
@property (readwrite,nonatomic,copy) NSString *back_version;
@property (readwrite,atomic,copy)    NSString *response;
@property (readwrite,atomic,retain)  NSArray* logArray;
@end

/* ---------------------------------------------------------------- */

@implementation ArchonIO

@synthesize uptime,frames,wbuf,rbuf;

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

+ (void)initialize
{
  extern BOOL use_shared;
  
  if (self != [ArchonIO class]) return;  // prevent sub-classes
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  archon_logger = [[Logger alloc] initWithBase:"archon" useShared:use_shared];
#ifdef NDEBUG
  archon_logger.mode = LOGGER_OPEN;
#else
  archon_logger.mode = LOGGER_FLUSH;
#endif
}

/* --- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
    case E_archon_error:       string = @"ArchonIO: error"; break;
    case E_archon_response:    string = @"ArchonIO: invalid response (short)"; break;
    case E_archon_response_id: string = @"ArchonIO: invalid response (id)"; break;
    case E_archon_response_lf: string = @"ArchonIO: invalid response (LF)"; break;
    case E_archon_command:     string = @"ArchonIO: invalid command"; break;
    case E_archon_status:      string = @"ArchonIO: invalid status"; break;
    case E_archon_parameter:   string = @"ArchonIO: invalid parameter"; break;
    default:                   string = [super errorString:err]; break;
  }
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)initWithHost:(const char*)hstr port:(int)pint online:(BOOL)on
{
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d,%d)\n",PREFUN,hstr,pint,on);
#endif

  self = [super initWithHost:hstr port:pint online:on];
  if (!self) return nil;

  sprintf(buf,"%s:%d (%d)",hstr,pint,on);
  [archon_logger append:buf];
  
  sock_timeout = 5;
  frames = (ArchonFrame*)malloc(ARCHON_NBUFS*sizeof(ArchonFrame));
  
  _p2v5_v =   2.5f;
  _p5v_v  =   5.0f;
  _p6v_v  =   6.0f;
  _n6v_v  =  -6.0f;
  _p17v_v =  17.0f;
  _n17v_v = -17.0f;
  _p35v_v =  35.0f;
  
  _lvlc_v = (float*)calloc(LVLC_N,sizeof(float)); // low voltage low current biases
  _lvlc_i = (float*)calloc(LVLC_N,sizeof(float));
  _lvhc_v = (float*)calloc(LVHC_N,sizeof(float)); // low voltage hight current biases
  _lvhc_i = (float*)calloc(LVHC_N,sizeof(float));

  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
   [_back_version release];
   [_logArray release];
   [_response release];
   [super dealloc];
}

/* ---------------------------------------------------------------- */

#ifndef NDEBUG 

- (int)open
{
  int err;
  @synchronized(self) {
    if (count == 0) { char buf[128];
      sprintf(buf,"%s (%s:%d)",PREFUN,host,port); [archon_logger append:buf];
    }
    err = [super open];
  }
  return err;
}

/* --- */

- (void)close
{
  @synchronized(self) {
    [super close];
    if (count == 0) { char buf[128];
      sprintf(buf,"%s (%s:%d)",PREFUN,host,port); [archon_logger append:buf];
    }
  }
}

#endif

/* ---------------------------------------------------------------- */

- (int)request:(const char*)cmd response:(char*)res
{
  char command[256],buf[RESPONSE_SIZE]="";
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%p)\n",PREFUN,cmd,res);
#endif

  int err = [self open]; if (err) return err;

  if (![self mylock:lock_timeout]) {   // open/lock is more MT-efficient
    [self close];
    return E_tcpip_lock;
  }

  [archon_logger append:cmd];

  assert(strlen(cmd) < sizeof(command)-5);
  sprintf(command,">%02X%s\n",messageID,cmd);  // IDEA: NSString
  messageID = (messageID+1) % 0x100;
  
  if (!online) {                       // Simulator
    Ysleep(5);
  } else {
    assert(sock > 0);
    err = TCPIP_Send(sock,command,strlen(command));
    if (err) {
      sprintf(buf,"err=%d",err); [archon_logger append:buf];
    } else
    if (res) { int i;
      for (i=0; i<RESPONSE_SIZE-1; i++) {
        err = TCPIP_Wait4Char(sock,&buf[i],1000*sock_timeout);
        if (err) break;
        if (buf[i] == terminator) break;
      }
      if (!err) {                      // check for valid response
        if      (i < 3)          err = E_archon_response;
        else if (buf[0] == '?')  err = E_archon_command;
        else if (buf[0] != '<')  err = E_archon_response_id; // todo more logging
        else if (buf[i] != '\n') err = E_archon_response_lf;
      }
      if (err) {
        buf[i] = '\0';
        sprintf(res,"err=%d (%s)",err,buf); 
      } else {                         // cleanup return string
        buf[i] = '\0';
        strcpy(res,buf+3);
      }
      if (*res) {
#ifndef NDEBUG
        [archon_logger append:res];
#else
        const u_int mlen=79;           // shorten log
        if (strlen(res) > mlen) {
          strcpy(buf,res); buf[mlen] = '\0'; strcat(buf," ...");
          [archon_logger append:buf];
        } else {
          [archon_logger append:res];
        }
#endif
      }
    }
  }
  
  [lock unlock];
  [self close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)command:(const char*)cmd
{
  char buf[RESPONSE_SIZE];

  return [self request:cmd response:buf];
}
  
/* ---------------------------------------------------------------- */

- (int)reopen:(int)delay
{
  int err=0;
  
  if (![self mylock:30]) return E_tcpip_lock;
  
  int cnt = count;
  while (count > 0) [self close];
  [NSThread sleepForTimeInterval:delay];
  for (int i=0; i<cnt; i++) {
    err = [self open]; if (err) break;
  }
  [lock unlock];

  return err;
}

/* ---------------------------------------------------------------- */
#pragma mark "Commands"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)getSystem
{
  char res[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  int err = [self request:"SYSTEM" response:res];

  if (!err) { 
    if (online) self.back_version = get_value(res,"BACKPLANE_VERSION");
    else        self.back_version = @"offline";
    self.response = @(res);
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getStatus     // allowed during fetchBytes -- uses one FETCH
{
  char res[RESPONSE_SIZE]="",key[32];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif

  int err = [self open]; if (err) return err;
  
  err = [self request:"STATUS" response:res];
  if (!err) {
    self.response = @(res);
    if (online) {                      // the real thing
      int valid = (int)get_long(res,"VALID");
      if (!valid) {
        [self close];
        return E_archon_status; // TODO auto-reconnnect ? -- never happend
      }
      _power = (int)get_long(res,"POWER");
      _nlogs = (int)get_long(res,"LOG");
      _p2v5_v = get_float(res,"P2V5_V");
      _p5v_v  = get_float(res,"P5V_V");
      _p6v_v  = get_float(res,"P6V_V");
      _n6v_v  = get_float(res,"N6V_V");
      _p17v_v = get_float(res,"P17V_V");
      _n17v_v = get_float(res,"N17V_V");
      _p35v_v = get_float(res,"P35V_V");
#ifdef LOG_PSU
      _p2v5_i = get_float(res,"P2V5_I");
      _p5v_i  = get_float(res,"P5V_I");
      _p6v_i  = get_float(res,"P6V_I");
      _n6v_i  = get_float(res,"N6V_I");
      _p17v_i = get_float(res,"P17V_I");
      _n17v_i = get_float(res,"N17V_I");
      _p35v_i = get_float(res,"P35V_I");
#endif
      _temp_backplane = get_float(res,"BACKPLANE_TEMP");
      _temp_mod1  = get_float(res,"MOD1/TEMP");
      _temp_mod2  = get_float(res,"MOD2/TEMP");
      _temp_mod3  = get_float(res,"MOD3/TEMP");
      _temp_mod6  = get_float(res,"MOD6/TEMP");
      _temp_mod7  = get_float(res,"MOD7/TEMP");
      _temp_mod11 = get_float(res,"MOD11/TEMP");
      for (int i=0; i<LVLC_N; i++) {   // read actual biases b0082 
        sprintf(key,"MOD11/LVLC_V%d",1+i); _lvlc_v[i] = get_float(res,key);
        sprintf(key,"MOD11/LVLC_I%d",1+i); _lvlc_i[i] = get_float(res,key);
      }
      for (int i=0; i<LVHC_N; i++) {
        sprintf(key,"MOD11/LVHC_V%d",1+i); _lvhc_v[i] = get_float(res,key);
        sprintf(key,"MOD11/LVHC_I%d",1+i); _lvhc_i[i] = get_float(res,key);
      }
    } else {                           // Simulator
      _power = 4;
      _temp_backplane = [MTwister gaussMean:29.0 sigma:0.5];
      for (int i=0; i<LVLC_N; i++) {
        _lvlc_v[i] = [MTwister gaussMean:i*0.4 sigma:0.1];
        _lvlc_i[i] = [MTwister gaussMean:0.01 sigma:0.001];
      }
      for (int i=0; i<LVHC_N; i++) {
        _lvhc_v[i] = [MTwister gaussMean:i*1.0 sigma:0.1];
        _lvhc_i[i] = [MTwister gaussMean:0.1 sigma:0.01];
      }
    }
  }
    
  [self close];
  
#if (TIME_TEST > 1)
  fprintf(stderr,"%s: run_time=%.3f\n",PREFUN,YsecSince(date));
#endif
  
  return err;
}

/* ---------------------------------------------------------------- */

- (NSArray*)getLog
{
  char res[RESPONSE_SIZE]="";
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  int err = [self open]; if (err) return nil;
  
  NSMutableArray *array = [NSMutableArray arrayWithCapacity:2];
  while (!err) {
    strcpy(res,ARCHON_NULL);            // Simulator
#if (UNIT_TEST == 1)
    static u_int counter=0;
    if (++counter % 2) sprintf(res,"counter=%u",counter);
#endif
    err = [self request:"FETCHLOG" response:res];
    if (!err) {
      if (!strcmp(res,ARCHON_NULL)) break;
      [array addObject:@(res)];
    }
  }
  [self close];
 
  return array;
}

/* ---------------------------------------------------------------- */

- (int)updateFrames  
{
  char res[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert(sizeof(long) == 8);           // 64-bit
  
#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif
  int err = [self request:"FRAME" response:res];
  if (!err) {
    if (online) {
      uptime = (double)get_hex(res,"TIMER") * ARCHON_TIMER_TIC;
      wbuf = (int)get_long(res,"WBUF");
      rbuf = (int)get_long(res,"RBUF");
      for (int i=0; i<ARCHON_NBUFS; i++) [self updateFrame:i fromResponse:res];
    } else {                           // Simulator
      wbuf++; if (wbuf > ARCHON_NBUFS) wbuf = 1;
    }
  }
#if (TIME_TEST > 1)
  fprintf(stderr,"%s: dt=%.2f [ms]\n",PREFUN,YmsSince(date));
#endif
  
  return err;
}

/* ---------------------------------------------------------------- */

- (void)updateFrame:(int)i fromResponse:(const char*)res
{
  char key[32];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,i,res);
#endif
  assert((i>=0) && (i<ARCHON_NBUFS));
  
#if (TIME_TEST > 2)
  NSDate *date = [NSDate date];
#endif
  sprintf(key,"BUF%dBASE",1+i);
  frames[i].base = (u_int)get_long(res,key);
  sprintf(key,"BUF%dFRAME",1+i);
  frames[i].frame = (int)get_long(res,key);
  sprintf(key,"BUF%dWIDTH",1+i);
  frames[i].width = (int)get_long(res,key);
  sprintf(key,"BUF%dHEIGHT",1+i);
  frames[i].height = (int)get_long(res,key);
  //sprintf(key,"BUF%dPIXELS",1+i);
  //value = get_long(res,key);
  sprintf(key,"BUF%dLINES",1+i);
  frames[i].lines = (int)get_long(res,key);
  sprintf(key,"BUF%dCOMPLETE",1+i);
  long value = get_long(res,key);
  frames[i].complete = (value) ? YES : NO;
  sprintf(key,"BUF%dTIMESTAMP",1+i);
  frames[i].timestamp = get_hex(res,key);
  sprintf(key,"BUF%dRETIMESTAMP",1+i);
  u_long re = get_hex(res,key);
  sprintf(key,"BUF%dFETIMESTAMP",1+i);
  u_long fe = get_hex(res,key);
  frames[i].deltatime = (double)(fe-re) * ARCHON_TIMER_TIC;
#if (TIME_TEST > 2)
  printf("%s: runtime=%.6f\n",PREFUN,YsecSince(date));
#endif
}

/* --- */

- (int)updateFrame:(int)b
{
  char res[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,b);
#endif
  assert((b>=1) && (b<=ARCHON_NBUFS));

  int err = [self request:"FRAME" response:res];

  if (!err && online) [self updateFrame:(b-1) fromResponse:res];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)getTimer
{
  char res[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert(sizeof(u_long) == 8);         // 64-bit

  int err = [self request:"TIMER" response:res];
  
  if (!err && online) {
    uptime = (double)get_hex(res,"TIMER") * ARCHON_TIMER_TIC;
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)uploadConfiguration:(NSArray*)config 
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,config);
#endif

  int err = [self open]; if (err) return err;

  NSDate *date = [NSDate date];

  @synchronized(self) { self.systemChange += 1; } 

  if (!err) err = [self command:"POLLOFF"];
  if (!err) err = [self command:"CLEARCONFIG"];
  for (int i=0; i<[config count]; i++) {
    if (err || !online) break; 
    NSString *line = [config objectAtIndex:i];
    NSString *cmd = [NSString stringWithFormat:@"WCONFIG%04X%@",i,line];
    err = [self command:[cmd UTF8String]];
  }
  if (!err) err = [self command:"POLLON"];
  if (!err) { char buf[128];
    sprintf(buf,"conf-time= %.3f (%lu)",YsecSince(date),[config count]);
    [main_logger append:buf];
  }
  if (!err) { int timeout=sock_timeout;
    sock_timeout = 30;                  // extend timeout on socket
    NSDate *date = [NSDate date];
    err = [self command:"APPLYALL"];    // turns off power and runs VCPU
    printf("APPLYALL took %.2f seconds\n",YsecSince(date)); // 0.61 [s]
    sock_timeout = timeout;
  }
  [self close];
  
  @synchronized(self) { self.systemChange -= 1; }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setParameterFast:(const char*)par to:(int)value // b0074
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,par,value);
#endif

  sprintf(buf,"FASTLOADPARAM %s %d",par,value);
  int err = [self command:buf];

  return err;
}

/* ---------------------------------------------------------------- */

- (u_short*)fetchBytes:(u_int)nbytes fromBuffer:(int)b offset:(u_int)offset
{
  const u_int block_size=1024,pre_size=4,blk_size=1024+4;
  char        buf[128],*data=NULL,*blkb=NULL;
  ssize_t     rval,rbytes=0,rb;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%u,%d,%u)\n",PREFUN,nbytes,b,offset);
#endif
#if (TIME_TEST > 0)
  u_int nchunks=0;
#endif
  assert(count > 0);                   // connection is open
  
  if (![self mylock:lock_timeout]) return NULL; // logged upstream

  u_int extra_bytes = offset % block_size;
  u_int nblocks = (nbytes+extra_bytes+block_size-1)/block_size;
  u_int buf_size = block_size*nblocks;
  assert(buf_size >= nbytes);
  data = (char*)malloc(buf_size);
  blkb = (char*)malloc(blk_size);
  
  assert((b>=1) && (b<=ARCHON_NBUFS));
  u_int address = frames[b-1].base+offset-extra_bytes;
  assert((address % block_size) == 0);
  sprintf(buf,"FETCH%08X%08X",address,nblocks);
  int err = [self request:buf response:NULL];

#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif

  for (int i=0; i<nblocks; i++) {      // copy data 
    rb = 0;
    while (!err) {                     // read block
      rval = TCPIP_Recv(sock,blkb+rb,blk_size-rb,5000); // [ms]
#if (TIME_TEST > 0)
      nchunks++;
#endif
      if (rval <= 0) { err = E_tcpip_recv; break; }
      rb += rval;
      if (rb == blk_size) {            // got a block
        if (i == 0) {
          memcpy(data+rbytes,blkb+pre_size+extra_bytes,block_size-extra_bytes);
          rbytes += block_size-extra_bytes;
        } else {
          memcpy(data+rbytes,blkb+pre_size,block_size);
          rbytes += block_size;
        }
        break;
      }
    }
    if (err) break;
  }
  [lock unlock];
  
#if (TIME_TEST > 0)
  double dt = YsecSince(date);
#endif
  if (err) {                           // logfile 
    free((void*)data); data = NULL;
    sprintf(buf,"%s: err=%d",PREFUN,err); [main_logger append:buf];
  }
  free((void*)blkb);
  
#if (TIME_TEST > 0)
  sprintf(buf,"kBytes=%6u nchunks=%7u (%6.1f) time=%6.3f (%5.1f)",
          nbytes/1024,nchunks,(float)nbytes/(float)nchunks,
          dt,((float)nbytes/(1024*1024))/dt);
  [main_logger append:buf];
#endif
  // printf("rbytes=%ld, nbytes=%u (%d)\n",rbytes,nbytes,err);
  
  return (u_short*)data;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
