/* ---------------------------------------------------------------- */
//
// CameraController.m
// Henrietta
//
// Created by Christoph Birk on 2022-09-22
// Copyright (c) 2022 Christoph Birk. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG      1
#define TIME_TEST  1

/* ---------------------------------------------------------------- */

#import  "CameraController.h"
#import  "FITS.h"
#import  "AppDelegate.h"
#import  "Notice.h"
#import  "Sound.h"
#import  "SwopeIO.h"   
#import  "Yutils.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

#define FRAME_CAMERA    "CameraWindow"

#define DEF_PREFIX      "hen"
#define MAX_FILENUM     9999   
#define MAX_EXPTIME     999.9          // TODO ?Nick

#define DBE_DISK_WARN   "dbe_disk_warning"
#define DBE_DISK_CRIT   "dbe_disk_critical"
#define DBE_RUN         "run_number"
#define DISK_INTERVAL   600

enum exposure_type_enum {              // 'IB' dependency
  ET_OBJECT,ET_DARK,ET_FLAT,ET_N       // ?Nick
};

enum readout_mode_enum {               // 'IB' dependency
  RM_RESET,RM_SINGLE,RM_CDS,RM_SUTR,RM_N
};

#define DEF_TELLAT      (-29.083)
#define DEF_TELLONG     (-70.698)
#define DEF_TELALT      (2280)

/* ---------------------------------------------------------------- */

@implementation CameraController
{
  AppDelegate *app;
  Ysemaphore *deviceMoving;
  GalilWheel *gWheel[GALIL_WHEELS];
  GalilSlide *gSlide;
  // Exposure
  double  cur_runtime;
  NSData *write_data;
  dispatch_queue_t sutr_queue;
  int nframes,rframes,dframes;
  // Telescope
  SwopeIO *tcs;
}

@synthesize loop_doing;
@synthesize panel_sta;                   // used by AppDelegate

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
  case E_camera_error:   string = @"Henrietta/H2RG: generic error"; break;
  case E_camera_timeout: string = @"Henrietta/H2RG: timeout during readout"; break;
  case E_camera_lines:   string = @"Henrietta/H2RG: insufficient data lines"; break;
  case E_camera_abort:   string = @"Henrietta/H2RG: readout aborted"; break;
  case E_camera_sync:    string = @"Henrietta/H2RG: no data from Archon"; break;
  case E_camera_fits:    string = @"Henrietta/H2RG: open FITS failed"; break;
  default:
    if (!string) string = [ArchonController errorString:err];
    if (!string) string = [SwopeIO errorString:err];
    break;
  }
  if (!string) string = [AppDelegate makeExcuse];
  assert(string);
  
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "NSWindowController"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)initWithWindowNibName:(NSString*)name
{
  char buf[256],host[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif

  if (!(self = [super initWithWindowNibName:name])) return nil;

  assert([NSThread isMainThread]);
  app = [NSApp delegate];          
  assert(app);

  nloops = 1;

  YprefGetChar(DBE_OBSERVER,DEF_OBSERVER,buf,sizeof(buf),NO);
  strcpy(observer,trim_string(buf,1+FITS_MAX_CHAR));
  
  datapaths = [[NSMutableArray alloc] initWithCapacity:N_FOLDERS];

  int online = YprefGetInt(DBE_TCS_ONLINE,DEF_STATE,use_shared);
  int port = DEF_TCS_PORT;
  YprefGetChar(DBE_TCS_HOST,DEF_TCS_HOST,host,sizeof(host),use_shared);
  sprintf(buf,"tcs_host= %s:%d (%d)",host,port,online); [main_logger append:buf];
  tcs = [[SwopeIO alloc] initWithHost:host port:port online:online];
#if 1 //def SIM_ONLY
  alpha   = d_alpha( "06:55:42"); // d_alpha( "06:58:40");
  delta   = d_delta("-28:36:21"); // d_delta("-28:58:19");
  equinox = 2000.0;
#else
  alpha   = get_sidereal(cor_time(0),DEF_TELLONG)/3600.0;
  delta   = DEF_TELLAT;
  equinox = get_epoch(0);
#endif
  pixscale = LO_MEM*(9.0/15.0)*0.095; // ?Nick 

  deviceMoving = YsemaphoreCreate("deviceMoving");
  
  // Note: the GalilWheels (gWheel[]) are taken from the Galil hardhat
  // Note: the GalilConnection for the slide is taken from its GalilAxis
  gSlide = [[GalilSlide alloc] initWithConnection:nil];

  sutr_queue = dispatch_queue_create(NULL,DISPATCH_QUEUE_SERIAL);
  
  next_disk = cor_time(0) + DISK_INTERVAL;
  
  [self showWindow:self];
  
  return self;
}

/* ---------------------------------------------------------------- */

- (void)windowDidLoad
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  [[self window] setFrameUsingName:@FRAME_CAMERA];
#ifdef NDEBUG
  [[self window] setStyleMask:([[self window] styleMask] & ~NSWindowStyleMaskClosable)];
#endif
  
  [self enableControls:FALSE];

  level_disks = [[NSArray alloc] initWithObjects:level_disk1,level_disk2,nil];
#ifndef NDEBUG
  { NSSet *set = [NSSet setWithArray:level_disks]; assert([set count] == N_FOLDERS); }
#endif

  run_number = YprefGetInt(DBE_RUN,1,NO);
  sprintf(buf,"%d",run_number); YeditSet(edit_file,buf);
  
  [self reloadPreferences];
  [self action_readmode:nil];          // preset 'readmode'

  dispatch_async(GLOBAL_QUEUE,^(void) {
    do {
      Ysleep(350);                     // NOTE: no timeout
    } while (!staController.configured);
    if (!staController.online)  {      // setup Simulator
      [staController setupSimulator:pixscale];
    }
    [self enableExposure:TRUE];
  });

  dispatch_async(GLOBAL_QUEUE,^(void) {  // get telescope position
    int err = [self sampleTCS:NO];
    if (err) [self showError:"cannot connect to TCSIS" error:err];
  });

  // Note: the order has to be by axis B,C,D,E
  drop_wheels = [@[drop_grism,drop_diffuser,drop_filter,drop_slit] retain];
#ifndef NDEBUG
  { NSSet *set = [NSSet setWithArray:drop_wheels]; assert([set count]==GALIL_WHEELS); }
#endif
  dispatch_async(GLOBAL_QUEUE,^(void) {  // setup galil motor controls
    // wait for the Galil hardhat to finish setting up the motors
    while (!app.galilHH.configured) { Ysleep(350); }
    int err = [gSlide setup:"slide" fromPath:r_path withAxis:app.galilHH.gAxes[0]];
    if (!err) YcontrolEnable(drop_slide,YES);
    [self updateSlide:NO];
    if (!err) {
      for (int i=0; i<GALIL_WHEELS; i++) {
        gWheel[i] = app.galilHH.gWheels[i]; // printf("gW[%d]=%p\n",i,gWheel[i]);
        dispatch_sync(MAIN_QUEUE,^(void) { // add menu entries
          NSPopUpButton *drop = drop_wheels[i];
          for (NSString *s in gWheel[i].positionNames) [drop addItemWithTitle:s];
          [drop setEnabled:YES]; // TODO disable if not in legal position?
          [self updateWheel:i fromGalil:NO];
        });
      } // endfor(i)
    } // endif(err)
  });  // run_setup Galil

#ifdef NDEBUG
  [self showQltool];
#endif

  assert([NSThread isMainThread]);
  [NSTimer scheduledTimerWithTimeInterval:5.0 target:self
           selector:@selector(fireTimer:) userInfo:nil repeats:YES];
}

/* ---------------------------------------------------------------- */

- (void)windowWillClose:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,note);
#endif

  YwindowFrameSave([self window],FRAME_CAMERA);
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)fireTimer:(NSTimer*)theTimer
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p)\n",PREFUN,theTimer);
#endif

  time_t now = cor_time(0);

  if (now >= next_disk) {              // check disk status
    if (!loop_doing) {
      next_disk = now + DISK_INTERVAL;
      [self updateDisk:NO];
    }
  }

  if (!staController.configured) return;
  
  if (now >= staController.last_check+staController.interval) {
    // todo last_check and interval should be IVAR at CameraController
    if (!loop_doing) {
      dispatch_async(GLOBAL_QUEUE,^(void) { [self sampleArchon:NO]; });
    }
  }
}

/* ---------------------------------------------------------------- */

- (void)reloadPreferences
{
  int  w,c;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert([level_disks count] == N_FOLDERS);  // after 'windowDidLoad'
  assert([NSThread isMainThread]);

  YprefGetChar(DBE_SND_EXPDONE,DEF_SND_EXPDONE,snd_expdone,sizeof(snd_expdone),NO);
  
  w = YprefGetInt(DBE_DISK_WARN,80,use_shared);   // yellow
  c = YprefGetInt(DBE_DISK_CRIT,95,use_shared);   // red
  for (NSLevelIndicator *level in level_disks) {
    [level setWarningValue:w];
    [level setCriticalValue:c];
    level.needsDisplay = YES;
  }
  [self reloadDatapath];

  _refPixelMode = YprefGetInt(DBE_REFPIXMODE,1,use_shared); // TODO shared ?Nick
  _showRaw = YprefGetInt(DBE_SHOWRAW,1,NO);
  _saveRaw = YprefGetInt(DBE_SAVERAW,0,NO);
  _showFrames = YprefGetInt(DBE_SHOWFRAMES,1,NO);
  _saveFrames = YprefGetInt(DBE_SAVEFRAMES,0,NO);
  _showCurrent = YprefGetInt(DBE_SHOWCURRENT,1,NO);

  [qltool reloadPreferences:YES];
  
  dispatch_async(GLOBAL_QUEUE,^(void) { [self sampleArchon:NO]; });
}

/* ---------------------------------------------------------------- */

- (void)reloadDatapath
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  NSArray *a = [app dataPaths];
  assert([a count] == N_FOLDERS);
  assert(datapaths);
  [datapaths removeAllObjects];
  [datapaths addObjectsFromArray:a];

  [self updateDisk:YES];
}

/* ---------------------------------------------------------------- */

- (void)showQltool
{
  if (!qltool) {
    qltool = [[QltoolController alloc] initWithWindowNibName:@"Qltool"];
  } else {
    [[qltool.mag_view window] orderFront:self];
    [qltool.window makeKeyAndOrderFront:self];
  }
}

/* ---------------------------------------------------------------- */

- (void)showWarning:(const char*)txt error:(int)err
{
  NSString *string;

  if (err) {
    string = [NSString stringWithFormat:@"%s, err=%d\r%@",txt,err,
                       [[self class] errorString:err]];
  } else {
    string = [NSString stringWithFormat:@"%s",txt];
  }
#ifdef NDEBUG
  [main_logger message:[string UTF8String] level:1 time:60];
#else
  [main_logger message:[string UTF8String] level:1 time:10];
#endif
}

/* --- */

- (void)showError:(const char*)txt error:(int)err
{
  NSString *string;

  if (err) {
    string = [NSString stringWithFormat:@"%s, err=%d\r%@",txt,err,
                       [[self class] errorString:err]];
  } else {
    string = [NSString stringWithFormat:@"%s",txt];
  }
#ifdef NDEBUG
  [main_logger message:[string UTF8String] level:2 time:600 file:YES];
#else
  [main_logger message:[string UTF8String] level:2 time:30 file:YES];
#endif
}

/* ---------------------------------------------------------------- */

- (void)loadFITS:(NSURL*)url asSky:(BOOL)sky
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,[url path]);
#endif

  if (!qltool) {
    dispatch_async(MAIN_QUEUE,^(void) {
      [self showQltool];
    });
  }
  dispatch_async(GLOBAL_QUEUE,^(void) { int n1,n2;
    int *data = fits_load32([[url path] UTF8String],&n1,&n2,0);
    if (data) {
      while (!qltool) { Ysleep(350); }
      if (!sky)  [qltool resetTitle:[url lastPathComponent]];
      [qltool copyData:data asSky:sky];  // bug-fix b0090
      free((void*)data);
    } // endfor(param)
    if (!data) {
      [Notice openNotice:"loading FITS file failed" title:P_TITLE time:30 level:1];
    }
  });
}



/* ---------------------------------------------------------------- */
#pragma mark "Enable GUI"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)enableControls:(BOOL)enable
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,enable);
#endif

  [self enableExposure:enable];
}

/* ---------------------------------------------------------------- */

- (void)enableExposure:(BOOL)enable
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,enable);
#endif

  dispatch_async(MAIN_QUEUE,^(void) { BOOL e;
    e = (staController.configured && enable);
    [edit_nloops setEnabled:e];
    e = (staController.configured && enable && !loop_doing);
    [popup_exptype setEnabled:e];
    [popup_readmode setEnabled:e];
    e = (staController.configured && enable);
    [but_start setEnabled:(e && !loop_doing && !YsemaphoreGet(deviceMoving))];
    [but_stop  setEnabled:(e && loop_doing)];
    e = ((readmode==RM_RESET) || (readmode==RM_SINGLE)) ? NO : (enable && !loop_doing);
    [edit_exptime setEnabled:e];
  });
}

/* ---------------------------------------------------------------- */

- (void)startMotion:(id)sender run:(BOOL)start
{
  if (start) {
    YcontrolEnable(sender,NO);
    YsemaphoreInc(deviceMoving);
    [self enableExposure:YES];   // TODO allow exposures while moving ?Nick,?Jason
  } else {
    YcontrolEnable(sender,YES);
    YsemaphoreDec(deviceMoving);
    [self enableExposure:YES];
  }
}

/* --- */

- (int)motorMoving                     // NEW b0096
{
  return deviceMoving->count;
}

/* ---------------------------------------------------------------- */
#pragma mark "Update GUI"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)updateExposure:(BOOL)updGUI
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,updGUI);
#endif

  cur_runtime = walltime() - staController.start_int;
  if (!updGUI) return;
  
  dispatch_async(MAIN_QUEUE,^(void) { char buf[32];
    sprintf(buf,"%d",(int)my_round(cur_runtime,0)); YeditSet(edit_runtime,buf);
    [prog_exposure setDoubleValue:(cur_runtime)];
  });
}

/* ---------------------------------------------------------------- */

- (void)updateDewar
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  // TODO from temps LakeShore -- Archon temps not connected ?Jason,?Greg

}

/* ---------------------------------------------------------------- */

- (void)updateDisk:(BOOL)resetWarnings   
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert([NSThread isMainThread]);

  if (resetWarnings) {
    for (int i=0; i<N_FOLDERS; i++) {
      warningPosted[i] = criticalPosted[i] = FALSE;
    }
  }
  
  NSFileManager *fm = [[NSFileManager alloc] init];

  assert([level_disks count] == N_FOLDERS);
  for (int i=0; i<N_FOLDERS; i++) {
    NSLevelIndicator *level = [level_disks objectAtIndex:i];
    NSString *path = [datapaths objectAtIndex:i];
    if ([path compare:@""] != NSOrderedSame) {
      [level setHidden:NO]; 
      NSDictionary *disk = [fm attributesOfFileSystemForPath:path error:NULL];
      if (disk) { double free,size,fs; NSString *str;
        size = [[disk objectForKey:NSFileSystemSize] doubleValue];
        free = [[disk objectForKey:NSFileSystemFreeSize] doubleValue];
        fs = floor(100.0*(1.0-(free/size)));
        str = [NSString stringWithFormat:@"'%@' %.0f%% full",path,fs];
        if (fs >= level.criticalValue) { // order assumes 'crit' > 'warn'
          if (!criticalPosted[i]) {
            [main_logger message:[str UTF8String] level:1 time:300];
            criticalPosted[i] = warningPosted[i] = TRUE;
          }
        } else criticalPosted[i] = FALSE;
        if (fs >= level.warningValue) {
          if (!warningPosted[i]) {
            [main_logger message:[str UTF8String] level:1 time:300];
            warningPosted[i] = TRUE;
          } 
        } else warningPosted[i] = FALSE; 
        [level setIntValue:fs];
      }
    } else { // no path set
      [level setHidden:YES];
    }
  }
  [fm release];
}

/* ---------------------------------------------------------------- */

- (int)updateWheel:(int)w fromGalil:(BOOL)refresh
{
  int  err=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,w,(int)refresh);
#endif
  assert((w>=0) && (w<=GALIL_WHEELS));

  if (gWheel[w].nPositions == 0) return 0;

  if (refresh) err = [gWheel[w] get];

  if (!err) {
    dispatch_async(MAIN_QUEUE,^(void) { NSString *string;
      GalilWheel *wheel = gWheel[w];
      NSPopUpButton *drop = drop_wheels[w];
      if (wheel.state < 0) string = @"?";
      else                 string = [drop itemTitleAtIndex:(1+wheel.state)];
      [drop setTitle:string];
    });
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)updateSlide:(BOOL)refresh 
{
  int  err=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,refresh);
#endif

  if (refresh) err = [gSlide get];

  if (!err) {
    dispatch_async(MAIN_QUEUE,^(void) { NSString *string;
      if (gSlide.state < 0) string = @"?";
      else                  string = [drop_slide itemTitleAtIndex:(1+gSlide.state)];
      [drop_slide setTitle:string];
    });
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (void)updateSUTR
{
  char buf[128];
  const int factors[]={3,4,5,7,11,13,17,19,23,29,31,37,41,43,47}; // TODO ?Nick,?Jason b0094
#if (DEBUG > 0)
  fprintf(stderr,"%s: rm=%d\n",PREFUN,readmode);
#endif
  
  if ((readmode == RM_RESET) || (readmode == RM_SINGLE)) {
    YeditSet(label_readmode,"");
    return;
  }
  
  nframes = 1+imax(1,my_round(staController.exptime/H2RG_FRAME_TIME,0));
  if (readmode == RM_CDS) {               // CDS XRdd..dddR
    rframes = 2;
    dframes = nframes-rframes;
    staController.exptime = (nframes-1)*H2RG_FRAME_TIME;
    [self action_exptime:nil];
  } else {
    switch (nframes) {
    case 2: case 3: case 4: case 6:      // use CDS
      rframes=2; dframes=nframes-rframes;
      break;
    case  5:                             // special case
      rframes=3; dframes=1;
      break;
    default:                             // 1+(R-1)*(D+1)
      rframes=2; dframes=nframes-2;      // default = CDS
      for (int i=0; i<sizeof(factors)/sizeof(int); i++) {
        if (((nframes-1) % factors[i]) == 0) {
          dframes = factors[i]-1;
          rframes = 1+(nframes-1)/(dframes+1);
          break;
        }
        if (factors[i] > nframes/2) break;
      }
      break;
    }
  }
  assert(rframes >= 2);
  assert(dframes >= 0);
  assert(nframes == 1+(rframes-1)*(dframes+1));
  sprintf(buf,"%dr%dd",rframes,dframes); YeditSet(label_readmode,buf);
}

/* ---------------------------------------------------------------- */
#pragma mark "Sample"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)sampleGUI
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert([NSThread isMainThread]);
  
  [self action_nloops:edit_nloops];
  
  exptype = (int)[popup_exptype indexOfSelectedItem];
  Ystr2chr([popup_exptype titleOfSelectedItem],etStr,sizeof(etStr));
  
  staController.exptime = [edit_exptime floatValue];
  [prog_exposure setMaxValue:(staController.exptime)];
  
  readmode = (int)[popup_readmode indexOfSelectedItem];
  Ystr2chr([popup_readmode titleOfSelectedItem],rmStr,sizeof(rmStr));
  
  assert(sizeof(object) > FITS_MAX_CHAR);
  YeditGet(edit_object,object,1+FITS_MAX_CHAR); // sampled again before write
  assert(sizeof(comment) > FITS_MAX_CHAR);
  YeditGet(edit_comment,comment,1+FITS_MAX_CHAR);
}

/* ---------------------------------------------------------------- */

- (int)sampleArchon:(BOOL)reset
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,reset);
#endif

  if (!staController.configured) return 0;
  
  int err = [staController statusCheck:reset];

  if (err) {
    [self showWarning:"reading Archon-status failed" error:err];
    return err;
  }

  [self updateDewar]; // TODO there's nothing? todo add sampleLakeshore

  return err;
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)sampleTCS:(BOOL)show
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  int err = [tcs open];
  
  if (!err) {
    if (!err) {
      err = [tcs getAlpha:&alpha delta:&delta equinox:&equinox];
    }
    [tcs close];
  }
  if (err && show) {
    [self showWarning:"cannot connect to TCSIS" error:err];
  }
  
  return err;
}

/* ---------------------------------------------------------------- */
#pragma mark "Actions/Other"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_file:(id)sender
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  if (sender == edit_file) {        // done
    [edit_file setBezeled:NO];
    [edit_file setEditable:NO];
    [edit_file setBordered:YES];
    [edit_file setDrawsBackground:NO];
    //[edit_file setTarget:nil];
    //[edit_file setAction:nil];
    run_number = imax(1,imin(MAX_FILENUM,[edit_file intValue]));
    sprintf(buf,"%d",run_number); YeditSet(edit_file,buf);
    YprefPutInt(DBE_RUN,run_number,NO);
  } else {                          // allow editing
    [edit_file setBezeled:YES];
    [edit_file setEditable:YES];
    [edit_file setEnabled:YES];
    //[edit_file setTarget:self];
    //[edit_file setAction:@selector(action_file:)];
    [[self window] makeFirstResponder:edit_file];
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_motor:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  
  NSInteger i = [drop_wheels indexOfObject:sender];
  [self startMotion:sender run:YES];
  if (i == NSNotFound) {                 // slide
    assert(sender == drop_slide);
    int p = (int)[sender indexOfSelectedItem]-1;
    // printf("move slide to %d\n",p);
    dispatch_async(GLOBAL_QUEUE,^(void) {
      int err = [gSlide move:p];
      if (err) { NSString *s;
        s = [NSString stringWithFormat:@"%@ failed, err=%d\r%@",
             gSlide.slideName,err,[GalilObject errorString:err]];
        [Notice openNotice:s.UTF8String title:p_exec time:30 level:1];
      }
      [self updateSlide:NO];
      [self startMotion:sender run:NO];
    });
  } else {                              // wheels
    assert(sender == drop_wheels[i]);
    GalilWheel *wheel = gWheel[i];
    int w = (int)i;
    int p = (int)[sender indexOfSelectedItem]-1;
    // printf("move wheel-%d(%c) to %d\n",1+w,wheel.axis.axis,p);
    dispatch_async(GLOBAL_QUEUE,^(void) {
      int err = [wheel move:p];
      if (err) { NSString *s;
        s = [NSString stringWithFormat:@"%@-wheel failed, err=%d\r%@",
             wheel.wheelName,err,[GalilObject errorString:err]];
        [Notice openNotice:s.UTF8String title:p_exec time:30 level:1];
      }
      [self updateWheel:w fromGalil:NO];
      [self startMotion:sender run:NO];
    });

  }

}

/* ---------------------------------------------------------------- */
#pragma mark "Actions/Exposure"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_exptime:(id)sender
{
  float exptime;
  char  buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  if (sender) exptime = [sender floatValue];
  else        exptime = staController.exptime;
  switch (readmode) {
  case RM_RESET:
    exptime = 0.0;
    break;
  case RM_SINGLE:                      // b0095
    exptime = H2RG_FRAME_TIME;
    break;
  default:
    exptime = (float)fmax(1,fmin(MAX_EXPTIME-H2RG_FRAME_TIME/2,exptime));
    break;
  }

  staController.exptime = exptime;
  if (sender) [self updateSUTR];
  exptime = staController.exptime;

  if (floor(exptime) == exptime) sprintf(buf,"%.0f",exptime);
  else if (exptime >= 100.0f)    sprintf(buf,"%.1f",exptime);
  else if (exptime >= 10.0f)     sprintf(buf,"%.2f",exptime);
  else                           sprintf(buf,"%.3f",exptime);
  YeditSet(edit_exptime,buf);
  sprintf(buf,"%.3f",exptime); self.tcpResponse = @(buf);
}

/* ---------------------------------------------------------------- */

- (IBAction)action_exptype:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  assert(!loop_doing);
  if (loop_doing) return;
  
  if (sender == popup_exptype) {         // from popup
    exptype = (int)[popup_exptype indexOfSelectedItem];
  } else {                               // as string or number
    assert(0);
  }
  
  [self enableExposure:YES];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_nloops:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  nloops = imax(1,imin(999,[sender intValue])); 
  [edit_nloops setIntValue:nloops];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_readmode:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert(!loop_doing);
  assert([NSThread isMainThread]);

  readmode = (int)[popup_readmode indexOfSelectedItem];
  switch (readmode) {
  case RM_RESET: case RM_SINGLE:
    [edit_exptime setEnabled:NO];
    break;
  default:
    [edit_exptime setEnabled:YES];
    break;
  }
  [self action_exptime:edit_exptime];
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_start:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  self.tcpResponse = @"-Efailed";
  if (loop_doing) return;
  
  if (!staController.array_power) { 
    [self showError:"Array power is off\rPlease restart Archon and GUI" error:0];
    return;
  }

  [self sampleGUI];
  
  [but_start setEnabled:NO];
  [self performSelectorInBackground:@selector(run_loop) withObject:nil];
  self.tcpResponse = @"ok";
}

/* ---------------------------------------------------------------- */

- (IBAction)action_stop:(id)sender
{
  int err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (loop_doing) {
    err = [staController abortExposure];
  }
  if (err) [self showWarning:"abort exposure failed" error:err];
}

/* ---------------------------------------------------------------- */
#pragma mark "Exposure/Loop"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)run_loop
{
  int  err=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  loop_doing = YES; [self enableControls:YES];
  
  if (!err) {
    assert(![NSThread isMainThread]);
    dispatch_sync(MAIN_QUEUE,^(void) { // mainThread
      [prog_exposure setMaxValue:(staController.exptime)]; 
      [prog_exposure setDoubleValue:0.0];
      [prog_exposure setHidden:NO];
    });
  }
  
  if (!err) {
    for (loop=1; loop<=nloops; loop++) {
      dispatch_async(MAIN_QUEUE,^(void) { [edit_loop setIntValue:loop]; });
      @autoreleasepool {
        err = [self readout_exposure];
      }
      if (err) break;
      if (staController.aborted) break;   // 'Stop' -- break loop
    } // endfor(loop)
  } // endif(err)

  if (err) {
    [self showError:"exposure loop failed" error:err];
    [Sound play:"'exposure failed'"];
  } else {
    [Sound play:snd_expdone];
  }
  dispatch_sync(MAIN_QUEUE,^(void) {
    [edit_loop setIntValue:0];
    [edit_runtime setIntValue:0];
    [prog_exposure setHidden:YES];
  });

  loop_doing = FALSE; [self enableControls:YES];

  [pool drain];
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)readout_exposure
{
  int err=0;
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p: loop=%d\n",PREFUN,self,loop);
#endif

  cur_runtime = 0.0;
#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif
  [self sampleTCS:YES];
  [self sampleArchon:NO];
#if (TIME_TEST > 0)
  sprintf(buf,"%s: samplingTime= %.3f",PREFUN,YsecSince(date));
  [main_logger append:buf];            // about 0.18 seconds TEST@SBS -- depends
  printf("%s\n",buf); //xxx
#endif
  
  if (!staController.online) {         // Simulator -- Note: after sampleTCS
    if (loop == 1) [staController setA:alpha D:delta E:equinox S:app.disperser];
    staController.shutter = (exptype==1) ? 0 : 1;  // b0091
    staController.objects = (exptype==0) ? 1 : 0;
  }

  [qltool resetTitle:nil];

  switch (readmode) {                  // readout modes
  case RM_RESET:
    err = [self readout_reset];
    break;
  case RM_SINGLE:
    err = [self readout_single];
    break;
  case RM_CDS:                        // special SUTR (R==2)
  case RM_SUTR:                       // generic SUTR
    err = [self readout_sutr:((readmode==RM_CDS) ? 1 : 0)];
    break;
  }
  if (err) return err;                // critital error 
  
  assert(write_data);                 // write result to disk
  [qltool addFrame:write_data];
  err = [self FITSsave:write_data run:run_number frame:0];
  if (err) [self showError:"Saving FITS failed" error:err];
  run_number++;
  if (run_number > MAX_FILENUM) {
    run_number = 1;
    [self showWarning:"file number overflow\rreset to '1'" error:0];
  }
  sprintf(buf,"%d",run_number); YeditSet(edit_file,buf);
  YprefPutInt(DBE_RUN,run_number,NO);
  [write_data release];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)readout_reset
{
  NSData *data=nil;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  int err = [staController startExposure:-1];
  if (err) return err;
  
  [NSThread sleepForTimeInterval:H2RG_FRAME_TIME];
  data = [staController fetchFrame:0]; // fetchNextFrame
  if (!data) return E_camera_sync;
#ifndef NDEBUG
  u_int pixels = staController.ntaps*staController.dataPixels;
  assert(data.length/(sizeof(u_short)*pixels) == staController.dataLines);
#endif

  // no reference pixel subtraction: write_data = [self makeInt:data];
  write_data = [[self refPixelSubtraction:data mode:_refPixelMode] retain];
  [data release];

  return err;
}

/* --- */

- (int)readout_sutr:(int)cdr
{
  int f=0;
  double next_sec=1.0;
  NSData *data=nil;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  [self updateSUTR];

  int err = [staController startExposure:rframes dummies:dframes];
  if (err) return err;
  
  register int npix = staController.dimx * staController.dimy;
  __block long W = 0;
  long *S = (long*)calloc(npix,sizeof(long));
  long *T = (long*)calloc(npix,sizeof(long));
  // int *Z = (int*)calloc(npix,sizeof(int));  // cosmic rays ?Nick -- saturation check TODO

  double frm_time = H2RG_FRAME_TIME;   // next frame
  double cyc_time = (1+dframes)*H2RG_FRAME_TIME;
  for (f=0; f<rframes; f++) {          // loop over read-frames
    do {
      if (staController.aborted) break;
      [NSThread sleepForTimeInterval:0.350];
      [self updateExposure:NO];  // printf("runtime=%.2f\n",cur_runtime);
      if (cur_runtime >= next_sec) { [self updateExposure:YES]; next_sec += 1.0; } // update GUI
    } while (cur_runtime < frm_time);
    if (staController.aborted) break;  // exposure stopped
    data = [staController fetchFrame:frm_time]; // fetchNextFrame
    if (!data) { printf("f=%d\n",f); break; } //xxx
#ifndef NDEBUG
    u_int pixels = staController.ntaps*staController.dataPixels;
    assert(data.length/(sizeof(u_short)*pixels) == staController.dataLines);
#endif
     // [staController printBufferState:0];
    dispatch_async(sutr_queue,^(void) { long *ptrS=S,*ptrT=T; NSData *d;
      // printf("sutr#%d -- start\n",1+f);
      NSDate *date = [NSDate date];
      if (_saveRaw) [self FITSsave:data run:run_number frame:-(1+f)];
      if (_showRaw) [qltool addRawFrame:data];
      d = [self refPixelSubtraction:data mode:_refPixelMode];
      if (_refPixelMode) {
        if (_saveFrames) [self FITSsave:d run:run_number frame:(1+f)];
        if (_showFrames) [qltool addFrame:d];
      }
      W += (f*(f+1))/2; // printf("W=%ld\n",W);     // SUTR summations
      int *ptrR = (int*)d.bytes;
      for (int i=0; i<npix; i++,ptrS++,ptrR++,ptrT++) {
        *ptrS += (long)*ptrR*f - *ptrT;
        *ptrT += (long)*ptrR;
        // *ptrZ  = *ptrR;
      }
      if (_showCurrent && (W>0) && (f<rframes-1)) { long *ptrS; int i,*ptr;
        int *result = (int*)malloc(npix*sizeof(int));
        assert(!cdr);
        double norm = (frm_time-H2RG_FRAME_TIME)/cyc_time; // printf("norm=%f\n",norm);
        for (i=0,ptrS=S,ptr=result; i<npix; i++,ptr++,ptrS++) {
          *ptr = (int)floor(0.5+(((double)*ptrS / (double)W) * norm));
        }
        [qltool copyData:result asSky:NO];
        free((void*)result);
      }
      printf("sutr#%d -- done (%.2f ms)\n",1+f,YmsSince(date)); //xxx
    });
    [data release];
    frm_time += cyc_time;           // next frame time
  }
  
  if (!data) { char buf[128];       // no data from Archon (synchronization?)
    sprintf(buf,"lost synchronization with Archon after %d frames",f);
    if (f > 1) {
      strcat(buf,"\rsaving what we received so far");
      staController.exptime = frm_time-cyc_time;
      [staController abortExposure];
    } else {
      strcat(buf,"\raborting exposure loop");
      err = E_camera_sync;
    }
    [self showWarning:buf error:err];
  } else
  if (staController.aborted) {      // stopped exposure (no failure)
    staController.exptime = frm_time-cyc_time;
    if (f < 2) err = E_camera_abort;
  }

  if (!err) {                    // no ciritcal error
    dispatch_sync(sutr_queue,^(void) { long *ptrS; int i,*ptr;
      NSDate *date = [NSDate date];
      assert(W >= 1);  assert(f > 1);
      int *result = (int*)malloc(npix*sizeof(int));
      double norm = (cdr) ? 1.0 : staController.exptime/cyc_time; // printf("norm=%f\n",norm);
      for (i=0,ptrS=S,ptr=result; i<npix; i++,ptr++,ptrS++) {
        *ptr = (int)floor(0.5+(((double)*ptrS / (double)W) * norm));
      }
      write_data = [[NSData alloc] initWithBytesNoCopy:result length:npix*sizeof(int)];
      printf("convert-time= %.2f ms\n",YmsSince(date)); //xxx
    });
  }
  free((void*)T); free((void*)S);

  return err;
}

/* --- */

- (int)readout_single  // todo fixed expTime=1.352 or XDDDDR ?
{
  double next=1.0;
  NSData *data=nil;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  nframes = 1;                         // fixed NEW b0095
  [self action_exptime:@(nframes*H2RG_FRAME_TIME)];
  
  int err = [staController startExposure:nframes];
  if (err) return err;
  
  for (int f=0; f<nframes; f++) {      // loop over frames
    do {
      if (staController.aborted) break;  // exposure stopped
      [NSThread sleepForTimeInterval:0.350];
      [self updateExposure:NO];        // update run time
      if (cur_runtime >= next) { [self updateExposure:YES]; next += 1.0; }
    } while (cur_runtime < (f+1)*H2RG_FRAME_TIME);
    if (staController.aborted) break;  // exposure stopped
    data = [staController fetchFrame:((f+1)*H2RG_FRAME_TIME)]; // fetchNextFrame
    if (!data) break;
#ifndef NDEBUG
    u_int pixels = staController.ntaps*staController.dataPixels;
    assert(data.length/(sizeof(u_short)*pixels) == staController.dataLines);
#endif
    if (f == nframes-1) { // last frame
      write_data = [self refPixelSubtraction:data mode:_refPixelMode];
      [write_data retain];
    }
    [data release];
  }
  if (!data) return E_camera_timeout; 
  if (staController.aborted) return E_camera_abort;
  
  return err;
}

/* ---------------------------------------------------------------- */
#pragma mark "FITS"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (NSData*)makeInt:(NSData*)d_in
{
  int *result,*r;
  
  int npix = staController.dimx*staController.dimy;
  assert(d_in.length == npix*sizeof(u_short));
  result = r = (int*)malloc(npix*sizeof(int));
  u_short *p = (u_short*)d_in.bytes;
  for (int i=0; i<npix; i++,r++,p++) { *r = (int)*p; }
  NSData *d_out = [NSData dataWithBytesNoCopy:result length:npix*sizeof(int)];
  return d_out;
}

/* --- */

- (int)FITSsave:(NSData*)data run:(int)run frame:(int)frame
{
  FITS *fits=nil;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d,%d)\n",PREFUN,data,run,frame);
#endif
  assert(![NSThread isMainThread]);

#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif
  dispatch_sync(MAIN_QUEUE,^(void) { char buf[256];
    YeditGet(edit_object,buf,sizeof(buf));
    strcpy(object,trim_string(buf,FITS_MAX_CHAR));
    YeditGet(edit_comment,buf,sizeof(buf));
    strcpy(comment,trim_string(buf,FITS_MAX_CHAR));
  });

  int npix = staController.dimx*staController.dimy;
  NSData *d=data;
  @autoreleasepool {
    if (d.length != npix*sizeof(int)) d = [self makeInt:data];
    fits = [self FITSopen:d run:run frame:frame];
    if (fits) {
      [self FITSwrite:d to:fits];
      [fits close:((frame==0) ? 1 : 0)];
    }
  }
#if (TIME_TEST > 1)
  fprintf(stderr,"%s: %.2f [ms]\n",PREFUN,YmsSince(date));
#endif
  if (!fits) return E_camera_fits;
  
  return 0;
}

/* ---------------------------------------------------------------- */

- (FITS*)FITSopen:(NSData*)data run:(int)run frame:(int)frame
{
  int d;
  char buf[128],com[128],file[512];
  struct tm res;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,run,frame);
#endif

  int dimx = staController.ntaps*staController.dataPixels;
  assert(staController.dimx == dimx);
  int dimy = staController.dataLines;
  assert(staController.dimy == dimy);
  assert(data.length == dimx*dimy*sizeof(int));
  FITS *fits = [[FITS alloc] initWithBase:DEF_PREFIX width:dimx height:dimy];

  int npix = staController.dimx * staController.dimy;
  assert(data.length == npix*sizeof(int));
  int *bytes = (int*)data.bytes;
  int min=2000000000,max=-2000000000;
  for (int i=0; i<npix; i++,bytes++) {
    if (*bytes < min) min = *bytes;
    if (*bytes > max) max = *bytes;
  }
  // printf("min=%d, max=%d\n",min,max); 
  if (max > 0x7fff) {
    fits.bzero = 0x8000 - 1000;
  } else {
    fits.bzero = 0.0;
  }

  assert(MAX_FILENUM == 9999);
  if      (frame > 0) sprintf(file,"%s%04d_%03d.fits",DEF_PREFIX,run,frame);
  else if (frame < 0) sprintf(file,"%s%04d_%03dr.fits",DEF_PREFIX,run,-frame);
  else                sprintf(file,"%s%04d.fits",DEF_PREFIX,run);
  if (![fits open:file paths:datapaths]) { NSString *text;
    text = [NSString stringWithFormat:@"failed to open %@",fits.failedFile];
    [main_logger message:[text UTF8String]];
    [fits release];
    return nil;
  }
  
  [fits charValue:"LCO/OCIW" forKey:"ORIGIN" comment:NULL];
  [fits charValue:observer forKey:"OBSERVER" comment:NULL];
  [fits charValue:"Swope" forKey:"TELESCOP" comment:NULL];
  [fits charValue:"LCO" forKey:"SITENAME" comment:NULL];
  [fits intValue:DEF_TELALT forKey:"SITEALT" comment:"meters"]; 
  [fits floatValue:DEF_TELLAT forKey:"SITELAT" prec:5 comment:"latitude"];
  [fits floatValue:DEF_TELLONG forKey:"SITELONG" prec:5 comment:"longitude"];
  [fits intValue:3 forKey:"TIMEZONE" comment:NULL];

  time_t ut_start = (time_t)my_round(staController.start_int,0);
  gmtime_r(&ut_start,&res);
  // was strftime(buf,sizeof(buf),"%Y-%m-%d",&res);
  strftime(buf,sizeof(buf),"%FT%H:%M:%S",&res);
  [fits charValue:buf forKey:"DATE-OBS" comment:"(start)"];
  strftime(buf,sizeof(buf),"%Y-%m-%d",&res);
  [fits charValue:buf forKey:"UT-DATE" comment:"UT date (start)"];

  strftime(buf,sizeof(buf),"%H:%M:%S",&res);
  sprintf(com,"UT time (start) %d seconds",get_uts(ut_start));
  [fits charValue:buf forKey:"UT-TIME" comment:com];

#if 0
  time_t ut_stop = (time_t)my_round(staController.stop_int,0); // ?Nick
  gmtime_r(&ut_stop,&res);
  strftime(buf,sizeof(buf),"%H:%M:%S",&res);
  sprintf(com,"UT time (end) %d seconds",get_uts(ut_stop));
  [fits charValue:buf forKey:"UT-END" comment:com];
#endif

  localtime_r(&ut_start,&res);
  strftime(buf,sizeof(buf),"%H:%M:%S",&res);
  sprintf(com,"local time (start) %d seconds",res.tm_hour*3600+res.tm_min*60+res.tm_sec);
  [fits charValue:buf forKey:"LC-TIME" comment:com];

  [fits charValue:get_night(buf,ut_start) forKey:"NIGHT" comment:"local night"];

  [fits charValue:P_TITLE forKey:"INSTRUME" comment:"instrument name"];
  [fits floatValue:pixscale forKey:"SCALE" prec:3 comment:"arcsec/pixel"];
  [fits floatValue:staController.egain forKey:"EGAIN" prec:2 comment:"electrons/DU"];
  [fits floatValue:staController.enoise forKey:"ENOISE" prec:2 comment:"electrons"];
  
  [fits charValue:alpha_str(buf,alpha) forKey:"RA" comment:"right ascension"];
  [fits floatValue:alpha*15.0 forKey:"RA-D" prec:7 comment:"degrees"];
  [fits charValue:delta_str(buf,delta) forKey:"DEC" comment:"declination"];
  [fits floatValue:delta forKey:"DEC-D" prec:7 comment:"degrees"];
  [fits floatValue:equinox forKey:"EQUINOX" prec:5 comment:"of above coordinates"];
  [fits floatValue:get_epoch(ut_start) forKey:"EPOCH" prec:5 comment:"epoch (start)"];
  double am = get_airmass(DEF_TELLAT,delta,get_ha(ut_start,DEF_TELLONG,alpha));
  [fits floatValue:am forKey:"AIRMASS" prec:3 comment:"airmass (start)"];
  double st = get_sidereal(ut_start,DEF_TELLONG);
  sprintf(com,"sidereal time: %s (start)",time_str(buf,ut_start));
  [fits floatValue:st forKey:"ST" prec:1 comment:com];
#if 0
  [fits floatValue:rota forKey:"ROTANGLE" prec:2 comment:"rotator offset angle"];
  [fits floatValue:para forKey:"PARANGLE" prec:2 comment:"parallactic angle"];
#endif
  [fits charValue:file forKey:"FILENAME" comment:NULL]; 
  [fits charValue:object forKey:"OBJECT" comment:"object name"];
  [fits charValue:rmStr forKey:"READMODE" comment:"read mode"];
  [fits charValue:etStr forKey:"EXPTYPE" comment:"exposure type"];

  d = (floor(staController.exptime) == staController.exptime) ? 1 : 3;
  [fits floatValue:my_round(staController.exptime,d) forKey:"EXPTIME"
        prec:d comment:"exposure time [seconds]"];
  [fits floatValue:staController.start_int forKey:"EXPSTART" prec:1 comment:"exposure start time"];

  [fits charValue:comment forKey:"COMMENT" comment:"object comment"];
  [fits intValue:nloops forKey:"NLOOPS" comment:"# of loops per sequence"];
  [fits intValue:loop forKey:"LOOP" comment:"# within this sequence"];
  [fits charValue:[staController.acf_file UTF8String] forKey:"ACF-FILE" comment:"Archon configuration"];
  [fits charValue:[staController.acf_version UTF8String] forKey:"ACF-VERS" comment:".acf version"];

#if 0 // TODO TRIM \reference pixels [2040 * 2040]
  sprintf(buf,"[%d:%d,%d:%d]",1,(t2-t1)*staController.dataPixels,1,staController.dataLines);
  [fits charValue:buf forKey:KEY_TRIMSEC comment:"NOAO: trim section"];
  [fits charValue:buf forKey:KEY_DATASEC comment:"NOAO: data section"]; 
  sprintf(buf,"[%d:%d,%d:%d]",(t2-t1)*staController.dataPixels+1,dw,1,dh);
  [fits charValue:buf forKey:"BIASSEC" comment:"NOAO: bias section"];
#endif
  [fits intValue:1 forKey:"DISPAXIS" comment:"dispersion axis"];

#if 0 // ?Nick
  [fits expValue:pfsController.dewar_vac forKey:"VGAUGE" prec:2
        comment:"dewar vacuum gauge [mbar]"];
  [fits expValue:pfsController.ion_vac forKey:"IONPUMP" prec:2
        comment:"ion-pump gauge [mbar]"];
  [fits expValue:pfsController.grating_vac forKey:"CGAUGE" prec:2
        comment:"grating vacuum gauge [mbar]"];
#endif

#if 0 // TODO from Lake
  [fits floatValue:staController.temp_a forKey:"TSENSORA" prec:1
        comment:"CCD temperature (controlled) [C]"];
  [fits floatValue:staController.temp_b forKey:"TSENSORB" prec:1
        comment:"CCD temperature (floating) [C]"];
  [fits floatValue:staController.temp_c forKey:"TSENSORC" prec:1
        comment:"getter temperature [C]"];
  [fits floatValue:staController.heat_a forKey:"HEATERA" prec:3
        comment:"CCD heater [W]"];
#endif

  sprintf(buf,"Version %s (%s)",p_version,p_build);
  [fits charValue:buf forKey:"SOFTWARE" comment:NULL];
  [fits  charValue:"0095" forKey:"FITSVERS" comment:"FITS header version"];

  [fits endHeader];
  
  return [fits autorelease]; 
}

/* ---------------------------------------------------------------- */

- (void)FITSwrite:(NSData*)data to:(FITS*)fits
{
  int     x,y,v,*idata;
  short   *sdata,*shortbuf;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p)\n",PREFUN,data,fits);
#endif
  assert(fits);
  
#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif
  int dimx = staController.dimx;
  int dimy = staController.dimy;
  shortbuf = (short*)malloc(dimx*sizeof(short));
  int bzero = (int)fits.bzero;

  idata = (int*)data.bytes;
  for (y=0; y<dimy; y++) {                // loop over lines
    sdata = shortbuf;
    for (x=0; x<dimx; x++,sdata++,idata++) {
      v = *idata - bzero;
      if      (v >  0x7fff) *sdata =  0x7fff;
      else if (v < -0x8000) *sdata = -0x8000;
      else                  *sdata = (short)v;
    }
    [fits writeShort:shortbuf];
  }
  free((void*)shortbuf);
#if (TIME_TEST > 1)
  printf("fitsTime= %.3f\n",YsecSince(date));
#endif
}

/* ---------------------------------------------------------------- */
#pragma mark "Referene Pixel Subtraction"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (NSData*)refPixelSubtraction:(NSData*)d_in mode:(int)mode
{
  NSData *d_out=nil;

#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif

  switch (mode) {
  case 0:                     // no reference pixel subtraction
    d_out = [self makeInt:d_in];
    break;
  case 1:                     // just column (bias) correction
    d_out = [self refPixelChannels:d_in];
    break;
  case 2:                     // FIRE/FourStar (time variable) b0085 TEST@SBS
    d_out = [self refPixelSubtraction:d_in];
    break;
  }
#if (TIME_TEST > 1)
  printf("refPixSubTime(%d)= %.2f ms\n",mode,YmsSince(date));
#endif
  return d_out;
}

/* --- */

- (NSData*)refPixelSubtraction:(NSData*)d_in
{
  int c,x,y,n,yy,x2,dy;
  int nc   = staController.ntaps;      // # channels
  int np   = staController.dataPixels;
  int dimx = np * nc;
  int dimy = (int)d_in.length / (sizeof(u_short)*dimx);
  int ref_x[N_CHANNELS],*ref_y;
  assert(dimx == dimy);

  u_short *data = (u_short*)d_in.bytes;
  int *sortval = (int*)malloc(dimy*np*sizeof(int)); // TODO IVAR ?
  
  for (y=0,n=0; y<dimy; y++) {         // left-most channel
    yy = y*dimx;
    x2 = (y < REF_PIXELS) ? np : REF_PIXELS;
    for (x=0; x<x2; x++,yy++) { sortval[n++] = data[yy]; }
  }
  q_sort(sortval,0,n-1);
  ref_x[0] = (n % 2) ? sortval[n/2] : (sortval[n/2]+sortval[n/2-1])/2;
  // printf("left=%d (%d)\n",ref_x[0],n);

  for (y=0,n=0; y<dimy; y++) {         // right-most channel
    yy = (y+1)*dimx - ((y < REF_PIXELS) ? np : REF_PIXELS);
    x2 = (y < REF_PIXELS) ? np : REF_PIXELS;
    for (x=0; x<x2; x++,yy++) { sortval[n++] = data[yy]; }
  }
  q_sort(sortval,0,n-1);
  ref_x[nc-1] = (n % 2) ? sortval[n/2] : (sortval[n/2]+sortval[n/2-1])/2;
  // printf("right=%d (%d)\n",ref_x[nc-1],n);
  
  ref_y = (int*)calloc(dimy,sizeof(int));
  const int ref_lines=3;               // 3+1+3 = 7 TODO on preferences
  assert(ref_lines <= REF_PIXELS);
  for (y=0; y<dimy; y++) {
    int y1 = imax(0,y-ref_lines);
    int y2 = imin(dimy-1,y+ref_lines);
    for (n=0,dy=y1; dy<=y2; dy++) {
      yy = dy*dimx;
      for (x=0; x<REF_PIXELS; x++) { sortval[n++] = data[yy++] - ref_x[0]; }
      yy = (dy+1)*dimx-REF_PIXELS;
      for (x=0; x<REF_PIXELS; x++) { sortval[n++] = data[yy++] - ref_x[nc-1]; }
    }
    q_sort(sortval,0,n-1);
    ref_y[y] = (n % 2) ? sortval[n/2] : (sortval[n/2]+sortval[n/2-1])/2;
    //printf("%d %d , %d %d (%d,%d)\n",y,ref_y[y],sortval[0],sortval[n-1],n,n/(2*REF_PIXELS));
  }

  for (c=1; c<nc-1; c++) {             // middle channels
    for (n=0,dy=0; dy<REF_PIXELS; dy++) {
      y = dy;         yy = y*dimx+c*np;
      for (x=0; x<np; x++) { sortval[n++] = data[yy++] - ref_y[y]; }
      y = dimy-dy-1; yy = y*dimx+c*np;
      for (x=0; x<np; x++) { sortval[n++] = data[yy++] - ref_y[y]; }
    }
    q_sort(sortval,0,n-1);
    ref_x[c] = (n % 2) ? sortval[n/2] : (sortval[n/2]+sortval[n/2-1])/2;
    //printf("%d %d (%d)\n",c,ref_x[c],n); 
  }
  free((void*)sortval);

//  y2 = dimy-REF_PIXELS;                // correct data area only TODO ?
//  x2 = dimx-REF_PIXELS;
  int *result = (int*)malloc(dimx*dimy*sizeof(int));
  for (x=0; x<dimx; x++) {
    c = x/np;
    for (y=0; y<dimy; y++) {
      result[x+y*dimx] = (int)data[x+y*dimx] - (ref_x[c]+ref_y[y]);
    }
  }
  NSData *d_out = [NSData dataWithBytesNoCopy:result length:dimx*dimy*sizeof(int)];
  free((void*)ref_y);

  return d_out;
}

/* -- */

- (NSData*)refPixelChannels:(NSData*)d_in
{
  int c,x,y,n,yy,dy;
  int nc   = staController.ntaps;      // # channels
  int np   = staController.dataPixels;
  int dimx = np * nc;
  int dimy = staController.dataLines;
  int ref_x[N_CHANNELS];
  assert(dimy == d_in.length / (sizeof(u_short)*dimx));
  assert(dimx == dimy);
  assert(nc == N_CHANNELS);

  u_short *data = (u_short*)d_in.bytes;
  int *sortval = (int*)malloc(dimy*np*sizeof(int)); // TODO IVAR ?
  
  for (c=0; c<nc; c++) {             // loop over channels
    for (n=0,dy=0; dy<REF_PIXELS; dy++) {
      y = dy;        yy = y*dimx+c*np;
      for (x=0; x<np; x++) { sortval[n++] = data[yy++]; }
      y = dimy-dy-1; yy = y*dimx+c*np;
      for (x=0; x<np; x++) { sortval[n++] = data[yy++]; }
    }
    q_sort(sortval,0,n-1);
    ref_x[c] = (n % 2) ? sortval[n/2] : (sortval[n/2]+sortval[n/2-1])/2;
    // printf("%d %d (%d)\n",c,ref_x[c],n);
  }
  free((void*)sortval);

  int *result = (int*)malloc(dimx*dimy*sizeof(int));
  for (x=0; x<dimx; x++) {
    c = x/np;
    for (y=0; y<dimy; y++) {
      result[x+y*dimx] = (int)data[x+y*dimx] - (ref_x[c]);
    }
  }
  NSData *d_out = [NSData dataWithBytesNoCopy:result length:dimx*dimy*sizeof(int)];

  return d_out;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
