/* ---------------------------------------------------------------- */
//
// DataSimulator.m
// Henrietta
//
// Created by Christoph Birk on 2022-09-20
// Copyright (c) 2022 Christoph Birk. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG           1
#define TIME_TEST       0

/* ---------------------------------------------------------------- */

#import  "DataSimulator.h"
#import  "Yutils.h"
#import  "main.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

#define N_THREADS       2
#define ANG             (1.e-10)       // Angstrom
#define N_SPECT         (400/LO_MEM)
#define N_SPREAD        (21/LO_MEM)

#define KTC_NOISE       60.0

#define SQD(x)          (((double)(x))*((double)(x)))

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation FovStar

+ (FovStar*)starAtX:(float)x Y:(float)y M:(float)m
{
  FovStar *s = [[FovStar alloc] initAtX:x Y:y M:m];
  
  return [s autorelease];
}

/* --- */

- (FovStar*)initAtX:(float)x Y:(float)y M:(float)m
{
  self = [super init];
  if (self) {
    _xc = x; _yc = y; _mag = m;
  }
  return self;
}

@end

/* ---------------------------------------------------------------- */

@implementation ArraySimulator
{
  int dimx,dimy;
  int *detector;                       // electrons
  float last_exptime;
  double cfzm;
  NSMutableArray *fovlist;
}

@synthesize name,data0,dimx,dimy,scat;

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

static char cname='A';

- (id)initW:(int)w H:(int)h X:(float)x Y:(float)y
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  self = [super init]; if (!self) return nil;

  assert([self class]);
  @synchronized([self class]) {
    name = [[NSString alloc] initWithFormat:@"%c%+.0f%+.0f",cname++,x,y];
    // NSLog(@"%@",name);
  }
  data0  = NULL;
  width  = w;                          // pixels
  height = h;
  choffx = x;                          // offset [arcsec]
  choffy = y;

  _egain  = 1.0;                       // maybe overwritten later
  _enoise = 0.5; // 9.0/sqrt(2); // TODO test with readmodes
  
  twister = [[MTwister alloc] init];
  fovlist = [[NSMutableArray alloc] initWithCapacity:32];
  
  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  [scat release];
  [twister release];
  if (data0) { free((void*)data0); data0=NULL; }
  if (detector) { free((void*)detector); detector=NULL; }
  [name release];
  
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (int)poisson:(double)mean
{
  int k=0;
  
  if (mean < 20.0) {                  // proper poisson
    double p = exp(-mean);
    double s = p;
    double u = [twister doubleValue:1.0];
    while (u > s) {
      k += 1;
      p *= mean/(double)k;
      s += p;
    }
  } else {                            // use gauss
    double x = [twister gaussMean:mean sigma:sqrt(mean)];
    if (x > 0.0) k = (int)floor(x+0.5);
  }
  return k;
}

/* ---------------------------------------------------------------- */

- (void)create_bias
{
  auto     int     y,x1,x2,y2;
  auto     double  bias;
  register int     i,j=0,x,nsamp;
  register int     *d,*array;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif

  bias  = _egain * (double)_bias_level;
  nsamp = 17+dimx+dimy;
  array = (int*)malloc(nsamp*sizeof(int));
  for (i=0; i<nsamp; i++) {             // create array of bias values
    do {
      array[i] = (int)floor([twister gaussMean:bias sigma:KTC_NOISE]+0.5);
    } while (array[i] <= 0);
  }

  d = detector;
  for (y=0; y<REF_PIXELS; y++) {        // top rows -- reference pixels
    for (x=0; x<dimx; x++,d++) *d = (int)bias;
  }
  x1 = (_leftRef) ? REF_PIXELS : 0;
  x2 = (_rightRef) ? dimx-REF_PIXELS : dimx;
  y2 = dimy-REF_PIXELS;
  for (y=REF_PIXELS; y<y2; y++) {       // middle rows
    if (_leftRef) {
      for (x=0; x<REF_PIXELS; x++,d++) *d = (int)bias;
    }
    for (x=x1; x<x2; x++,d++) {
      if (j >= nsamp) j = [twister intValue:nsamp]; // Clang wants '>='
      *d = array[j++];
    }
    if (_rightRef) {
      for (x=dimx-REF_PIXELS; x<dimx; x++,d++) *d = (int)bias;
    }

  }
  for (y=y2; y<dimy; y++) {             // bottom rows -- reference pixels
    for (x=0; x<dimx; x++,d++) *d = (int)bias; 
  }
  free((void*)array);

#if 1 // mark as simulated data
  int x0=8,y0=9,*logo;                  /* add smiley */
  static int logo1[]={  0,0 , 0,+1, 0,+2 ,     /*  :-)  */
                       -2,+3,-1,+4, 0,+4 , +1,+4 , +2,+3 ,
                       -2,-1 , +2,-1 };
  static int logo2[]={  0,0 , 0,+1, 0,+2 ,     /*  :-(  */
                       -2,+5,-1,+4, 0,+4 , +1,+4 , +2,+5 ,
                       -2,-1 , +2,-1 };
  logo = [twister intValue:9] ? logo1 : logo2;
  for (i=0; i<sizeof(logo1)/sizeof(int); i+=2) {
    detector[x0+logo[i]+dimx*(y0-logo[i+1])] = 0x4000;
  }
#endif

#if (TIME_TEST > 1)
  fprintf(stderr,"%s(%s): %.2f sec\n",PREFUN,[name UTF8String],YsecSince(date));
#endif
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)add_back:(double)itime
{
  auto     int    y;
  auto     double bkgr,flux,dark;
  const    double dark_cur=1.0;     
  register int    i,x,nsamp,*d,*array;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%.3f)\n",PREFUN,[name UTF8String],itime);
#endif
#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif

  dark = dark_cur * itime;            // dark current in electrons
  bkgr = (_shutter) ? mag2flux(_back,cfzm) *itime*_pixscale*_pixscale : 0.0; 
  flux = dark + bkgr * _egain;        // background in electrons
#if (DEBUG > 1)
  fprintf(stderr,"%s: dark=%.1f, bkgr=%.1f (%.1f)\n",PREFUN,dark,bkgr,flux);
#endif

  nsamp = 17+dimx+dimy;
  array = (int*)malloc(nsamp*sizeof(int));
  for (i=0; i<nsamp; i++) array[i] = [self poisson:flux];
  
  int y2 = dimy-REF_PIXELS;
  int x1 = (_leftRef) ? REF_PIXELS : 0;
  int x2 = (_rightRef) ? dimx-REF_PIXELS : dimx;
  for (y=REF_PIXELS; y<y2; y++) {   // loop over lines
    d = detector + y*dimx + x1;
    for (x=x1; x<x2; x++,d++) {  // loop over columns
      if (i >= nsamp) i = [twister intValue:nsamp]; // Clang wants '>='
      *d += array[i++];
    }
  }
  free((void*)array);

#if (TIME_TEST > 1)
  fprintf(stderr,"%s(%s): %.2f msec\n",PREFUN,[name UTF8String],YmsSince(date));
#endif
}

/* ---------------------------------------------------------------- */

- (void)add_stars:(double)itime
{
  int    x1,x2,y2,f3,px,py,x,y;
  float  cx,cy;
  double f2,peak,flux,fw,c0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%.3f): %lu\n",PREFUN,[name UTF8String],itime,[fovlist count]);
#endif
#if (TIME_TEST > 2)
  NSDate *date = [NSDate date];
#endif

  fw = _fwhm * sqrt(1.0+fabs(_blur));
  c0 =  cfzm * sqrt(1.0+_gain);
  f2 = 2.0 * pow(fw/(2.355*_pixscale),2.0);
  f3 = 1 + (int)(3.0*fw/_pixscale);

  for (FovStar *star in fovlist) {
    cx = (flip_x) ? (float)dimx-star.xc-1.0 : star.xc;
    cy = (flip_y) ? (float)dimy-star.yc-1.0 : star.yc;
    px = (int)floor(cx+0.5);
    py = (int)floor(cy+0.5);
    //assert(dimx == dimy);
    //if (_transpose) { int h=px; px=py; py=h; float f=cx; cx=cy; cy=f; }
    peak = _egain * flux2peak(itime*mag2flux(star.mag,c0),fw,_pixscale); // [e-]
    peak /= (1.0+_neutral_filter);
    y2 = imin(dimy-REF_PIXELS,py+f3);
    x1 = (_leftRef) ? imax(REF_PIXELS,px-f3) : imax(0,px-f3);
    x2 = (_rightRef) ? imin(dimx-REF_PIXELS,px+f3) : imin(dimx,px+f3);
    for (y=imax(REF_PIXELS,py-f3); y<y2; y++) {
      for (x=x1; x<x2; x++) {
        flux = peak*exp(-SQD((float)x-cx)/f2-SQD((float)y-cy)/f2); // [e-]
        detector[x+y*dimx] += [self poisson:flux];
      } // endfor(x)
    } // endfor(y)
  } // endfor(i)

#if (TIME_TEST > 2)
  fprintf(stderr,"%s(%s): %.2f sec\n",PREFUN,[name UTF8String],YsecSince(date));
#endif
}

/* ---------------------------------------------------------------- */

- (double)star_temp:(float)mag
{
  double ave,noise,r;

  ave = fmax(0.0,10000.0-mag*500.0);
  noise = sqrt(fmax(1.0,10000.0-mag*250.0));
  do { r = 2000.0+[twister gaussMean:ave sigma:noise]; } while (r < 1000);
  return r;
}

/* --- */

static double fill_spectrum(double lc,double rs,double t,double* spc,int n_spc,
                            double f2,double* spr,int n_spr)
{
  int    i,j;
  double lambda,s=0.0;
  const double C=2.998e+8,K=1.381e-23,H=6.626e-34;

  for (i=0; i<n_spr; i++) {            /* spread */
    spr[i] = exp(-SQD(i-n_spr/2)/f2);
  }
  for (i=0; i<n_spc; i++) {            /* spectrum */
    lambda = ANG * (lc * pow(1.+1/rs,(double)(i-n_spc/2)));
#if 0 /* energy distribution */
    spc[i] = 8.0*M_PI*H*C* pow(lambda,-5.0) / (exp((H*C)/(lambda*K*t))-1.0);
#else /* photon distribution */
    spc[i] = 8.0*M_PI*     pow(lambda,-4.0) / (exp((H*C)/(lambda*K*t))-1.0);
#endif
  }
  for (i=0; i<n_spc; i++) {            /* calculate norm */
    for (j=0; j<n_spr; j++) s += spc[i] * spr[j];
  }

  return s;
}

/* --- */

- (void)add_spect:(double)itime
{
  const int DEF_DORIENT=0;
  int       x1,x2,y2,px,py,x,y,n_spect,spike;
  double    f2,flux,norm,f,lambda0,res,t,fw,c0;
  double    rho[N_SPECT+1],spr[N_SPREAD];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s): %lu\n",PREFUN,[name UTF8String],[fovlist count]);
#endif
#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif

  lambda0 = 5000.0;                    /* center wavelength */
  n_spect = N_SPECT;
  res     = 600.0;
  spike   = -n_spect/4;
  fw      = _fwhm * sqrt(1.0+fabs(_blur));
  c0      =  cfzm * sqrt(1.0+_gain);
  f2      = fmax(5.0,2.0*pow(fw/(2.355*_pixscale),2.0));

  for (FovStar *s in fovlist) {
    px = (flip_x) ? dimx-(int)s.xc-1 : (int)s.xc;
    py = (flip_y) ? dimy-(int)s.yc-1 : (int)s.yc;
    //assert(dimx == dimy);
    //if (_transpose) { int h=px; px=py; py=h; }
    t  = [self star_temp:s.mag];       /* star temperature */
    flux = _egain * itime*mag2flux(s.mag,c0);   /* total flux [e-] */
    flux /= (1.0+_neutral_filter);
    norm = fill_spectrum(lambda0,res,t,rho,n_spect,f2,spr,N_SPREAD);
    if (DEF_DORIENT == 0) {            /* normal orientation */
      y2 = imin(dimy-REF_PIXELS,py+1+n_spect/2);
      x1 = (_leftRef) ? imax(REF_PIXELS,px-N_SPREAD/2) : imax(0,px-N_SPREAD/2);
      x2 = (_rightRef) ? imin(dimx-REF_PIXELS,px+1+N_SPREAD/2) : imin(dimx,px+1+N_SPREAD/2);
      for (y=imax(REF_PIXELS,py-n_spect/2); y<y2; y++) {
        for (x=x1; x<x2; x++) {
          f = (flux/norm) * rho[y-py+n_spect/2] * spr[x-px+N_SPREAD/2];
#if 1 // TEST
          if (y==py)       f /= 2.0;   /* center */
          if (y==py+spike) f *= 2.0;   /* halfway to 'color' end */
#endif
          detector[x+y*dimx] += [self poisson:f];
        } /* endfor (x) */
      } /* endfor (y) */
    } else {                           /* nod/shuffle orientation */
      x1 = (_leftRef) ? imax(REF_PIXELS,px-n_spect/2) : imax(0,px-n_spect/2);
      x2 = (_rightRef) ? imin(dimx-REF_PIXELS,px+1+n_spect/2) : imin(dimx,px+1+n_spect/2);
      y2 = imin(dimy-REF_PIXELS,py+1+N_SPREAD/2);
      for (x=x1; x<x2; x++) {
        for (y=imax(REF_PIXELS,py-N_SPREAD/2); y<y2; y++) {
          f = (flux/norm) * rho[x-px+n_spect/2] * spr[y-py+N_SPREAD/2] * _egain;
#if 1 // TEST
          if (x==px)       f /= 2.0;   // center
          if (x==px+spike) f *= 2.0;   // halfway to 'color' end
#endif
          detector[x+y*dimx] += [self poisson:f];
        } /* endfor (y) */
      } /* endfor (x) */
    }
  }  /* endfor (i) */

#if (TIME_TEST > 1)
  fprintf(stderr,"%s(%s): %.2f sec\n",PREFUN,[name UTF8String],YsecSince(date));
#endif
}

/* ---------------------------------------------------------------- */

- (void)reset
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  cfzm = _efzm/_egain;
  dimx = width;                      // no binning
  dimy = height;
  detector = (int*)realloc(detector,dimx*dimy*sizeof(int)); // electrons
  data0 = (u_short*)realloc(data0,dimx*dimy*sizeof(u_short)); // ADU

  last_exptime = 0.0;
  [self create_bias];
  [fovlist removeAllObjects];

  if (_shutter) { StarList *list=nil;
    if (scat) {
      float  w = _pixscale * width;
      float  h = _pixscale * height;
      double a = _alpha + choffx/(3600.0*15.0*cos(_delta/RADS));
      double d = _delta + choffy/3600.0;
      list = [scat getList:a D:d E:_equinox W:(w/1.0f) H:(h/1.0f) M:30];
    } else {
      for (int i=0; i<16/LO_MEM; i++) {  // random list
        float xc = 4+[twister intValue:(dimx-8)];
        float yc = 4+[twister intValue:(dimy-8)];
        float mg = [twister gaussMean:(12.0+LO_MEM/2) sigma:2.0];
        FovStar *star = [FovStar starAtX:xc Y:yc M:mg];
        [fovlist addObject:star];
      }
    }
    if ([list.array count]) { double da; float xc,yc;
      double cosd15=3600.0*15.0*cos(_delta/RADS);
      assert(fovlist);
      for (CatStar *s in list.array) {
        da = (list.alpha - s.alpha);
        if      (da >  12.0) da = da-24.0;
        else if (da < -12.0) da = da+24.0;
        xc = dimx/2+(float)[twister gaussMean:(da*cosd15/_pixscale) sigma:0.05];
        yc = dimy/2+(float)[twister gaussMean:(3600.0*(list.delta-s.delta)/_pixscale) sigma:0.05];
        if ((xc < 1.0) || (xc > dimx-2)) continue;
        if ((yc < 1.0) || (yc > dimy-2)) continue;
        FovStar *star = [FovStar starAtX:xc Y:yc M:s.mag];
        [fovlist addObject:star];
      }
    }
  }
}

/* ---------------------------------------------------------------- */

- (void)update:(float)cur_exptime
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.3f)\n",PREFUN,cur_exptime);
#endif
  assert(fovlist);
  
  if (cur_exptime == 0) { [self reset]; return; } 

  double itime = cur_exptime - last_exptime;
  assert(itime > 0);
  
  [self add_back:itime];        // uses 'shutter'
  if (_objects) {               // b0091
    if (_disperser) [self add_spect:itime];
    else            [self add_stars:itime];
  }
  last_exptime = cur_exptime;
}

/* ---------------------------------------------------------------- */

// TODO verify readnoise reduction in various modes

- (void)readout:(double)phase
{
  int *array,i,x,y;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  int nsamp = 17+dimx+dimy;
  array = (int*)malloc(nsamp*sizeof(int));
  for (i=0; i<nsamp; i++) {                 // readnoise
    array[i] = (int)floor([twister gaussMean:0.0 sigma:_enoise]+0.5);
    // poisson(_enoise*_enoise);
  }

  int *d = detector;
  u_short *p = data0;                       // add readnoise and drift
  for (y=0; y<dimy; y++) {
    double drift = _drift_a*sin(phase+y*_drift_f);  // [DU]
    for (x=0; x<dimx; x++) {
      if (i >= nsamp) i = [twister intValue:nsamp];
      *p = (u_short)floor(0.5+drift+(double)(*d+array[i++])/_egain);
      d++; p++;
    }
  }
  free((void*)array);
  
#if 0 // mark as simulated data
  int x0=8,y0=9,*logo;                  /* add smiley */
  static int logo1[]={  0,0 , 0,+1, 0,+2 ,     /*  :-)  */
                       -2,+3,-1,+4, 0,+4 , +1,+4 , +2,+3 ,
                       -2,-1 , +2,-1 };
  static int logo2[]={  0,0 , 0,+1, 0,+2 ,     /*  :-(  */
                       -2,+5,-1,+4, 0,+4 , +1,+4 , +2,+5 ,
                       -2,-1 , +2,-1 };
  logo = [twister intValue:9] ? logo1 : logo2;
  for (i=0; i<sizeof(logo1)/sizeof(int); i+=2) {
    data0[x0+logo[i]+dimx*(y0-logo[i+1])] = 0x4000;
  }
#endif
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

@implementation DataSimulator
{
  float pixscale;
  dispatch_queue_t disp_queue[N_THREADS];
  dispatch_group_t disp_group;
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)initWithPixscale:(float)ps
{
  BOOL k=FALSE;
  int corder[]={SCAT_GAIA17,SCAT_USNO,SCAT_USNO15,SCAT_TYCHO,SCAT_SAO};
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.3f)\n",PREFUN,ps);
#endif

  self = [super init]; if (!self) return nil;

  for (int i=0; i<N_THREADS; i++) {
    disp_queue[i] = dispatch_queue_create(NULL,DISPATCH_QUEUE_SERIAL);
  }
  disp_group = dispatch_group_create();

  arrays = [[NSMutableArray alloc] initWithCapacity:32];

  _disperser = 0;
  _shutter = 1;
  pixscale = ps;

  scat = [[StarCatalog alloc] init];
  for (int i=0; i<sizeof(corder)/sizeof(int); i++) {
    k = [scat loadCatalog:corder[i] path:nil];
    if (k) break;
  }
  if (!k) { [scat release]; scat = nil; }
  
  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  [arrays removeAllObjects];
  [arrays release];
  [scat release];
  dispatch_release(disp_group);
  for (int i=0; i<N_THREADS; i++) dispatch_release(disp_queue[i]);
 
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (void)addArrayW:(int)w H:(int)h X:(float)x Y:(float)y L:(BOOL)lr R:(BOOL)rr
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d,%.1f,%.1f)\n",PREFUN,w,h,x,y);
#endif

  ArraySimulator *a = [[ArraySimulator alloc] initW:w H:h X:x Y:y];
  
  int c = (int)arrays.count;
  int b = c/2 + ((c % 2) ? N_CHANNELS/2 : 0); // high contrast b/w channels
  a.bias_level = 1000+10*b;
  a.efzm = 3.6e9;               // [e-]
  a.fwhm = 0.7;
  a.blur = 0.0;
  a.back = 14.0;                 // background J=14,H=12,K=11
  a.pixscale = pixscale;
  a.scat = scat;
  a.leftRef = lr;
  a.rightRef = rr;
  a.drift_a = 0.0;             // [DU] TODO on GUI this and other simulator parameters -- file?
  a.drift_f = 0.03;
  
  [arrays addObject:[a autorelease]];
}

/* ---------------------------------------------------------------- */

- (void)setEgain:(double)eg Enoise:(double)en
{
  for (ArraySimulator *a in arrays) {
    if (eg > 0) a.egain = eg;
    if (en > 0) a.enoise = en;
  }
}

/* ---------------------------------------------------------------- */

- (void)setAlpha:(double)alp delta:(double)del equinox:(double)equ
{
  for (ArraySimulator *a in arrays) {
    a.alpha   = alp;
    a.delta   = del;
    a.equinox = equ;
  }
}

/* ---------------------------------------------------------------- */

- (void)startExposure
{
  int q=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  for (ArraySimulator *a in arrays) { 
    a.shutter = _shutter;
    a.objects = _objects;
    a.disperser = _disperser;
  }

#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif
  for (ArraySimulator *a in arrays) {
    // printf("reset=%s\n",[a.name UTF8String]);
    dispatch_group_async(disp_group,disp_queue[q++ % N_THREADS],^(void) {
      [a reset];
    });
  }
#if (TIME_TEST > 0)
  dispatch_group_wait(disp_group,DISPATCH_TIME_FOREVER);
  fprintf(stderr,"%s: %.3f sec\n",PREFUN,YsecSince(date));
#endif
}

/* ---------------------------------------------------------------- */

- (void)updateDetector:(float)et
{
  int q=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.3f)\n",PREFUN,et);
#endif

  double phase = [MTwister doubleValue:2.0*M_PI];

  dispatch_group_wait(disp_group,DISPATCH_TIME_FOREVER);
#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif
  for (ArraySimulator *a in arrays) {
    // printf("name=%s\n",[a.name UTF8String]);
    dispatch_group_async(disp_group,disp_queue[q++ % N_THREADS],^(void) {
      [a update:et];
      [a readout:phase];
    });
  }
#if (TIME_TEST > 0)
  dispatch_group_wait(disp_group,DISPATCH_TIME_FOREVER);
  fprintf(stderr,"%s: %.2f ms\n",PREFUN,YmsSince(date));
#endif
}

/* ---------------------------------------------------------------- */

- (NSData*)fetchFrame
{
  u_short *p;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
#ifndef NDEBUG
  char buf[32];
#endif

  dispatch_group_wait(disp_group,DISPATCH_TIME_FOREVER);
 
  ArraySimulator *a = [arrays firstObject];
  assert(a);
  size_t nbytes = [arrays count]*a.dimy*a.dimx*sizeof(u_short);
  NSMutableData *data = [[NSMutableData alloc] initWithCapacity:nbytes];
  for (int y=0; y<a.dimy; y++) {   // stitch image from arrays
#ifndef NDEBUG
    int c = 0;
#endif
    for (ArraySimulator *a in arrays) {
#ifndef NDEBUG
      Ystr2chr(a.name,buf,sizeof(buf)); assert(buf[0] == 'A'+c); c++;
#endif
      p = a.data0 + y*a.dimx;
      [data appendBytes:p length:(a.dimx*sizeof(u_short))];
    }
  }
  assert([data length] == nbytes);
  
  return data;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
