/* ---------------------------------------------------------------- */
//
// Galil.m
// TeleGUI/Swope/Henrietta
//
// Created by Christoph Birk on 2013-07-17
// Copyright 2013 Carnegie Observatories. All rights reserved.
//
// Mavericks Arp Cache Validation (RFC1122-2.3.2.1) workaround
// sudo sysctl -w net.link.ether.inet.arp_unicast_lim=0
// /etc/sysctl.conf
//   net.link.ether.inet.arp_unicast_lim=0
//
// http:www.galil.com/download/comref/com30010/index.html
//
/* ---------------------------------------------------------------- */

#define DEBUG           0              // debug level
#define MAGIC_PRESET    0              // Simulator

enum galil_stop_codes { STOP_OK=1,STOP_REV=3,STOP_STOP=4,STOP_ABORT=7 };

/* ---------------------------------------------------------------- */

#import  "Galil.h"
#import  "Notice.h"
#import  "Yutils.h"
#import  "MTwister.h"

#if (DEBUG > 0)
extern Logger *main_logger;
#endif
extern BOOL use_shared;

static int magicPreset[9]={0,0,0,0,0,0,0,0,0}; // Simulator

/* ---------------------------------------------------------------- */

@implementation GalilConnection

Logger *galil_logger;

/* ---------------------------------------------------------------- */

+ (void)initialize
{
  if (self != [GalilConnection class]) return;  // avoid call by sub-class

  galil_logger = [[Logger alloc] initWithBase:"galil" useShared:use_shared];
#ifdef NDEBUG
  galil_logger.mode = LOGGER_OPEN;     // keep open, no auto-flush
#else
  galil_logger.mode = LOGGER_FLUSH;    // keep open, auto-flush
#endif
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)initWithHost:(const char*)hstr port:(int)pint online:(BOOL)on
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d,%d)\n",PREFUN,hstr,pint,on);
#endif

  if (!(self = [super initWithHost:hstr port:pint online:on])) return nil;
  
  // Note: The Galil controller terminates with '?' or ':'
  //       There might be <CR><LF> sequences before that
  terminator = '\0';                   // no terminator
  lock_timeout = 10;                   // Mavericks 'arp' fixed
  assert(lock_timeout > sock_timeout);
  
  return self;
}

/* ---------------------------------------------------------------- */

#if (DEBUG > 1)

- (int)open
{
  int err;
  
  @synchronized(self) {
    if (count == 0) { char buf[128];
      sprintf(buf,"%s (%d)",PREFUN,port); [galil_logger append:buf];
    }
    err = [super open];
  }
  return err;
}

/* --- */

- (void)close
{
  @synchronized(self) {
    [super close];
    if (count == 0) { char buf[128];
      sprintf(buf,"%s (%d)",PREFUN,port); [galil_logger append:buf];
    }
  }
}

#endif

/* ---------------------------------------------------------------- */

- (int)request:(const char*)cmd response:(char*)res max:(int)len
{
  int  err,i,nterms=1;
  char buf[GALIL_RESPONSE_SIZE+128],*p;
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s,%p,%d)\n",PREFUN,cmd,res,len);
#endif
  assert(len > 260);                   // 'ID' returns a long string

  sprintf(buf,"%s (%d) %s",PREFUN,port,cmd); [galil_logger append:buf];
  [galil_logger message:cmd];
  
  p = (char*)cmd; while ((p=strchr(p,';')) != NULL) { nterms++; p++; }
  
  err = [self open]; if (err) return err;

  if (![self mylock:lock_timeout]) {   // open/lock is more MT-efficient
    [self close];
    return E_tcpip_lock; 
  }

  if (online) {
    assert(sock > 0);
    sprintf(buf,"%s\r",cmd);           // append <CR>
    err = TCPIP_Send(sock,buf,strlen(buf));
    if (!err) {
      assert(res);
      for (i=0; i<len-1; i++) {
        err = TCPIP_Wait4Char(sock,&res[i],1000*sock_timeout); if (err) break;
        if      ((res[i] == ':') || (res[i] == '?')) nterms--;
        else if (res[i] == '\r') res[i] = ' ';
        else if (res[i] == '\n') res[i] = ' ';
        if (nterms == 0) { i++; break; }
      }
      res[i] = '\0';                   // terminate string
    }
  } else {
    assert(sock == 0);
    Ysleep(5);
  }
  [lock unlock];
  [self close];

  if (err) {
    sprintf(buf,"%s (%d) err=%d (%s)",PREFUN,port,err,cmd);
    [galil_logger append:buf];
    sprintf(buf,"err=%d",err); [galil_logger message:buf]; // hardhat message
  } else
  if (online) { int len=(int)strlen(res);
    for (i=0; i<len; i++) {            // write response to logfile
      if ((res[i] != ':')) {           // if there's anything but ':'
        sprintf(buf,"%s (%d) %s",PREFUN,port,res); [galil_logger append:buf];
        [galil_logger message:res];
        break;
      }
    }
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)request:(const char*)cmd response:(char*)res
{
  int err;
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s,%p)\n",PREFUN,cmd,res);
#endif
  err = [self request:cmd response:res max:GALIL_RESPONSE_SIZE];
  if (!err && online) { 
    if (strchr(res,'?')) err = E_galil_command;
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)command:(const char*)cmd
{
  int  err;
  char res[GALIL_RESPONSE_SIZE]="";  // Clang wants an initialization
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s)\n",PREFUN,cmd);
#endif

  err = [self request:cmd response:res];
  
  return err;
}

/* ---------------------------------------------------------------- */
/* instance methods */

- (int)errorMessage:(char*)string   
{
  int  err;
  char res[GALIL_RESPONSE_SIZE],*p;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,string);
#endif

  err = [self request:"TC1" response:res];
  if (err) return err;

  p = strchr(res,' '); if (!p) return E_galil_error;
  assert(strlen(res) < 120);
  sprintf(string,"Galil: %s",(char*)(p+1));
  p = string+strlen(string);
  while (--p > string) {
    if ((*p == ':') || (*p == '\r') || (*p == '\n')) {
      *p = '\0';
    } else {
      break;
    }
  }

  return 0;
}

/* ---------------------------------------------------------------- */

- (int)set_var:(const char*)key withInt:(int)value
{
  int  err;
  char cmd[128];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,key,value);
#endif
  sprintf(cmd,"%s=%d",key,value);
  err = [self command:cmd]; 
  
  return err;  
}

- (int)get_var:(const char*)key intPtr:(int*)value
{
  int  err;
  char cmd[128],res[GALIL_RESPONSE_SIZE];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s,%p)\n",PREFUN,key,value);
#endif
  
  sprintf(cmd,"%s=?",key);
  err = [self request:cmd response:res];
  if (!err && online) *value = atoi(res);
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)get_input:(int)bit intPtr:(int*)value
{
  int  err;
  char cmd[128],res[GALIL_RESPONSE_SIZE];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,bit,value);
#endif
  assert(bit > 0);

  sprintf(cmd,"MG @IN[%d]",bit);
  err = [self request:cmd response:res];
  if (!err && online) *value = atoi(res) ? 0 : 1; // inverted
  //printf("Ivalue(%d)=%d\n",bit,*value);
  
  return err;
}

/* --- */

- (int)get_output:(int)bit intPtr:(int*)value
{
  int  err;
  char cmd[128],res[GALIL_RESPONSE_SIZE];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d,%p)\n",PREFUN,bit,value);
#endif
  assert(bit > 0);

  sprintf(cmd,"MG @OUT[%d]",bit);
  err = [self request:cmd response:res];
  if (!err && online) *value = atoi(res);
  // printf("Ovalue(%d)=%d\n",bit,*value);
  
  return err;
}

/* --- */

- (int)set_output:(int)bit value:(int)value
{
  int  err;
  char cmd[128];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,bit,value);
#endif
  assert(bit > 0);

  if (value) sprintf(cmd,"SB %d",bit);
  else       sprintf(cmd,"CB %d",bit);
  err = [self command:cmd];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setupFromFile:(const char*)file  // Connection
{
  int  magic,err;
#if (DEBUG > 0)
  { char *p=strstr(file,".app"); fprintf(stderr,"%s(%s)\n",PREFUN,(p) ? p : file); }
#endif

  magic = (int)([MTwister hash32FromFile:file] >> 4);

  err = [self open];
  if (!err) {
    int value = magicPreset[0];         // Simulator preset
    // use 'magic' to only upload the settings once to the Galil controller
    err = [self get_var:"magic" intPtr:&value];
    if (!err) {
      printf("magic=%08x, value=%08x\n",magic,value); //xxx
      if (magic != value) {
        NSArray *setup = [NSArray arrayWithContentsOfFile:@(file)];
        if (setup) {
          for (NSString *line in setup) {
            // NSLog(@"%@",line);
            err = [self command:[line UTF8String]];
            if (err) break;
          }
          _fileUploaded = TRUE;
        } else { 
          err = E_galil_fnotf;
        }
        if (!err) [self set_var:"magic" withInt:magic];
#if (MAGIC_PRESET > 0)
        magicPreset[0] = magic;
#endif
      }
    }
    [self close];
  }
  printf("gConn: fileUploaded=%d\n",_fileUploaded);
 
  return err;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

@implementation GalilObject

@synthesize name,state,galil;

/* ---------------------------------------------------------------- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
  case E_galil_error: string = @"Galil: error"; break;
  case E_galil_command: string = @"Galil: invalid command"; break;
  case E_galil_fnotf: string = @"Galil: setup file not found"; break;
  case E_galil_forlim: string = @"Galil: FWD limit switch active"; break;
  case E_galil_revlim: string = @"Galil: REV limit switch active"; break;
  case E_galil_stop: string = @"Galil: axis received STOP command"; break;
  case E_galil_timeout: string = @"Galil: timeout on motion"; break;
  case E_galil_pnotf: string = @"Galil: position not found"; break;
  case E_galil_nohome: string = @"Galil: HOME not found"; break;
  case E_galil_abort: string = @"Galil: motion aborted"; break;
  case E_galil_nodetent: string = @"Galil: DETENT not found"; break;
  case E_galil_target: string = @"Galil: invalid wheel position"; break;
  default:
    if (!string) string = [TCPIP_Connection errorString:err];
    break;
  }
  return string;
}

/* ---------------------------------------------------------------- */

- (id)initWithConnection:(GalilConnection*)connection
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%p): self=%p\n",PREFUN,connection,self);
#endif

  if (!(self = [super init])) return nil;
  
  galil = [connection retain];
  
  return self;
}

/* ---------------------------------------------------------------- */

- (int)setup:(const char*)base
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s)\n",PREFUN,base);
#endif

  name = [@(base) retain];

  return 0;
}

/* ---------------------------------------------------------------- */

- (int)get
{
  assert(0);
  
  return 0;
}

/* ---------------------------------------------------------------- */

- (int)move:(int)pos
{
  assert(0);
  
  return 0;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  
  [galil release];
  [name release];
  [super dealloc];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

@interface GalilAxis ()                // private class extension
@property (readwrite) BOOL running,jogging;
@property (readwrite) int  limfor,limrev,homesw;
@end

@implementation GalilAxis

@synthesize acc,dec,motor_off; // TODO all as "modern" properties
@synthesize speed;
@synthesize limfor,limrev;
@synthesize axis;

/* ---------------------------------------------------------------- */

- (id)initWithConnection:(GalilConnection*)controller axis:(char)a
{
  char buf[32];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%c)\n",PREFUN,controller,a);
#endif

  self = [super initWithConnection:controller];
  if (!self) return nil;

  sprintf(buf,"Axis%c",a); [super setup:buf];  // set name
  axis = a;
  _velocity = 0.0f;
  _step_pos = 0;
  has_limits = YES;
  motor_off = NO;
  
  return self;
}

/* ---------------------------------------------------------------- */

- (int)set_par:(const char*)key withInt:(int)par
{
  int  err;
  char cmd[128];
  
  sprintf(cmd,"%s%c=%d",key,axis,par);
  err = [galil command:cmd]; 
  
  return err;  
}

/* --- */

- (int)get_par:(const char*)key intPtr:(int*)value
{
  int  err;
  char cmd[128],res[GALIL_RESPONSE_SIZE];
  
  sprintf(cmd,"MG_%s%c",key,axis);
  err = [galil request:cmd response:res];
  if (!err && galil.online) *value = atoi(res);
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setupFromGalil
{
  int  err,v=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p, axis=%c\n",PREFUN,self,axis);
#endif

  err = [galil open]; if (err) return err;

  if (galil.online) { char cmd[128],res[GALIL_RESPONSE_SIZE];
    // NOTE: default for YA=2, YB=200, YC=4000
    sprintf(cmd,"MG_YA%c;MG_YB%c",axis,axis);
    err = [galil request:cmd response:res];
    if (!err) { float a,b;
      sscanf(res,"%f : %f",&a,&b);  // printf("a=%f b=%f\n",a,b);
      _spt = (int)(a*b);
      assert(_spt != 0);
    }
  }
  // printf("stepsPerTurn=%d\n",_spt);
  
  if (!err) err = [self get_spdpar];
  _run_speed = speed;

  if (!err) err = [self get_pospar];

  if (!err) err = [self get_par:"LD" intPtr:&v];
  if (!err && galil.online) has_limits = (v) ? NO : YES;
  // printf("has_limits%c=%d\n",axis,(int)has_limits);
  if (!err) err = [self get_limits];

  if (!err) err = [self get_gain]; // todo always from .xml file xxxyyy
  // printf("gain%c=%d\n",axis,_gain);

  if (!err) err = [self get_par:"MO" intPtr:&v];
  // Note: "BN" after using a new controller
  // Warning: reading "MO" from the controller fails if the
  //          software is shutdown during a move
  if (!err && galil.online) motor_off = (v) ? YES : NO;
  // printf("motor_off%c=%d\n",axis,motor_off); // todo always from .xml file xxxyyy
  
  [galil close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setupFromFile:(const char*)path   // Axis
{
  int  err;
  char file[512];

  sprintf(file,"%s%c.xml",path,axis);
#if (DEBUG > 0)
  { char *p=strstr(file,".app"); fprintf(stderr,"%s(%s)\n",PREFUN,(p) ? p : file); }
#endif

  err = [galil open]; if (err) return err;
#if (DEBUG > 2)
  if (galil.online) { err = [galil request:"WH" response:cmd]; printf("%s\n",cmd); }
#endif

  if (!err) { int magic; char key[32];
    int value = magicPreset[1+axis-'A'];  // Simulator
    sprintf(key,"magic%c",axis);
    magic = (int)([MTwister hash32FromFile:file] >> 4);
    // use 'magic' to only upload the settings once to the Galil controller
    err = [galil get_var:key intPtr:&value];
    if (!err) {
      // printf("%s=%08x, value=%08x\n",key,magic,value); 
      if (magic != value) {            // upload from file 
        NSArray *setup = [NSArray arrayWithContentsOfFile:@(file)];
        if (setup) { int a=0,b=0;
          for (NSString *line in setup) { NSString *str=nil,*kst,*vst=@"";
            // NSLog(@"%@",line);
            NSArray *cmps = [line componentsSeparatedByString:@"="];
            kst = [cmps objectAtIndex:0];
            if        ([cmps count] == 2) {
              vst = [cmps objectAtIndex:1];
              str = [NSString stringWithFormat:@"%@%c=%@",kst,axis,vst];
            } else if ([cmps count] == 1) {
              str = [NSString stringWithFormat:@"%@%c",kst,axis];
            } else {
              assert(0);
            }
            if (!str) continue;
            // NSLog(@"%@",str);
            err = [galil command:[str UTF8String]]; if (err) break;
            if (!galil.online) {           // Simulator preset
              if        ([kst isEqualToString:@"SP"]) {
                speed = [vst intValue];
              } else if ([kst isEqualToString:@"AC"]) {
                acc   = [vst intValue];
              } else if ([kst isEqualToString:@"DC"]) {
                dec   = [vst intValue];
              } else if ([kst isEqualToString:@"YA"]) {
                a     = [vst intValue];
              } else if ([kst isEqualToString:@"YB"]) {
                b     = [vst intValue];
              } else if ([kst isEqualToString:@"LD"]) {
                has_limits = ([vst intValue] > 0) ? FALSE : TRUE;
              } else if ([kst isEqualToString:@"MO"]) {
                motor_off = TRUE;
              } else if ([kst isEqualToString:@"AG"]) {
                _gain = [vst intValue]; // printf("gain%c=%d\n",axis,_gain);
              }
            }
          }
          _fileUploaded = TRUE;
          if (!galil.online) {         // Simulator preset
            _spt = a*b;                // TODO needs to be in .xml file
          }
          if (!err) [galil set_var:key withInt:magic];
#if (MAGIC_PRESET > 0)
          magicPreset[1+axis-'A'] = magic;
#endif
        } else {
          err = E_galil_fnotf;        
        } // endif(setup)
      } // endif(magic==value)
    } // endif(err)
  } // endif(err)
//  printf("gAxis%c: fileUploaded=%d\n",axis,_fileUploaded); //xxx
  
  if (!err) err = [self setupFromGalil];
  
  [galil close];
  
  if (err) { NSString *str;
    str = [NSString stringWithFormat:@"%s\r%s\rfailed, err=%d\r%@",
            PREFUN,file,err,[[self class] errorString:err]];
    [Notice openNotice:[str UTF8String] title:__FILE__ time:60 level:2];
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)set_gain:(int)val
{
  int err;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,val);
#endif
  
  err = [self set_par:"AG" withInt:val];
  if (!err) _gain = val;
  
  return err;
}

/* --- */

- (int)get_gain
{
  int err,value=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  err = [self get_par:"AG" intPtr:&value];
  if (!err && galil.online) _gain = value;
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)set_speed:(int)val
{
  int  err;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,val);
#endif
  
  err = [self set_par:"SP" withInt:val]; 
  if (!err) _run_speed = speed = val;
  
  return err;  
}

/* --- */

- (int)get_speed
{
  int err,value=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  err = [self get_par:"SP" intPtr:&value];
  if (!err && galil.online) speed = value;
  
  return err;
}

/* --- */

- (int)get_velocity
{ 
  int  err;
  char cmd[128],res[GALIL_RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  sprintf(cmd,"TV%c",axis);
  err = [galil request:cmd response:res];
  if (!err && galil.online) {
    int value = atoi(res);
    if (abs(value) < 5) value = 0;  // Note: Galil bug
    _velocity = (float)value;
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)set_acceleration:(int)val
{
  int  err;
  
  err = [self set_par:"AC" withInt:val]; 
  if (!err) acc = val;
  
  return err;  
}

/* --- */

- (int)get_acceleration
{
  int err,value=acc;
  
  err = [self get_par:"AC" intPtr:&value];
  if (!err && galil.online) acc = value;
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)set_deceleration:(int)val
{
  int  err;
  
  err = [self set_par:"DC" withInt:val]; 
  if (!err) dec = val;
  
  return err;  
}

/* --- */

- (int)get_deceleration
{
  int err,value=dec;
  
  err = [self get_par:"DC" intPtr:&value];
  if (!err && galil.online) dec = value;
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)set_step_pos:(int)val
{
  int err = [self set_par:"DP" withInt:val];
  if (!err) _step_pos = val;
  
  return err;  
}

/* --- */

- (int)get_step_pos;
{
  char cmd[128],res[GALIL_RESPONSE_SIZE];
  
  sprintf(cmd,"TD%c",axis);
  int err = [galil request:cmd response:res];
  if (!err && galil.online) _step_pos = atoi(res);
  
  return err;  
}

/* ---------------------------------------------------------------- */

- (int)get_limits
{
  int err=0;
  
  if (has_limits) { char cmd[128],res[GALIL_RESPONSE_SIZE];
    sprintf(cmd,"MG_LF%c;MG_LR%c",axis,axis);
    err = [galil request:cmd response:res];
    if (!err && galil.online) { float f,r;
      sscanf(res,"%f : %f",&f,&r);
      self.limfor = (f) ? 0 : 1;         // KVO
      self.limrev = (r) ? 0 : 1;
    }
  } else { 
    limfor = limrev = 0; 
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)get_pospar
{
  int  err;
  char cmd[128],res[GALIL_RESPONSE_SIZE];
  
  sprintf(cmd,"TD%c;TV%c",axis,axis);
  err = [galil request:cmd response:res];
  if (!err && galil.online) { float d,v;
    sscanf(res,"%f : %f",&d,&v);
    if (fabs(v) < 5) v = 0;      // Note: Galil bug
    _step_pos = (int)d;
    _velocity = v;
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)get_spdpar
{
  int err;
  char cmd[128],res[GALIL_RESPONSE_SIZE];
  
  sprintf(cmd,"MG_SP%c;MG_AC%c;MG_DC%c",axis,axis,axis);
  err = [galil request:cmd response:res];
  if (!err && galil.online) { float s,a,d;
    sscanf(res,"%f : %f : %f",&s,&a,&d);
    speed = (int)s;  acc = (int)a;  dec = (int)d;
  }

  return err;
}

/* ---------------------------------------------------------------- */

- (float)est_time:(int)steps
{
  double stps = (double)abs(steps);
  double spd = sqrt((2.0*acc*dec*stps)/(acc+dec));
  if (spd > speed) spd = speed;
  double t1 = spd/acc;
  double s1 = 0.5*acc*t1*t1;
  double t2 = spd/dec;
  double s2 = 0.5*dec*t2*t2;
  
  return (float)(t1+t2+(stps-s1-s2)/spd);
}

/* --- */

- (int)moveSteps:(int)steps wait:(BOOL)wait
{
  char cmd[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%+d,%d): axis=%c (%d)\n",PREFUN,steps,wait,axis,motor_off);
#endif

  int err = [galil open]; if (err) return err;

  if (motor_off) err = [self motorPower:YES];

  if (!err) {
    sprintf(cmd,"PR%c=%d;BG%c",axis,steps,axis);
    err = [galil command:cmd];
  }
  if (!err) {
    self.running = TRUE;               // KVO
    if (!galil.online) {               // Simulator
      _step_pos += steps;
      _stop_code = STOP_OK;
    }
    if (wait) { const float tic=0.25f,toc=2.0f;
      float timeout = [self est_time:steps];
      while (timeout > 0.0f) {
        [NSThread sleepForTimeInterval:toc]; timeout -= toc;
        if ([self hasStopped]) break;
      }
      if (_running) {                  // +tic
        timeout = 2.0f;
        while (timeout > 0.0f) {
          [NSThread sleepForTimeInterval:tic]; timeout -= tic;
          if ([self hasStopped]) break;
        }
      }
      if (_running) {
        [self stop]; self.running = FALSE; err = E_galil_timeout;
      } else {
        switch (_stop_code) {
          case 1:  err = 0; break;
          case 2:  err = E_galil_forlim; break;
          case 3:  err = E_galil_revlim; break;
          case 4:  err = E_galil_stop; break;
          case 7:  err = E_galil_abort; break;
          default: err = E_galil_error; break;
        }
      }
    }
  }

  if (wait && motor_off) {
    if (!err) {
      float dt = 0.5f*(float)speed/(float)dec;
      [NSThread sleepForTimeInterval:dt];
      err = [self motorPower:NO];
    }
  }
  [galil close];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)motorPower:(BOOL)power
{
  int  err;
  char cmd[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,(int)power);
#endif
  
  err = [galil open]; if (err) return err;

  do {                       // try until motor has come to a full stop
    if (power) sprintf(cmd,"SH%c",axis);
    else       sprintf(cmd,"MO%c",axis);
    err = [galil command:cmd];
    if (err == E_galil_command) {
#if (DEBUG > 0)
      char buf[256]; (void)[galil errorMessage:buf]; printf("%s\n",buf);
#endif
      Ysleep(350);
    }
  } while (err == E_galil_command);
  [galil close];

  if (!err && power) Ysleep(50);
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)stop
{
  int  err;
  char cmd[128];
  
  _stop_code = STOP_STOP;              // Simulator preset
  
  sprintf(cmd,"ST%c",axis);
  err = [galil command:cmd];
  if (motor_off) [self motorPower:NO];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)jog:(int)spd                    // [steps/sec]
{
  int  err;
  char cmd[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,spd);
#endif
  
  err = [galil open]; if (err) return err;
  
  if (spd != 0) {                      // set 'jog' speed
    self.jogging = TRUE;               // KVO
    if (motor_off && (self.jog_speed==0)) { // no holding current
      err = [self motorPower:YES];
    }
    sprintf(cmd,"JG%c=%d;BG%c",axis,spd,axis);
    if (!err) err = [galil command:cmd];
    if (!galil.online) _stop_code = 0;
  } else {                             // stop jogging
    err = [self stop];
    if (!err) err = [self set_speed:self.run_speed]; // restore speed
    self.jogging = FALSE;
  }
  _jog_speed = spd;
    
  [galil close];
    
  return err;
}

/* ---------------------------------------------------------------- */

- (BOOL)hasStopped
{
  int  err;
  char cmd[128],res[GALIL_RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%c): stop_code=%d\n",PREFUN,axis,_stop_code);
#endif

  if (!galil.online) sprintf(res,"%d",_stop_code); // Simulator

  sprintf(cmd,"MG_SC%c",axis);
  err = [galil request:cmd response:res]; if (err) return FALSE;
  _stop_code = atoi(res);
  if (_stop_code != 0) {
    _running = _jogging = FALSE;
  }
  return (_running || _jogging) ? NO : YES;
}

/* ---------------------------------------------------------------- */

- (BOOL)isRunning
{
#if (DEBUG > 2)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  if (!galil.online) _stop_code = 1;    // Simulator preset

  if ([self hasStopped]) {
    self.running = _jogging = NO;
  }
  return _running;
}

/* ---------------------------------------------------------------- */

- (BOOL)isJogging
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  if (!galil.online) _stop_code = 1;    // Simulator preset

  if ([self hasStopped]) {
    self.jogging = _running = NO;
  }
  return _jogging;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation GalilSlide
{
  int steps;
  int posA,posB;
  int limfor,limrev;                   // used with Simulator
  BOOL motionThread;
}

/* ---------------------------------------------------------------- */

- (int)setup:(const char*)base fromPath:(const char*)path withAxis:(GalilAxis*)a
{
  int  err=0;
  char file[512],buf[128];
#if (DEBUG > 0)
  { char *p=strstr(path,".app"); fprintf(stderr,"%s(%s,%s,%p)\n",PREFUN,base,(p) ? p : path,a); }
#endif

  [super setup:base];                  // just sets name as (base)

  sprintf(file,"%s/%s.xml",path,base);
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:@(file)];
  assert(dict);

  _slideName = [[dict objectForKey:@"name"] retain];

  steps = [[dict objectForKey:@"steps"] intValue];
  // printf("steps=%d\n",steps);
  
  posA = [[dict objectForKey:@"A"] intValue];
  posB = [[dict objectForKey:@"B"] intValue];
  // printf("A=%d B=%d\n",posA,posB);

  if (a) {                       // used supplied axis
    _axis = [a retain];
    if (!galil) {                // take connection from the GalilAxis
      // printf("take galil connection from axis (%p)\n",_axis.galil);
      galil = _axis.galil;
      assert(galil);
    }
  } else {                       // create my own
    Ystr2chr([dict objectForKey:@"axis"],buf,sizeof(buf));
    _axis = [[GalilAxis alloc] initWithConnection:galil axis:buf[0]];
    sprintf(file,"%s/galil",path);
    err = [_axis setupFromFile:file];
  }
  if (!err) err = [self get];

  return err;
}


/* ---------------------------------------------------------------- */

- (int)get
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%c)\n",PREFUN,_axis.axis);
#endif

  int err = [galil open];
  if (!err) {
    if (!err) err = [_axis get_step_pos];
    if (!err) err = [_axis get_limits];
    [galil close];
  }

  if (galil.online) {
    limfor = _axis.limfor;
    limrev = _axis.limrev;
  } else {                          // Simulator
    limfor = (_axis.step_pos >= steps) ? 1 : 0;
    limrev = (_axis.step_pos <=  0)    ? 1 : 0;
    // printf("lim= %d %d\n",limrev,limfor);
  }
  if (!err) {
    if      (!limfor && limrev) state =  1;
    else if (!limrev && limfor) state =  0;
    else  {
      if (motionThread) state =  2;
      else              state = -1;
    }
  }
  return err;
}

/* ---------------------------------------------------------------- */

- (int)move:(int)pos
{
  const int wobbleState=2;                // hardcoded from GUI
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,name.UTF8String,pos);
#endif

  int err = [galil open]; if (err) return err;

  [self get];                             // update position
  if (pos != self.state) {
    if (pos == wobbleState) {             // 'Wobble'
      [_axis moveSteps:-steps wait:YES];  // move to NEG limit -- inserted
      [_axis set_step_pos:0];             // clear step count
      int old_gain = _axis.gain;
      [_axis set_gain:0];                 // reduce heat TEST@SBS
      [galil open];                       // keep open when closing below
      dispatch_async(GLOBAL_QUEUE,^(void) {
        motionThread = TRUE;
        state = wobbleState;
        while (self.state == wobbleState) {
          @autoreleasepool {
            [_axis moveSteps:(posA - _axis.step_pos) wait:YES];
            if (self.state != wobbleState) break;
            [_axis get_step_pos]; printf("pos=%d\n",_axis.step_pos); //xxx
            Ysleep(350); if (self.state != wobbleState) break;
            [_axis moveSteps:(posB - _axis.step_pos) wait:YES];
            if (self.state != wobbleState) break;
            [_axis get_step_pos]; printf("pos=%d\n",_axis.step_pos); //xxx
            Ysleep(350);
          }
        } // endwhile(Wobble)
        printf("gain=%d, restore to %d\n",_axis.gain,old_gain); //xxx
        [_axis set_gain:old_gain];      // restore gain
        [galil close];
        motionThread = FALSE;
      });
      Ysleep(350);
    } else {                             // move 'in' or 'out'
      state = -1; while (motionThread) { printf("."); Ysleep(50); } //xxx
      int s = (pos) ? -steps : steps;    // IN:neg  OUT:pos
      [_axis moveSteps:s wait:YES];      // todo use err but ignore limits
      err = [self get];
    } // endif(Wobble)
  } else {
    // printf("already in state=%d\n",self.state);
  } // endif(pos==state)
  if (!err) {
    if (self.state != pos) err = E_galil_pnotf;
  }
  [galil close];

  return err;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

@interface GalilWheel ()                // private class extension
@property (readwrite) BOOL homing;
@end

@implementation GalilWheel
{
  int   homePolarity;
  u_int getCounter;                   // debug only
}

/* ---------------------------------------------------------------- */

- (int)setup:(const char*)base fromPath:(const char*)path withAxis:(GalilAxis*)a
{
  int  err=0;
  char file[512],buf[128];
#if (DEBUG > 0)
  { char *p=strstr(path,".app"); fprintf(stderr,"%s(%s,%s,%p)\n",PREFUN,base,(p) ? p : path,a); }
#endif
  
  [super setup:base];                 // just sets name as (base)

  sprintf(file,"%s/%s.xml",path,base);
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:@(file)];
  assert(dict);

  _wheelName = [[dict objectForKey:@"name"] retain];

  assert(!_positionNames);
  _positionNames = [dict objectForKey:@"positions"];
  // for (NSString *s in _positionNames) NSLog(@"%@",s);
  [_positionNames retain];

  _nPositions = (int)[_positionNames count];
  // printf("n_postitions=%d\n",_nPositions);
  assert(_nPositions > 0);

  _stepsPerPos = [[dict objectForKey:@"steps_per_pos"] intValue];
  // printf("steps_per_pos=%d\n",stepsPerPos);
  assert(_stepsPerPos != 0);
  
  homePolarity = [[dict objectForKey:@"home_polarity"] intValue];
  // printf("home_polarity=%d\n",homePolarity);
  assert((homePolarity==0) || (homePolarity==1));

  if (a) {                            // use supplied axis
    _axis = [a retain];
    if (!galil) {                     // take connection from GalilAxis
      // printf("take galil connection from axis (%p)\n",_axis.galil);
      galil = _axis.galil;
      assert(galil);
    }
  } else {                            // create my own axis
    Ystr2chr([dict objectForKey:@"axis"],buf,sizeof(buf));
    _axis = [[GalilAxis alloc] initWithConnection:galil axis:buf[0]];
    sprintf(file,"%s/galil",path);
    err = [_axis setupFromFile:file];
  }
  if (!err) err = [self get];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  [_axis release];
  [_positionNames release];

  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (int)get                             // inherited from super class
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%c): spp=%d\n",PREFUN,_axis.axis,_stepsPerPos);
#endif
  assert(_nPositions);
  assert(_stepsPerPos);
  
  int err = [galil open];
  if (!err) {
    if (!err) err = [_axis get_step_pos];
    if (!err) err = [self get_switches];
    [galil close];
  }
  if (err) {
    state = -1;
  } else {
    if ((abs(_axis.step_pos) % abs(_stepsPerPos)) != 0) {
      state = -1;                      // invalid step position
    } else {
      if (!self.detent) {              // no detent switch
        state = -1;
      } else {
        state = (abs(_axis.step_pos) / abs(_stepsPerPos)) % _nPositions;
        if ((state == 0) && (!_homesw)) state = -1; // require home at '0'
      }
    }
  }
  getCounter++; // printf("state=%d (%u)\n",state,getCounter);
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)get_switches
{
  char cmd[128],res[GALIL_RESPONSE_SIZE];
  
  sprintf(cmd,"MG_TS%c",self.axis.axis);
  int err = [galil request:cmd response:res];
  if (!err) {
    if (galil.online) { float v; int i;
      sscanf(res,"%f",&v); i = (int)v;
      _detent = (i & 0x0c) ? 0 : 1;
      _homesw = (i & 0x02) ? homePolarity : 1-homePolarity;
    } else {                            // Simulator
      _detent = ((abs(_axis.step_pos) % abs(_stepsPerPos)) == 0) ? 1 : 0;
      int spr = abs(_stepsPerPos) * _nPositions;
      _homesw = (_detent && ((_axis.step_pos % spr) == 0)) ? 1 : 0;
    }
  }
  // printf("%s(%c): D=%d H=%d\n",PREFUN,_axis.axis,_detent,_homesw);
  
  return err;
}


/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)move:(int)target
{
  char buf[128];
  sprintf(buf,"%s(%c,%d)",PREFUN,_axis.axis,target);
#if (DEBUG > 0)
  fprintf(stderr,"%s\n",buf);
#endif
  [main_logger append:buf];

  if ((target < 0) || (target >= _nPositions)) return E_galil_target;
  
  int err = [galil open]; if (err) return err;

  err = [self get];
  if (!err) {
    if (self.state < 0) err = E_galil_nodetent; // TODO _invpos
  }
  if (!err) {
    while (state != target) {
      err = [_axis moveSteps:_stepsPerPos wait:YES];  // powers up the motor
      if (err) break;
      Ysleep(2*350); [self get];       // wait for detent
      if (self.state < 0) { err=E_galil_pnotf; break; }
    }
  }
  [galil close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)find_home
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p: name=%s\n",PREFUN,self,[name UTF8String]);
#endif

  int err = [galil open]; if (err) return err;
  
  _foundHome = FALSE;
  _homing = TRUE;

  err = [self find_detent];           // motor off

  if (!err) {
    for (int i=0; i<_nPositions; i++) {
      if (self.homesw) break;
      err = [_axis moveSteps:_stepsPerPos wait:YES];  // powers up the motor
      if (err) break;
      Ysleep(2*350); [self get_switches];
    }
  }
  if (!err) {
    if (self.homesw) {                // at home
      [_axis set_step_pos:0];         // step_count=0
      [self get];                     // update state
      if (state == 0) _foundHome = TRUE;
    }
  }
  if (_axis.motor_off) [_axis motorPower:NO]; // bug-fix b0087
  [galil close];
  self.homing = FALSE;

  return err;
}

/* ---------------------------------------------------------------- */

- (int)find_detent
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%c)\n",PREFUN,_axis.axis);
#endif

  int err = [galil open]; if (err) return err;
  
  int save = _axis.speed;
  [_axis set_speed:save/5];           // works for all 4 wheels in parallel
  
  if (galil.online) {
    [_axis moveSteps:_stepsPerPos wait:NO];  // powers up the motor
    do {
      [self get_switches];
      if (self.detent) { [_axis stop]; Ysleep(350); } // powers down motor
      Ysleep(50);
    } while (![_axis hasStopped]);
  } else {                            // Simulator
    int step = _axis.step_pos;
    Ysleep(1000);
    do {
      step += _stepsPerPos/abs(_stepsPerPos);
    } while (step % abs(_stepsPerPos));
    [_axis set_step_pos:step]; 
  }
  [_axis set_speed:save];

  if (_axis.motor_off) [_axis motorPower:NO];
  if (!err) {      
    Ysleep(2*350); [self get_switches];
    if (!_detent) err = E_galil_nodetent;
  }
  [galil close];
  
  return err;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
