/* ---------------------------------------------------------------- */
//
// GraphController.m
// Henrietta
//
// Created by Christoph Birk on 2014-03-21
//
/* ---------------------------------------------------------------- */

#define DEBUG           1     // debug level TODO common with FourStar/PFS

/* ---------------------------------------------------------------- */

#import  "GraphController.h"
#import  "Yutils.h"

#if (DEBUG > 0)
#import  "Logger.h"
extern Logger *main_logger;
#endif

#include "utils.h"

/* ---------------------------------------------------------------- */

#define TMAX  0xff00ff00ff             // sizeof(time_t)==8
#define MNS   (60)
#define HRS   (60*MNS)

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@interface GraphDataPoint ()
@property (readwrite,nonatomic) time_t time;
@property (readwrite,nonatomic) float  value;
@end

@implementation GraphDataPoint
@synthesize time,value;

+ (id)pointWithTime:(time_t)t value:(float)v
{
  GraphDataPoint *point = [[GraphDataPoint alloc] initWithTime:t value:v];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%ld,%f) -> %ld,%f\n",PREFUN,t,v,point.time,point.value);
#endif
  return [point autorelease];
}

/* --- */

- (id)initWithTime:(time_t)t value:(float)v
{
  if ((self = [super init]) != nil) {
    time = t;
    value = v;
  }
  return self;
}
@end

/* ---------------------------------------------------------------- */
/* GraphController */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation GraphController

@synthesize view_graph;

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

#ifndef NDEBUG
+ (void)initialize
{
  if (self != [GraphController class]) return;  // prevent sub-classes
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
}
#endif

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */

- (id)initWithWindowNib:(NSString*)name context:(int)myid
                                        title:(const char*)tstr
                                        unit:(const char*)ustr
                                        N:(int)nc
                                        colors:(const GraphColor*)cols
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%s,%s,%d,%p)\n",PREFUN,myid,tstr,ustr,nc,cols);
#endif

  if ((self = [super initWithWindowNibName:name]) == nil) return nil;

  time0 = time(NULL);

  YwindowTitle([self window],tstr);    // forces lazy window
  key_frame = [[NSString stringWithFormat:@"Graphwindow_%s_%d",tstr,myid] retain];
  [[self window] setFrameUsingName:key_frame];
  [panel_gpars setTitle:[NSString stringWithFormat:@"Graph Parameters (%s)",tstr]];
 
  NSMutableArray *array = [NSMutableArray arrayWithCapacity:nc];
  for (int i=0; i<nc; i++) {
    NSMutableArray *data = [[NSMutableArray alloc] initWithCapacity:128];
    [array addObject:data];
    [data release];
  }
  assert([array count] == nc);
  dataArrays = [[NSArray alloc] initWithArray:array];
  assert(view_graph);
  [view_graph setDataArray:dataArrays colors:cols];
  view_graph.unit = @(ustr);

  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
  [dataArrays release];
  [key_frame release];
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (void)windowWillClose:(NSNotification*)notification
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,notification);
#endif
  YwindowFrameSave(self.window,[key_frame UTF8String]);
}

/* ---------------------------------------------------------------- */

- (void)addData:(int)c time:(time_t)t value:(float)v redraw:(BOOL)r
{
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d,%ld,%.3f): %p\n",PREFUN,c,t,v,self);
#endif
  assert(view_graph);
  
  NSMutableArray *array = [dataArrays objectAtIndex:c];
  @synchronized (array) {
    [array addObject:[GraphDataPoint pointWithTime:t value:v]];
  }
  if (r) {
    dispatch_async(MAIN_QUEUE,^void() {
      if ([[self window] isVisible]) view_graph.needsDisplay = YES;
    });
  }
}

/* -- */

- (void)addData:(int)c time:(time_t)t value:(float)v
{
  [self addData:c time:t value:v redraw:YES];
}

/* -- */

- (void)addData:(int)c value:(float)v redraw:(BOOL)r
{
  [self addData:c time:time(NULL) value:v redraw:r];
}

/* -- */

- (void)addData:(int)c value:(float)v
{
  [self addData:c time:time(NULL) value:v redraw:YES];
}

/* ---------------------------------------------------------------- */

- (void)thinData:(double)threshold after:(double)maxage
{
  int    i,keep;
  GraphDataPoint *point,*p1,*p2;
  double age,dt;
  time_t tmin=TMAX;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%f,%f): self=%p\n",PREFUN,threshold,maxage,self);
#endif
  assert(view_graph);
  assert(sizeof(tmin) == 8);        // sizeof(time_t)==8

  time_t ut = time(NULL);
  assert(ut < tmin);
  
  for (NSMutableArray *array in dataArrays) {
    keep = (maxage > 0) ? 4 : 0; // todo re-think this
    if ([array count] <= keep) continue;
    @synchronized(array) { // remove all data older than 'maxage' [hrs]
      NSMutableArray *rem = [NSMutableArray arrayWithCapacity:32];
      keep = (maxage > 0) ? 2 : 0;
      for (i=0; i<(long)[array count]-keep; i++) { // don't touch newest
        point = (GraphDataPoint*)[array objectAtIndex:i];
        age = (double)(ut - point.time)/3600.0; // [hrs]
        if (age < maxage) break;           // too young
        [rem addObject:point];
        if (point.time < tmin) tmin = point.time;
      }
      // remove data points closer than 'dt'
      for ( i+=1 ; i<(long)[array count]-2; i++) {    // don't touch newest
        p1 =    (GraphDataPoint*)[array objectAtIndex:(i-1)];
        point = (GraphDataPoint*)[array objectAtIndex:i];
        p2 =    (GraphDataPoint*)[array objectAtIndex:(i+1)];
        dt = threshold*log10(ut-point.time);
        if (((p2.time - p1.time) > dt)) continue;
#if (DEBUG > 0) //xxx1
        printf("dt=%.1f, p2-p1=%ld\n",dt,p2.time-p1.time);
#endif
        [rem addObject:point]; i++;
      } // endfor(i)
      [array removeObjectsInArray:rem];
#if (DEBUG > 0)
      NSString *str = [NSString stringWithFormat:@"%s(%p,%@): count=%lu, removed=%lu",
        PREFUN,array,view_graph.unit,[array count],[rem count]];
      [main_logger append:[str UTF8String]];
#endif
    } // end(synchronized)
  } // endfor(data arrays)

  if (tmin != TMAX) {
    if (tmin > time0) time0 = tmin;  // new zero time
  }
#ifndef NDEBUG // todo why
  dispatch_async(MAIN_QUEUE,^(void) {
    if ([[self window] isVisible]) YviewNeedsDisplay(view_graph);
  });
#endif
}

/* ---------------------------------------------------------------- */

- (void)setChannel:(int)c monitor:(BOOL)m
{
  assert(view_graph);
  [view_graph setChannel:c monitor:m];
  if ([[self window] isVisible]) YviewNeedsDisplay(view_graph);
}

/* ---------------------------------------------------------------- */

- (void)updateTscale
{
  char  buf[128];
  float tmin,tmax,tlst;
  
  assert([NSThread isMainThread]);

  [popup_tmode selectItemAtIndex:view_graph.scale_tmode];
  
  tmin = (float)(view_graph.scale_tmin-time0)/3600.0f;
  tmax = (float)(view_graph.scale_tmax-time0)/3600.0f;
  tlst = (float)(view_graph.scale_tlst)/3600.0f;
  
  switch (view_graph.scale_tmode) {   // 'IB' dependency
  case GRAPH_SCALE_MAN:
  case GRAPH_SCALE_FULL:
    sprintf(buf,"%.2f",tmin); YeditSet(edit_tmin,buf);
    sprintf(buf,"%.2f",tmax); YeditSet(edit_tmax,buf);
    [edit_tmin setHidden:NO];
    [edit_tmin setEnabled:!view_graph.scale_tmode];
    [edit_tmax setEnabled:!view_graph.scale_tmode];
    break;
  case GRAPH_SCALE_LAST:
    sprintf(buf,"%.2f",tlst); YeditSet(edit_tmax,buf);
    [edit_tmin setHidden:YES];
    [edit_tmax setEnabled:YES];
    break;
  }
}

/* ---------------------------------------------------------------- */

- (void)updateVscale
{
  char  buf[128];
  
  assert([NSThread isMainThread]);

  [popup_vmode selectItemAtIndex:view_graph.scale_vmode];

  sprintf(buf,"%.1f",view_graph.scale_vmin); YeditSet(edit_vmin,buf);
  sprintf(buf,"%.1f",view_graph.scale_vmax); YeditSet(edit_vmax,buf);
  [edit_vmin setEnabled:!view_graph.scale_vmode];
  [edit_vmax setEnabled:!view_graph.scale_vmode];
}

/* ---------------------------------------------------------------- */

-(IBAction)action_gpars:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);
  
  if (sender == popup_tmode) {
    view_graph.scale_tmode = (int)[popup_tmode indexOfSelectedItem];
    [self updateTscale];
  } else
  if (sender == popup_vmode) {
    view_graph.scale_vmode = (int)[popup_vmode indexOfSelectedItem];
    [self updateVscale];
  } else
  if (sender == edit_tmin) {
    assert(view_graph.scale_tmode == GRAPH_SCALE_MAN);
    view_graph.scale_tmin = time0+(long)([edit_tmin floatValue]*3600);
  } else
  if (sender == edit_tmax) {
    assert(view_graph.scale_tmode != GRAPH_SCALE_FULL);
    if (view_graph.scale_tmode == GRAPH_SCALE_MAN) {
      view_graph.scale_tmax = time0+(long)([edit_tmax floatValue]*3600);
    } else {
      view_graph.scale_tlst = (long)([edit_tmax floatValue]*3600);
      if (view_graph.scale_tlst < 60) view_graph.scale_tlst = 60;
    }
  } else
  if (sender == edit_vmin) {
    assert(view_graph.scale_vmode == GRAPH_SCALE_MAN);
    view_graph.scale_vmin = [edit_vmin floatValue];
  } else
  if (sender == edit_vmax) {
    assert(view_graph.scale_vmode == GRAPH_SCALE_MAN);
    view_graph.scale_vmax = [edit_vmax floatValue];
  } else
  if (sender == but_apply) {
    [panel_gpars orderOut:self];
    if (view_graph.scale_tmode == GRAPH_SCALE_MAN) {
      view_graph.scale_tmin = time0+(long)([edit_tmin floatValue]*3600);
      view_graph.scale_tmax = time0+(long)([edit_tmax floatValue]*3600);
    } else
    if (view_graph.scale_tmode == GRAPH_SCALE_LAST) {
      view_graph.scale_tlst = (long)([edit_tmax floatValue]*3600);
      if (view_graph.scale_tlst < 60) view_graph.scale_tlst = 60;
    }
    if (view_graph.scale_vmode == GRAPH_SCALE_MAN) {
      view_graph.scale_vmin = [edit_vmin floatValue];
      view_graph.scale_vmax = [edit_vmax floatValue];
    }
  } else
  if (strstr([[sender title] UTF8String],"Pars")) {
    if (![panel_gpars isVisible]) {
      [self updateTscale];
      [self updateVscale];
    }
    [panel_gpars makeKeyAndOrderFront:self];
  } else {
    assert(0);
  }
  if (view_graph.scale_tmin > view_graph.scale_tmax) { time_t h;
    h = view_graph.scale_tmin;
    view_graph.scale_tmin = view_graph.scale_tmax;
    view_graph.scale_tmax = h;
  }
  if (view_graph.scale_vmin > view_graph.scale_vmax) { float h;
    h = view_graph.scale_vmin;
    view_graph.scale_vmin = view_graph.scale_vmax;
    view_graph.scale_vmax = h;
  }
  YviewNeedsDisplay(view_graph);
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* GraphView */
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

@implementation GraphView

@synthesize scale_tmode,scale_tmin,scale_tmax,scale_tlst;
@synthesize scale_vmode,scale_vmin,scale_vmax;
@synthesize unit;

/* ---------------------------------------------------------------- */

- (id)initWithFrame:(NSRect)frame
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.0f,%.0f)\n",PREFUN,frame.size.width,frame.size.height);
#endif

  if ((self = [super initWithFrame:frame]) == nil) return nil;

  scale_tmode = GRAPH_SCALE_FULL;
  scale_vmode = GRAPH_SCALE_FULL;
  scale_tlst  = 3600;                  // 1 [hrs]
  
  font_size = 14.0;
  font = CtextGetFont("Helvetica",font_size);

  return self;
}

/* ---------------------------------------------------------------- */

- (void)dealloc
{
  [dataArrays release];
  if (monitor) free((void*)monitor);
  if (colors) free((void*)colors);
  [unit release];
  [super dealloc];
}

/* ---------------------------------------------------------------- */

- (void)setDataArray:(NSArray*)array colors:(const GraphColor*)cols
{
  dataArrays = [array retain];
  monitor = (BOOL*)malloc([array count]*sizeof(BOOL));
  colors = (GraphColor*)malloc([array count]*sizeof(GraphColor));
  for (int i=0; i<[array count]; i++) {
    monitor[i] = TRUE;
    colors[i] = cols[i];
  }
}

/* ---------------------------------------------------------------- */

- (void)setChannel:(int)c monitor:(BOOL)m
{
  assert(c < [dataArrays count]);
  monitor[c] = m;
}

/* ---------------------------------------------------------------- */

static char* graph_timestr(char* str,time_t ut,const char* fmt)
{
  struct tm *ltime,res;
  
  ltime = localtime_r(&ut,&res);
  strftime(str,32,fmt,ltime); 
#if (DEBUG > 3)
  fprintf(stderr,"%s(%ld,'%s')=%s\n",PREFUN,ut,fmt,str);
#endif
  return str;
}

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)drawRect:(NSRect)rect
{
  int           i,xtics=1,ytics=1,itic,move_flag,nc;
  float         mtic,xtic,ytic,vmin,vmax,vdyn;
  double        t,v,x,y=0,tdyn,lmargin=0;
  char          buf[128];
  time_t        ut,md,tmin=0,tmax=0;
  BOOL          dots_only=TRUE;
  const int     ttic[]={MNS,5*MNS,10*MNS,30*MNS,HRS,2*HRS,6*HRS,24*HRS,0};
  const float   vtic[]={1,2,5};
  const CGFloat rmargin=10,bmargin=18,tmargin=18;
  const CGFloat dashes[]={10,5};
#if (DEBUG > 1)
  fprintf(stderr,"%s([%.0f,%.0f]): %p\n",PREFUN,rect.size.width,rect.size.height,self);
#endif
  assert([NSThread isMainThread]);     // NSView is mainThread only

  if (!dataArrays) return;
  nc = (int)[dataArrays count];
  
  if (scale_tmode) {                   // 'full' or 'last' scaling
    scale_tmin = 0x7fffffff; scale_tmax = 0;
    for (i=0; i<nc; i++) {
      if (!monitor[i]) continue;
      NSArray *array = [dataArrays objectAtIndex:i];
      @synchronized (array) {
        if ([array count] > 0) {
          tmin = [(GraphDataPoint*)[array firstObject] time];
          tmax = [(GraphDataPoint*)[array lastObject] time];
        } else {
          tmin = scale_tmin; tmax = scale_tmax;
        }
      }
      scale_tmin = (long)fmin(scale_tmin,tmin);
      scale_tmax = (long)fmax(scale_tmax,tmax);
    } // endfor(nc)
    if (scale_tmode==2) scale_tmin = (long)fmax(scale_tmin,scale_tmax-scale_tlst);
  }
#if (DEBUG > 1)
  printf("time_min=%ld, time_max=%ld (%ld)\n",scale_tmin,scale_tmax,scale_tmax-scale_tmin);
#endif
  if (scale_tmin >= scale_tmax) return;   // not enough data
  
  if (scale_vmode) {                   // 'full' scaling
    scale_vmin = 9e9; scale_vmax = -9e9;
    for (i=0; i<nc; i++) {
      if (!monitor[i]) continue;
      NSArray *array = [dataArrays objectAtIndex:i];
      @synchronized (array) {
        for (GraphDataPoint *point in array) {
          if ([point time] < scale_tmin) continue;
          if ([point time] > scale_tmax) continue;
          scale_vmin = (float)fmin(scale_vmin,[point value]);
          scale_vmax = (float)fmax(scale_vmax,[point value]);
        }
      }
    }
    if (scale_vmax < 1.0) {
      if (scale_vmax-scale_vmin < 0.01f) scale_vmax += 0.01f;
    } else
    if (scale_vmax < 10.0) {
      if (scale_vmax-scale_vmin < 0.1f) scale_vmax += 0.1f;
    } else {
      if (scale_vmax-scale_vmin < 1.0f) scale_vmax += 1.0f; 
    }
  }
#if (DEBUG > 1)
  printf("data_min=%.1f, data_max=%.1f\n",scale_vmin,scale_vmax);
#endif

  CGContextRef context = [[NSGraphicsContext currentContext] CGContext];

  float yb = bmargin+self.bounds.origin.y;
  float yt = self.bounds.origin.y+self.bounds.size.height-tmargin;
  float h = yt-yb; if (h < 1) return;

  itic=0; mtic=0.1f;                   // design data (y) axis
  mtic = (scale_vmin < 10.0) ? 0.1f : 1.0f;
  do {
    ytic = vtic[itic]*mtic;
    vmin = floor(scale_vmin/ytic)*ytic;
    vmax = floor(scale_vmax/ytic+0.99)*ytic;
    if (vmax == vmin) {
      vmax = floor(scale_vmax/ytic+1.00)*ytic;
    }
    if (ytic*ytics >= vmax-vmin) break;
    ytics++;
    if (ytics*30 > h) {
      ytics=1; itic++;
      if (itic == (sizeof(vtic)/sizeof(float))) { itic=0; mtic *= 10.0; }
    }
  } while (1);
  vdyn = vmax - vmin;
#if (DEBUG > 1)
  printf("vmin=%f, vmax=%f, vdyn=%f\n",vmin,vmax,vdyn);
#endif
  assert(vdyn > 0.0f);

  for (i=0; i<=ytics; i++) {           // get required lmargin
    v = vmin + i*ytic;
    if (v < 10.0) sprintf(buf,"%.1f",v);
    else          sprintf(buf,"%.0f",v);
    v = 10 + 0.45*font_size*strlen(buf); if (v > lmargin) lmargin = v;
  }
  float xl = lmargin+self.bounds.origin.x;
  float xr = self.bounds.origin.x+self.bounds.size.width-rmargin;
  float w  = xr-xl; if (w < 1) return;

  itic=0; mtic=1.0f;                   // design time axis
  do {
    xtic = mtic*ttic[itic];
    md = (ttic[itic] < 24*HRS) ? (long)(xtic) : HRS;
    tmin = scale_tmin - (scale_tmin % md);       // round down
    tmax = scale_tmax - (scale_tmax % md) + md;  // round up
    if (ttic[itic] >= 24*HRS) {        // find start/end of local day
      while (atof(graph_timestr(buf,tmin,"%H.%M")) != 0.0) tmin -= HRS;
      // while (atof(graph_timestr(buf,tmax,"%H.%M")) != 0.0) tmax += HRS;
    }
    if (xtic*xtics >= tmax-tmin) break;
    xtics++;
    if (xtics*60 > w) { xtics=1; itic++; if (ttic[itic]==0) { itic=7; mtic *= 2.0; } }
  } while (1);
  if (ttic[itic] >= 24*HRS) tmax = tmin + xtics*xtic;
  tdyn = (double)(tmax - tmin);
  assert(tdyn > 0.0f);

  CGContextSetRGBStrokeColor(context,0,0,0,1);   // black
  CGContextSetLineWidth(context,2);              // draw frame
  CGContextStrokeRect(context,CGRectMake(xl,yb,w,h));

  CGContextSetTextMatrix(context,CGAffineTransformIdentity);

  CGContextSetLineWidth(context,1);    // draw tics
  CGContextSetLineDash(context,0.0,dashes,sizeof(dashes)/sizeof(CGFloat));
  CGContextBeginPath(context);

  for (i=0; i<=xtics; i++) {           // draw time (x) legend
    t = i*xtic; x = xl + w*t/tdyn;
    ut = (long)my_round(t+tmin,0);
    (void)graph_timestr(buf,ut,(xtic < 24*HRS) ? "%H:%M" : "%b/%d");
    CtextCenterText(buf,font,context,x-((i==xtics) ? 0.6*font_size : 0),yb-0.6*font_size);
    if ((i==0) || (i == xtics)) continue;
    CGContextMoveToPoint(context,x,yb);
    CGContextAddLineToPoint(context,x,yt);
  }

  for (i=0; i<=ytics; i++) {           // draw data (y) legend
    v = i*ytic; y = yb + h*v/vdyn;
    v += vmin;
    if (v < 10.0) sprintf(buf,"%.1f",v);
    else          sprintf(buf,"%.0f",v);
    CtextRightText(buf,font,context,xl-0.4*font_size,y+0.25*font_size);
    if ((i == 0) || (i == ytics)) continue;
    CGContextMoveToPoint(context,xl,y);
    CGContextAddLineToPoint(context,xr,y);
  }
  if ([unit compare:@""] != NSOrderedSame) {
    sprintf(buf,"[%s]",[unit UTF8String]);
    CtextRightText(buf,font,context,xl-0.6*font_size,y-1.0*font_size);
  }
  CGContextDrawPath(context,kCGPathStroke);  // draw dashed lines

  /* draw sensor data --------------------------------------------- */

  CGContextSetLineDash(context,0.0,NULL,0);  // solid line
  CGContextSetLineWidth(context,(self.bounds.size.height > 260) ? 2 : 1);
  CGContextClipToRect(context,CGRectMake(xl+1,yb+1,w-2,h-2));
  for (i=0; i<nc; i++) {               // loop over channels
    if (!monitor[i]) continue;
    CGContextSetRGBStrokeColor(context,colors[i].r,colors[i].g,colors[i].b,1);
    CGContextSetRGBFillColor(context,colors[i].r,colors[i].g,colors[i].b,1);
    if (!dots_only) CGContextBeginPath(context);
    move_flag = 1;  
    NSArray *array = [dataArrays objectAtIndex:i];
    @synchronized (array) {
      for (GraphDataPoint *point in array) {
        t = point.time;
        if ((t < tmin) || (t > tmax)) { move_flag=1; continue; }
        v = point.value;
        x = xl + w*(t-tmin)/tdyn;
        y = yb + h*(v-vmin)/vdyn;
        if (dots_only) {
          CGContextBeginPath(context);
          CGContextAddArc(context,x,y,3,0,2*3.14159,0);
          CGContextFillPath(context);
        } else {
          if (move_flag) CGContextMoveToPoint(context,x,y);
          else           CGContextAddLineToPoint(context,x,y);
          move_flag = 0;
        }
      }
    }
    if (!dots_only) CGContextDrawPath(context,kCGPathStroke);  // draw line
  }
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
