/* ---------------------------------------------------------------- */
//
// MagView.m
// M2FS/Swope/PFS
//
// Created by Christoph Birk on 2012-03-12
// Copyright 2012 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG           1              // debug level
#endif
#define TIME_TEST       0

/* ---------------------------------------------------------------- */

#import "MagView.h"
#import "QltoolController.h"

#import "Yutils.h"                     // Cocoa wrappers

/* ---------------------------------------------------------------- */

static const u_int outcol=0x0000a000;  // 0bGr

static void set_col(u_int* data,int x0,int y0,int w,int m,u_int c)
{
  int x,y,x1=x0+m,y1=y0+m;
  
  for (y=y0; y<y1; y++) {
    for (x=x0; x<x1; x++) {
      data[x+y*w] = c;
    }
  }
}

/* ---------------------------------------------------------------- */

@implementation MagView

@synthesize controller;
@synthesize mag_x,mag_y;
@synthesize radius,mag;

/* ---------------------------------------------------------------- */

- (void)awakeFromNib
{
  u_int *idata,*d;
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  
  safeBounds = [self bounds];
  int w = (int)self.bounds.size.width;
  int h = (int)self.bounds.size.height;
  d = idata = (u_int*)malloc(w*h*sizeof(u_int));
  for (int j=0; j<w*h; j++) *d++ = 0x00f0a0a0; // 0BGR
  YimageData(self,idata,w,h);
  free((void*)idata);
}

/* ---------------------------------------------------------------- */

- (void)drawRect:(NSRect)rect
{
#if (DEBUG > 1)
  fprintf(stderr,"%s([%.0f,%.0f,%d])\n",PREFUN,rect.size.width,rect.size.height,radius);
#endif
  assert([NSThread isMainThread]);     // NSView is mainThread only
  
  [super drawRect:rect];               // draw image

  safeBounds = [self bounds];  

  if (radius > 0) { float x,y;         // draw center pixel and radius
    CGContextRef context = [[NSGraphicsContext currentContext] CGContext];
    x = (int)(rect.size.width/2.0); y = (int)(rect.size.height/2.0);
    CGContextSetRGBStrokeColor(context,0.0f,1.0f,0.0f,1);
    CGContextSetLineWidth(context,1);
    CGContextAddArc(context,x,y,radius,0,2*3.14159,0);
    CGContextStrokePath(context);
    CGContextStrokeRect(context,CGRectMake(x-mag/2-1,y-mag/2-1,mag+1,mag+1));
  }
}

/* ---------------------------------------------------------------- */

- (void)newData:(int*)data min:(int)mi dyn:(int)dy 
{
  int    raw_w,raw_h,mag_w,mag_h,img_w,img_h,x,y,xx,yy,colour,satlev;
  u_int  c,*image_data,*pixels;
#if (TIME_TEST > 0)
  NSDate *date = [NSDate date];
#endif
  assert(controller);
  assert(data);

  satlev = [controller satlev];
  pixels = [controller pixels];
  raw_w  = [controller raw_w];
  raw_h  = [controller raw_h];
  mag_w = (int)safeBounds.size.width/mag;  if (mag_w > 1) mag_w -= ((mag_w+1)%2);
  mag_h = (int)safeBounds.size.height/mag; if (mag_h > 1) mag_h -= ((mag_h+1)%2);
  assert((mag_w > 0) && (mag_h > 0));
  img_w = mag_w * mag;
  img_h = mag_h * mag;
  assert(sizeof(u_int) == 4);
  image_data = (u_int*)malloc(img_w*img_h*sizeof(u_int));

  for (xx=0; xx<mag_w; xx++) {         // create image
    x = mag_x - mag_w/2 + xx;
    for (yy=0; yy<mag_h; yy++) {
      y = mag_y - mag_h/2 + yy;
      if ((x<0) || (x>=raw_w) || (y < 0) || (y >= raw_h)) c = outcol;
      else if (data[x+y*raw_w] >= satlev)                 c = pixels[0xff];
      else {
        colour = (QLT_NCOLS*(data[x+y*raw_w] - mi))/dy;
        if (colour >= QLT_NCOLS) c = pixels[QLT_NCOLS-1];
        else if (colour < 0)     c = pixels[0];
        else                     c = pixels[colour];
      }
      set_col(image_data,xx*mag,yy*mag,img_w,mag,c);
    }
  }
  YimageData(self,image_data,img_w,img_h); // centered on view
  free((void*)image_data);
#if (TIME_TEST > 0)
  fprintf(stderr,"%s(): runtime= %.3f\n",PREFUN,YsecSince(date));
#endif
}

/* ---------------------------------------------------------------- */
/* NSView overrides */
/* ---------------------------------------------------------------- */

- (void)mouseDown:(NSEvent*)event
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,event);
#endif
  assert([NSThread isMainThread]); 

  [[self window] makeFirstResponder:self];
}

/* ---------------------------------------------------------------- */

- (void)keyDown:(NSEvent*)event
{
  int   dx=0,dy=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p), keyCode=%d\n",PREFUN,event,[event keyCode]);
#endif
  assert(controller);

  NSString *chars = [event charactersIgnoringModifiers]; 
  switch ([chars characterAtIndex:0]) {
    case NSLeftArrowFunctionKey:  dx = -1; dy =  0; break;
    case NSRightArrowFunctionKey: dx = +1; dy =  0; break;
    case NSDownArrowFunctionKey:  dx =  0; dy = +1; break;
    case NSUpArrowFunctionKey:    dx =  0; dy = -1; break;
  }
  if ((dx == 0) && (dy == 0)) {
    if ([chars compare:@" "] == NSOrderedSame) {   // FWHM fit
      [controller update_fwhm:YES];
    }
    return;
  }
  
  NSUInteger mod = [event modifierFlags];
  if (mod & NSEventModifierFlagControl) { dx *= 5; dy *=5; }
  if (mod & NSEventModifierFlagOption)  { dx *= 5; dy *=5; }
  if (mod & NSEventModifierFlagCommand) { dx *= 5; dy *=5; }
  if (mod & NSEventModifierFlagShift)   { dx *= 5; dy *=5; }

  float x = [controller mag_x] + (float)dx;
  float y = [controller mag_y] + (float)dy;
  [controller update_magnifierX:x Y:y];
  
  if ([controller fwhm] > 0) [controller update_fwhm:YES];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

