/* ---------------------------------------------------------------- */
//
// QltoolController.m
// M2FS/Swope/PFS/Henrietta
//
// Created by Christoph Birk on 2011-12-08
// Copyright 2011 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG           0              // debug level

#define TIME_TEST       1
#define GFIT_TEST       0

/* ---------------------------------------------------------------- */

#import  "QltoolController.h"
#import  "CameraController.h"
#import  "main.h"

#import  "MTwister.h"
#import  "Yutils.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

#define FRAME_QLT       "QlToolWindow"
#define FRAME_QLM       "QlToolMagnifier"


#define MAX_USHORT      0xffff
#define MAX_BYTE        0xff

/* ---------------------------------------------------------------- */

static const u_int satcol=0x000000ff;  // 0bgR

/* ---------------------------------------------------------------- */

static int linear(int value,double x1,double x2,int ncolors)
{
  double  x;
 
  x = (double)value/(double)ncolors;
  x = (x-x1)/(x2-x1);
  if (x < 0.0) x = 0.0;
  if (x > 1.0) x = 1.0;
 
  return (int)(x*MAX_BYTE);
}

/* ---------------------------------------------------------------- */
 
static int circle(int value,double m,double radius,int virtcols)
{
  double x;
 
  x = (double)value/(double)(virtcols-1);
 
  if ((x-m)*(x-m) >= radius*radius) return 0;
  
  return (int)(MAX_BYTE*sqrt(radius*radius - (x-m)*(x-m))/radius);
}

/* ---------------------------------------------------------------- */

#if (GFIT_TEST > 0)

static double fn_gauss(double x,double a,double b)
{
  return(a * exp(-(x*x/(2.0*b*b))));
}
 
#endif

/* ---------------------------------------------------------------- */

static double get_background(int* data,int dimx,int dimy,
                             int x0,int y0,int r,double* noise)
{
  int     r2=r*r,n=0,x,y;
  int     *sortval,d;
  double  s1=0.0,s2=0.0,back=-1.0;

  sortval = (int*)malloc(4*r2*sizeof(int));

  for (x=x0-r; x<=x0+r; x++) {         /* find background */
    for (y=y0-r; y<=y0+r; y++) {       /* circular aperture */
      if ((x<0) || (x>=dimx) || (y<0) || (y>=dimy)) continue;
      if ((x-x0)*(x-x0)+(y-y0)*(y-y0) < r2) continue; /* use outside */
      d = data[x+y*dimx];
      if (d > 0) {
        sortval[n] = d; n++;
        if (noise) { s1 += (double)d; s2 += (double)d * (double)d; }
      }
    }
  }
  if (n > 0) {                         /* at least 1 pixel */
    if (n > 1) q_sort(sortval,0,n-1);
    back = (double)sortval[n/2];
    if (noise) *noise = sqrt(s2/n - (s1/n)*(s1/n));
  }
  free((void*)sortval);

#if (DEBUG > 2)
  fprintf(stderr,"%s(): x0=%d, y0=%d, back=%.0f (n=%d)\n",PREFUN,x0,y0,back,n);
#endif

  return (double)back;
}

/* ---------------------------------------------------------------- */

static double get_centroid(int* data,int dimx,int dimy,
                           int x0,int y0,int r,double back,double* cx,double* cy)
{
  int    r2=r*r,x,y;
  double m=0.0,h;

  if (back < 0) back = get_background(data,dimx,dimy,x0,y0,r,NULL);
  if (back < 0) return back;
  
  *cx = *cy = 0.0;
  for (x=x0-r; x<=x0+r; x++) {         /* find centeroid */
    for (y=y0-r; y<=y0+r; y++) {
      if ((x<0) || (x>=dimx) || (y<0) || (y>=dimy)) continue;
      if ((x-x0)*(x-x0)+(y-y0)*(y-y0) > r2) continue; /* use inside */
      h = (double)data[x+y*dimx] - back;
      if (h < 0.0) continue;           /* signal too low */
      *cx += h*x;
      *cy += h*y;
      m   += h;
    }
  }
  if (m <= 0.0) return -1.0;
  *cx /= m;                            /* normalize centroid */
  *cy /= m;

#if (DEBUG > 2)
  fprintf(stderr,"%s(): cx=%.1f, cy=%.1f (m=%.1f)\n",PREFUN,*cx,*cy,m);
#endif

  return back;
}

/* ---------------------------------------------------------------- */
/* modified from 'Numerical Recipes in C' */

static double lfit(double* x,double* y,double* e,int n,double* a,double* b)
{
  int    i;
  double delta,h;
  double s=0.0,sx=0.0,sy=0.0,sxx=0.0,sxy=0.0,chi2=0.0;

  for (i=0; i<n; i++) {
    h    = e[i]*e[i];                  
    assert(h > 0);                     
    s   += 1.0/h;                      /* Warning: assumes all e[i] > 0 */
    sx  += x[i]/h;
    sy  += y[i]/h;
    sxx += x[i]*x[i]/h;
    sxy += x[i]*y[i]/h;
  }
  delta = s*sxx - sx*sx;
  *a    = (sxx*sy-sx*sxy)/delta;       /* y = a + bx */
  *b    = (s*sxy-sx*sy)/delta;
#if (DEBUG > 2)
  fprintf(stderr,"%s(): a=%f, b=%f\n",PREFUN,*a,*b);
#endif

  for (i=0; i<n; i++) {                /* chi-square */
    chi2 += sqd((y[i] - *a  - (*b)*x[i]) / e[i]);
  }

  return chi2;
} 

/* ---------------------------------------------------------------- */

static double gfit(double* x,double* y,int n,double* a,double* b,
                   double egain,double enoise)
{
  int    i;
  double *xval,*yval,*eval,chi2;

  xval = (double*)malloc(n*sizeof(double));
  yval = (double*)malloc(n*sizeof(double));
  eval = (double*)malloc(n*sizeof(double));

  for (i=0; i<n; i++) {                /* 'linearize' data */
    xval[i] = x[i]*x[i];               
    assert(y[i] > 0);                  /* Warning: assumes all y[i] > 0 */
    yval[i] = log(y[i]);               /* eval ~ 1/weight */
    eval[i] = (xval[i]+1) * ((sqrt(y[i]*egain+enoise*enoise)/egain)/y[i]);
  }
#if (GFIT_TEST > 0)
  { FILE *fp;
    fp = fopen("/tmp/linear.dat","w");
    for (i=0; i<n; i++) fprintf(fp,"%f  %f  %f\n",xval[i],yval[i],eval[i]);
    fclose(fp);
  }
#endif

  chi2 = lfit(xval,yval,eval,n,a,b);   /* linear regression fit */
  *a = exp(*a);                        /* 'de-linearize' */
  *b = sqrt(-0.5/(*b));
#if (DEBUG > 1)
  fprintf(stderr,"%s(): a=%f, b=%f\n",PREFUN,*a,*b);
#endif
  
  free((void*)xval); free((void*)yval); free((void*)eval);

  return chi2;
}
 
/* ---------------------------------------------------------------- */

static double get_fwhm(int* data,int dimx,int dimy,int x0,int y0,
                       int r,float* flx)
{
  int     x,y,r2=r*r,n;
  double  *flux,*dist,back,nbck=0,fwhm=0,a,b,d,cx=0,cy=0;
  const double  egain=1.0,enoise=1.0; // TODO as parameters
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%d,%d,%d,%d,%d)\n",PREFUN,data,dimx,dimy,x0,y0,r);
#endif
  
  *flx = 0.0f;
  if (!data) return 0.0;

#if (TIME_TEST > 1)
  NSDate *date = [NSDate date];
#endif
  back = get_background(data,dimx,dimy,x0,y0,r,&nbck);
#if (DEBUG > 1)
  fprintf(stderr,"%s(): back=%.0f, noise=%.1f\n",PREFUN,back,nbck);
#endif
  back = get_centroid(data,dimx,dimy,x0,y0,r,back,&cx,&cy);
#if (DEBUG > 1)
  fprintf(stderr,"%s(): cx=%.1f, cy=%.1f\n",PREFUN,cx,cy);
#endif

  flux = (double*)malloc(4*r2*sizeof(double));
  dist = (double*)malloc(4*r2*sizeof(double));
  n = 0;
  for (x=x0-r; x<=x0+r; x++) {         /* get flux */
    for (y=y0-r; y<=y0+r; y++) {       /* circular aperture */
      if ((x<0) || (x>=dimx) || (y<0) || (y>=dimy)) continue;
      d = ((double)x-cx)*((double)x-cx)+((double)y-cy)*((double)y-cy);
      if (d > r2) continue;            /* use inside */
      flux[n] = (double)data[x+y*dimx] - back;
      if (flux[n] < 1+3.0*nbck) continue;  /* flux too low */
      dist[n] = sqrt(d);
      n++;
    }
  }
#if (DEBUG > 1)
  fprintf(stderr,"%s(): n=%d\n",PREFUN,n);
#endif

  if (n > 2) {
#if (DEBUG > 0)
    double chi2 = gfit(dist,flux,n,&a,&b,egain,enoise);
#else
    gfit(dist,flux,n,&a,&b,egain,enoise);
#endif
    if ((a > 0) && (b > 0)) { double f; // sanity check
      fwhm = 2.355*b;                   // 2*sqrt(-2*ln(0.5))
      f = (float)peak2flux(a,fwhm);     // NOTE: valid only if pixscale<<fwhm
#if (DEBUG > 0)
      fprintf(stderr,"%s(): fwhm=%.1f, flux=%f, chi=%f\n",PREFUN,fwhm,f,sqrt(chi2/n));
#endif
      if (flx) *flx = (float)f;
#if (GFIT_TEST > 0)
      { int i; double d; FILE *fp;
        fp = fopen("/tmp/pixel.dat","w");
        for (i=0; i<n; i++) fprintf(fp,"%f %f\n",dist[i],flux[i]);
        fclose(fp);
        fp = fopen("/tmp/gauss.dat","w");
        for (d=0.; d<(double)r; d+=.05) fprintf(fp,"%f %f\n",d,fn_gauss(d,a,b));
        fclose(fp);
      }
#endif
    }
  } 
  free((void*)flux); free((void*)dist);

#if (TIME_TEST > 1)
  fprintf(stderr,"%s(): runtime= %4.2f\n",PREFUN,YmsSince(date));
#endif

  return fwhm;
}

/* ---------------------------------------------------------------- */

@implementation QltoolController
{
  int     flip_x,flip_y;
  BOOL    subtract_sky;
  u_int   raw_w,raw_h;
  int     *raw_data,*sky_data,*img_data;
  int     data_min,data_max,data_dyn;
  NSMutableArray *frames;
}

@synthesize raw_w,raw_h;
@synthesize contrast,slevel;
@synthesize fwhm,pixels,satlev,lut;
@synthesize mag_view,mag,mag_x,mag_y;

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

// none

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)initWithWindowNibName:(NSString*)name
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif
  
  if (!(self = [super initWithWindowNibName:name])) return nil;

  twister = [[MTwister alloc] init];

  assert(!raw_data);
  raw_w = raw_h = 2048/LO_MEM;
  raw_data = (int*)malloc(raw_w*raw_h*sizeof(int));
  sky_data = (int*)calloc(raw_w*raw_h,sizeof(int));
  img_data = (int*)malloc(raw_w*raw_h*sizeof(int));

  frames = [[NSMutableArray alloc] initWithCapacity:100];
  
  flip_x = 0;
  flip_y = 1;
  data_dyn = 1;
  contrast = 1.0;
  slevel = 0.0;

  [[self window] setFrameUsingName:@FRAME_QLT];
#ifdef NDEBUG
  [[self window] setStyleMask:([[self window] styleMask] & ~NSWindowStyleMaskClosable)];
#endif

  mag_view.controller = self;
  mag_view.mag = 4; 

  [[mag_view window] setFrameUsingName:@FRAME_QLM];
#ifdef NDEBUG
  [[mag_view window] setStyleMask:([[mag_view window] styleMask] & ~NSWindowStyleMaskClosable)]; 
#endif

  return self;
}

/* ---------------------------------------------------------------- */

- (void)windowDidLoad
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  assert(over_view);
  over_view.controller = self;

#if 0 // unused
  [self updateCompass];
#endif

  [self reloadPreferences:NO];

  stretch = (int)[popup_stretch indexOfSelectedItem];
  lut = (int)[popup_lut indexOfSelectedItem];
  [self fill_pixels];
  
  radius = [step_radius intValue]; 
  [edit_radius setIntValue:radius];    // < 1000

  [self action_mag:nil];
  [self action_radius:nil];
}

/* ---------------------------------------------------------------- */

- (void)windowWillClose:(NSNotification*)note
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,note);
#endif

  if ([note object] == [self window]) {
    YwindowFrameSave([self window],FRAME_QLT);
  } else
  if ([note object] == [mag_view window]) {
    YwindowFrameSave([mag_view window],FRAME_QLM);
  }
}

/* ---------------------------------------------------------------- */

- (void)reloadPreferences:(BOOL)redraw
{
#if (DEBUG > 1) 
  fprintf(stderr,"%s(%d)\n",PREFUN,(int)redraw);
#endif

  satlev = YprefGetInt(DBE_SATLEV,DEF_SATLEV,NO); 
  
  if (redraw) [self redrawPixels];
}

/* ---------------------------------------------------------------- */

- (void)fill_pixels
{
  int     i,s,virt,offset,ncols=QLT_NCOLS;
  u_int   col;

#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,lut);
#endif
  assert(sizeof(u_int) == 4);
  assert(ncols < 0xfe);
  
  if (!pixels) pixels = (u_int*)malloc(0x100*sizeof(u_int));

  offset = (int)((double)ncols/2.0*(1.0-slevel-fabs(contrast)));
  for (i=0; i<ncols; i++) {
    virt = (int)((i-offset)/fabs(contrast)); /* virtual color cell  */
    if (contrast < 0.0)  virt = ncols-virt;
    if (virt < 0)        virt = 0;
    if (virt >= ncols)   virt = ncols-1;
    switch (lut){
    case LUT_BBDY:                     /* black body */
      col  = linear(virt,0.00,0.50,ncols);       // R
      col += linear(virt,0.25,0.75,ncols) << 8;  // G
      col += linear(virt,0.50,1.00,ncols) << 16; // B
      break;
    case LUT_RAIN:                     /* rainbow */
      col  = circle(virt,1.,.55,ncols);          // R
      col += circle(virt,.5,.45,ncols) << 8;     // G
      col += circle(virt,0.,.65,ncols) << 16;    // B
      break;
    case LUT_GREY:                     /* grayscale */
    case LUT_INV:                      /* inverted grayscale */
    default:
      if (lut == LUT_INV) s = (int)(((ncols-virt)*MAX_BYTE)/ncols);
      else                s = (int)(MAX_BYTE-((ncols-virt)*MAX_BYTE)/ncols);
      col = s + (s << 8) + (s << 16);
      break;
    }
    pixels[i] = col; // printf("col[%3d]=%06x\n",i,pixels[i]);
  }
  pixels[0xff] = satcol;
}

/* ---------------------------------------------------------------- */
#pragma mark "Actions"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_lut:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  lut = (int)[sender indexOfSelectedItem];
  [self fill_pixels];
  [self redrawPixels];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_stretch:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  stretch = (int)[sender indexOfSelectedItem]; 
  [edit_min setEnabled:(stretch) ? NO : YES];
  [edit_max setEnabled:(stretch) ? NO : YES];
  [over_view squeeze_colormap:nil];  // reset contrast/level
  [self redrawWithScaling:YES];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_minmax:(id)sender // 'manual' scaling
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif

  if (sender == edit_min) data_min = [sender intValue];
  if (sender == edit_max) data_max = [sender intValue];
  data_dyn = data_max - data_min;
  if (data_dyn == 0) data_dyn = 1;
  
  [self redrawWithScaling:YES];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_mag:(id)sender
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  Ystr2chr([popup_mag titleOfSelectedItem],buf,sizeof(buf));
  mag = atoi(buf);
  mag_view.mag = mag;
  mag_view.radius = mag*radius;
  if (sender) { 
    [over_view setNeedsDisplay:YES];
    [self update_magnifierX:mag_x Y:mag_y];
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_radius:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  radius = [step_radius intValue];
  [edit_radius setIntValue:radius];    // < 1000
  mag_view.radius = mag * radius;
  if (sender) {
    [self update_magnifierX:mag_x Y:mag_y];
    if (fwhm > 0) [self update_fwhm:YES]; 
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_sky:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  subtract_sky = ([sender state]==NSControlStateValueOn) ? YES : NO;
  
  [self updateImageData];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_frame:(id)sender
{
#if (DEBUG > 1)
  fprintf(stderr,"%s: step=%d\n",PREFUN,[step_frame intValue]);
#endif

  int f = (int)step_frame.intValue;
  if (f > 0) {
    dispatch_async(GLOBAL_QUEUE,^(void) { [self displayFrame:f]; });
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Data"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)resetTitle:(NSString*)tstr
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,tstr);
#endif

  [frames removeAllObjects];
  dispatch_async(MAIN_QUEUE,^(void) {
    [step_frame setMinValue:0];
    [step_frame setMaxValue:0];
    [edit_frame setIntValue:0];
    if (tstr) YwindowTitle([self window],[tstr UTF8String]);
    else      YwindowTitle([self window],"Overview");
  });
}

/* ---------------------------------------------------------------- */

- (void)copyData:(int*)data asSky:(BOOL)sky
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d)\n",PREFUN,data,sky);
#endif

  int *ptr = (sky) ? sky_data : raw_data;
  memcpy(ptr,data,raw_w*raw_h*sizeof(int));

  [self updateImageData];
}

/* ---------------------------------------------------------------- */

- (void)displayFrame:(int)f
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,f);
#endif
  assert(f > 0);
  assert(f <= frames.count);

  dispatch_async(MAIN_QUEUE,^(void) {
    assert(edit_frame);
    [edit_frame setIntValue:f];
    assert(f <= [step_frame maxValue]);
    [step_frame setIntValue:f];
  });
  NSData *d = frames[f-1];
  [self copyData:(int*)d.bytes asSky:NO];
}


/* ---------------------------------------------------------------- */

- (void)addFrame:(NSData*)data
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%ld)\n",PREFUN,frames.count);
#endif
  assert(![NSThread isMainThread]);
  
  [frames addObject:data];
  dispatch_async(MAIN_QUEUE,^(void) {
    assert(step_frame);
    [step_frame setMaxValue:frames.count];
    [step_frame setMinValue:1];
  });
  [self displayFrame:(int)frames.count];
}

/* --- */

- (void)addRawFrame:(NSData*)data // convert to (int)
{
  int *b,*buffer;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%ld)\n",PREFUN,frames.count);
#endif
  assert(![NSThread isMainThread]);
  
  u_short *bytes = (u_short*)data.bytes;
  u_int n = (u_int)(data.length/sizeof(u_short));
  assert(n == raw_w*raw_h);
  buffer = b = (int*)malloc(n*sizeof(int));
  for (int i=0; i<n; i++,bytes++,b++) *b = (int)*bytes;
  NSData *d = [NSData dataWithBytesNoCopy:buffer length:(n*sizeof(int))];
  [self addFrame:d];
}

/* ---------------------------------------------------------------- */
#pragma mark "Update"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)updateImageData
{
  const int n=raw_w*raw_h;
  register int i;
#if (DEBUG > 1)
  fprintf(stderr,"%s: sub=%d\n",PREFUN,subtract_sky);
#endif

#if (TIME_TEST > 1)
  NSDate *date1 = [NSDate date];
#endif
  register int *img = img_data;
  if (subtract_sky) {
    register int *raw = raw_data;
    register int *sky = sky_data;
    for (i=0; i<n; i++,img++,raw++,sky++) *img = *raw - *sky;
  } else {
    memcpy(img_data,raw_data,raw_w*raw_h*sizeof(int));
  }
  assert(!flip_x);
  if (flip_y) {
    img = img_data; register int h,*p1,*p2; const int h2=raw_h/2; 
    for (int x=0; x<raw_w; x++) {
      p1 = img + x; p2 = p1 + (raw_h-1)*raw_w;
      for (i=0; i<h2; i++) {
        assert(p1 != p2);
        h = *p1; *p1 = *p2; *p2 = h; p1 += raw_w; p2 -= raw_w;
      }
    }
  }
#if (TIME_TEST > 1)
  fprintf(stderr,"%s(): run-time= %.3f\n",PREFUN,YsecSince(date1));
#endif

  [self redrawWithScaling:YES];
}

/* ---------------------------------------------------------------- */

- (void)redrawWithScaling:(BOOL)rescale
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,rescale);
#endif

  if (rescale) [self scale_data];
  [over_view newData:img_data min:data_min dyn:data_dyn];
  [self update_magnifierX:mag_x Y:mag_y];
}

/* ---------------------------------------------------------------- */

- (void)redrawPixels
{
  [over_view newPixels];
 
  [self update_magnifierX:mag_x Y:mag_y];
}

/* ---------------------------------------------------------------- */

- (void)update_magnifierX:(float)mx Y:(float)my
{
  const int stat_mode=1; //0:no clip, 1:4sigma, 2:3sigma TODO stat_mode -- action_stat
  int     x,y,v,n,nlast,*sortval=NULL;
  int     aper_min=MAX_USHORT,aper_max=-MAX_USHORT;
  int     clip_min=-MAX_USHORT,clip_max=MAX_USHORT;
  double  aper_ave=0.0,aper_sig=0.0,s1,s2;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.1f,%.1f): mainThread=%d\n",PREFUN,mx,my,[NSThread isMainThread]);
#endif
  assert(img_data);
  
  mag_x = (int)my_round(fmin(raw_w-1,fmax(0,mx)),0); 
  mag_y = (int)my_round(fmin(raw_h-1,fmax(0,my)),0); 

  const int r2 = radius*radius; n=4*r2;
  if (stat_mode) sortval = (int*)malloc(n*sizeof(int));
  do {
    nlast=n; n=0; s1=s2=0.0;
    for (x=mag_x-radius; x<=mag_x+radius; x++) {   // statistics
      if ((x<0) || (x>=raw_w)) continue;
      for (y=mag_y-radius; y<=mag_y+radius; y++) {
        if ((y<0) || (y>=raw_h)) continue;
        if ((x-mag_x)*(x-mag_x)+(y-mag_y)*(y-mag_y) > r2) continue;
        v = img_data[x+raw_w*y];
        if (nlast == 4*r2) {   // min/max on first pass only
          if (v > aper_max) aper_max = v;
          if (v < aper_min) aper_min = v;
        }
        if ((v >= clip_min) && (v <= clip_max)) {
          if (sortval) sortval[n] = v;
          s1 += (double)v; s2 += (double)v * (double)v;
          n++;
        }
      } // endfor(y)
    } // endfor(x)
    if (n < 2) break;
    aper_ave = s1/n;
    aper_sig = sqrt(s2/n-aper_ave*aper_ave);
    if (!sortval) break;
    q_sort(sortval,0,n-1);
    int med = sortval[n/2];
    clip_min = (int)fmax(-MAX_USHORT,med-(5.0-stat_mode)*aper_sig);
    clip_max = (int)fmin( MAX_USHORT,med+(5.0-stat_mode)*aper_sig);
#if (DEBUG > 2)
    fprintf(stderr,"\tclip: %d %d %.3f (%d)\n",clip_min,clip_max,aper_sig,n);
#endif
  } while (n < nlast);
  if (sortval) free((void*)sortval);
  
  [self update_fwhm:NO];

  mag_view.mag_x = mag_x;
  mag_view.mag_y = mag_y;
  [mag_view newData:img_data min:data_min dyn:data_dyn];

  dispatch_async(MAIN_QUEUE,^(void) { char buf[128];
    sprintf(buf,"%d,%d = %d",flip_x ? raw_w-mag_x : mag_x+1,
                             flip_y ? raw_h-mag_y : mag_y+1,img_data[mag_x+mag_y*raw_w]);
    YeditSet(edit_pixel,buf);
    sprintf(buf,"%d,%d",aper_min,aper_max);
    YeditSet(edit_minmax,buf);
    sprintf(buf,"%.1f, %.2f",aper_ave,aper_sig);
    if      (aper_ave > satlev)   YeditColor(edit_avesig,1.0f,0.0f,0.0f,1.0f);
    else if (aper_ave > satlev/2) YeditColor(edit_avesig,0.7f,0.7f,0.0f,1.0f);
    else                          YeditColor(edit_avesig,0.0f,0.0f,0.0f,1.0f);
    YeditSet(edit_avesig,buf);
    over_view.needsDisplay = YES;
  });

}

/* ---------------------------------------------------------------- */

- (void)update_fwhm:(BOOL)recalc
{
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,recalc);
#endif

  if (recalc) {
    fwhm = (float)get_fwhm(img_data,raw_w,raw_h,mag_x,mag_y,radius,&flux);
  }
  if (fwhm > 0.0f) sprintf(buf,"%.1f / %.0f",fwhm,flux);
  else              strcpy(buf,"n/a");
  YeditSet(edit_fwhm,buf);            // mainThread
}

/* ---------------------------------------------------------------- */

- (void)update_mimawin
{
  char buf[32];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  sprintf(buf,"%d",data_min); YeditSet(edit_min,buf); // 1000->1,000
  sprintf(buf,"%d",data_max); YeditSet(edit_max,buf); // 1000->1,000
}

/* ---------------------------------------------------------------- */

- (int)arrayStatsLo:(int)lo cutHi:(int)hi med:(int*)median sig:(int*)sigma
{
  const int n=raw_w*raw_h;
  const int step=(raw_w/(2*N_CHANNELS));
  int       *array,*d,j=0,x,y,x0,y0;
  double    s1=0.0,s2=0.0;
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,lo,hi);
#endif
#if (TIME_TEST > 2)
  NSDate *date = [NSDate date];
#endif

  assert(img_data);
  array = (int*)malloc((n/(step*step))*sizeof(int));

  assert(step >= 2*REF_PIXELS);
  y0 = REF_PIXELS + [twister intValue:(1+step-2*REF_PIXELS)];
  for (y=y0; y<raw_h; y+=step) {
    d = img_data + y*raw_w;
    x0 = REF_PIXELS + [twister intValue:(1+step-2*REF_PIXELS)];
    for (x=x0; x<raw_w; x+=step,d+=step) {
      if (*d < lo) continue;
      if (*d > hi) continue;
      array[j] = *d;  j++;
      s1 += (double)*d; s2 += (double)*d * (double)*d;
    }
  }
  assert(j <= n/(step*step));
  if (j < 3) {                         // not enough data
    *median = 0; *sigma = 1;
  } else {
    q_sort(array,0,j-1);
    *median = array[j/2];
    *sigma = (int)sqrt(s2/j - (s1/j)*(s1/j)); if (*sigma < 1) *sigma = 1;
  }
#if (DEBUG > 2)
  fprintf(stderr,"%s(): j=%d, m=%d, s=%d\n",PREFUN,j,*median,*sigma);
#endif
  free((void*)array);
#if (TIME_TEST > 2)
  fprintf(stderr,"%s(): cpu= %4.2f [ms]\n",PREFUN,YmsSince(date));
#endif
  
  return j;
}

/* ---------------------------------------------------------------- */

- (void)scale_medLo:(int)lo Hi:(int)hi
{
  int median,sigma,min,max=0xfff0;
#if (DEBUG > 2)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,lo,hi);
#endif

#if (TIME_TEST > 2)
  NSDate *date = [NSDate date];
#endif
 
  min = (subtract_sky) ? -0xfff0 : 0x000f;
  [self arrayStatsLo:min cutHi:max med:&median sig:&sigma]; // pass1
  min = median-5*sigma;
  max = median+5*sigma;
  [self arrayStatsLo:min cutHi:max med:&median sig:&sigma]; // pass2

  data_min = median - lo*sigma;
  data_max = median + hi*sigma;

#if (TIME_TEST > 2)
  fprintf(stderr,"%s(): cpu= %4.2f [ms] (%d,%d)\n",PREFUN,YmsSince(date),data_min,data_max);
#endif
}

/* ---------------------------------------------------------------- */

- (void)scale_data
{
  if (!img_data) {
    data_min = data_max = 0; data_dyn = 1;
    return;
  }

  switch (stretch) { 
  case STRETCH_MED3:               /* median -3/+3 sigma */
    [self scale_medLo:3.0 Hi:3.0];
    break;
  case STRETCH_MED5:               /* median -5/+5 sigma */
    [self scale_medLo:5.0 Hi:5.0];
    break;
  case STRETCH_310:                /* median -3/+10 sigma */
    [self scale_medLo:3.0 Hi:10.0];
    break;
  case STRETCH_MIMA:               /* min/max scaling */
    { register int i,npix=raw_w*raw_h;
      register int *data=img_data;
      data_min = MAX_USHORT; data_max = -MAX_USHORT;
      for (i=0; i<npix; i++, data++) {
        if ((int)*data > data_max) data_max = (int)*data;
        if ((int)*data < data_min) data_min = (int)*data;
      }
    }
    break;
  case STRETCH_CUTS:               /* manual scaling */
  default:
    break;
  }
  data_dyn = data_max - data_min;
  if (data_dyn == 0) data_dyn = 1;
  [self update_mimawin];
}

/* ---------------------------------------------------------------- */
#pragma mark "NSWindow delegates"
/* ---------------------------------------------------------------- */

#if 0 // unused
- (void)windowDidResize:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): live=%d\n",PREFUN,note,[[note object] inLiveResize]);
#endif
  
  if ([[note object] inLiveResize]) return;
}

/* --- */

- (void)windowDidEndLiveResize:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): object=%p\n",PREFUN,note,[note object]);
#endif
  assert([NSThread isMainThread]);

  [self performSelector:@selector(windowDidResize:) withObject:note
        afterDelay:0.25];
}
#endif

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
