/* ---------------------------------------------------------------- */
//
// LakeIO.m
// FourStar / Henrietta
//
// Created by Christoph Birk on 2013-09-10
// Copyright (c) 2014 Christoph Birk. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG           1              // debug level

#define RESPONSE_SIZE   512

/* ---------------------------------------------------------------- */

#import "LakeIO.h"

#import "Logger.h"

extern BOOL use_shared;

/* ---------------------------------------------------------------- */

@implementation LakeIO

Logger *lakeio_logger;

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (void)initialize
{
  if (self != [LakeIO class]) return;  // prevent sub-classes
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  lakeio_logger = [[Logger alloc] initWithBase:"lakeio" useShared:use_shared];
  lakeio_logger.mode = LOGGER_FLUSH; // keep-open, auto-flush

}

/* ---------------------------------------------------------------- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
    case E_lakeio_error: string = @"LakeIO: error"; break;
    case E_lakeio_resistor: string = @"LakeIO: invalid resistor value"; break;
    case E_lakeio_lock: string = @"LakeIO: failed to acquire lock"; break;
    case E_lakeio_sensor: string = @"LakeIO: sensor error"; break;
    default: string = [super errorString:err]; break;
  }
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)request:(const char*)cmd response:(char*)res
{
  int  err;
  char command[256];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%p)\n",PREFUN,cmd,res);
#endif

  err = [self open]; if (err) return err;
  if (![self mylock:lock_timeout]) {   // open/lock order is more MT-efficient
    [self close];
    return E_lakeio_lock;
  }
#if (DEBUG > 0)
  sprintf(command,"<- %s",cmd);
  [lakeio_logger append:command];
#endif

  assert(strlen(cmd) < sizeof(command)-3);
  sprintf(command,"%s\r\n",cmd);       // append <CR><LF>
  if (online) {
    assert(sock > 0);
    err = TCPIP_Send(sock,command,strlen(command));
  }

  if (!err && online) {
    if (res) { int i;
      for (i=0; i<RESPONSE_SIZE-1; i++) {
        err = TCPIP_Wait4Char(sock,&res[i],1000*sock_timeout);
        if (err) break;
        if (res[i] == terminator) break;
      }
      res[i] = '\0';                   // remove <LF>
      if (i > 0) res[i-1] = '\0';      // remove <CR>
      if (err) sprintf(res,"err=%d",err);
#if (DEBUG > 0)
      sprintf(command,"-> %s",res);
      [lakeio_logger append:command];
#endif
    }
  }
  if (!err) [NSThread sleepForTimeInterval:0.050]; // throttle (p.101)
  
  [lock unlock];
  [self close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)command:(const char*)cmd  // Note: does not expect an response
{
  int  err;
  
  err = [self request:cmd response:nil];
  
  return err;
}

/* ---------------------------------------------------------------- */

#if (DEBUG > 2)
- (void)close
{
  [super close];
  printf("lakeShore.count=%d\n",count);
}
#endif

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)getVersion:(char*)version
{
  if (!online) strcpy(version,"LakeSim v1.0");
  int err = [self request:"*IDN?" response:version];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setupCurrent:(float)max_cur resistor:(int)r channel:(int)c
{
  int  err,res=0;
  char cmd[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%f,%d)\n",PREFUN,max_cur);
#endif

  err = [self open]; if (err) return err;
  
  switch (r) {                         // resistor code {1,2}
    case 25: res = 1; break;
    case 50: res = 2; break;
  }
  int enabled = (res && (max_cur>0)) ? 1 : 0;
  int mode = (enabled) ? 1 : 0;        // 1=PID
  int input = (enabled) ? 1+c : 0;     // 1->A,2->B
  if (!err) {
    sprintf(cmd,"OUTMODE %d,%d,%d,0",1+c,mode,input);  // not <powerup enable>
    err = [self command:cmd];
  }
  if (!err) {                          // turn ramping OFF
    err = [self setRamp:0 channel:c];
  }
  if (!err) { 
    if (enabled) {
      sprintf(cmd,"HTRSET %d,%d,0,%.3f,2",1+c,res,max_cur); // display 'power'
      err = [self command:cmd];
    }
  }
  [self close];

  return err;
}


/* ---------------------------------------------------------------- */

- (int)setRamp:(float)max_ramp channel:(int)c
{
  int  err;
  char cmd[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.1f)\n",PREFUN,max_ramp);
#endif

  if (max_ramp > 0.0f) sprintf(cmd,"RAMP %d,1,%.1f",1+c,max_ramp);
  else                 sprintf(cmd,"RAMP %d,0",1+c);
  err = [self command:cmd];
  if (!err) [NSThread sleepForTimeInterval:0.100];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setP:(int)p_reg I:(int)i_reg D:(int)d_reg channel:(int)c
{
  int  err;
  char cmd[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d,%d)\n",PREFUN,p_reg,i_reg,d_reg);
#endif

  sprintf(cmd,"PID %d,%d,%d,%d",1+c,p_reg,i_reg,d_reg);
  err = [self command:cmd];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getP:(int*)p_reg I:(int*)i_reg D:(int*)d_reg channel:(int)c
{
  int  err;
  char cmd[128],buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p,%p,%p)\n",__func__,c,p_reg,i_reg,d_reg);
#endif

  sprintf(cmd,"PID? %d",1+c);
  err = [self request:cmd response:buf];
  if (!err && online) { float p,i; int d;
    if (sscanf(buf,"%f,%f,%d",&p,&i,&d) == 3) {
      *p_reg = (int)p; *i_reg = (int)i; *d_reg = d;
    }
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setRange:(int)range channel:(int)c
{
  int  err;
  char cmd[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,range);
#endif

  sprintf(cmd,"RANGE %d,%d",1+c,range);
  err = [self command:cmd];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getRange:(int*)range channel:(int)c
{
  int  err;
  char cmd[128],buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,range);
#endif

  sprintf(cmd,"RANGE? %d",1+c);
  err = [self request:cmd response:buf];
  if (!err && online) {
    *range = atoi(buf);              // {0,1,2,3}
    assert((*range >= 0) && (*range <= 3));
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setSetpoint:(float)setp channel:(int)c
{
  int  err;
  char cmd[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.2f)\n",PREFUN,setp);
#endif

  sprintf(cmd,"SETP %d,%.2f",1+c,setp);
  err = [self command:cmd];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)getSetpoint:(float*)setp channel:(int)c
{
  int  err;
  char cmd[128],buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,setp);
#endif

  sprintf(cmd,"SETP? %d",1+c);
  err = [self request:cmd response:buf];
  if (!err && online) *setp = (float)atof(buf);

  return err;
}

/* ---------------------------------------------------------------- */

- (int)getPower:(float*)power channel:(int)c
{
  char cmd[128],buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,power);
#endif

  sprintf(cmd,"HTR? %d",1+c);
  int err = [self request:cmd response:buf];
  if (!err && online) *power = (float)fmin(100.0,fmax(0.0,atof(buf)));
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getTemp:(float*)temp channel:(int)c
{
  char cmd[128],buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,temp);
#endif

  sprintf(cmd,"KRDG? %c",'A'+c);
  int err = [self request:cmd response:buf];
  if (!err && online) {
    *temp = (float)atof(buf);
//todo    if (*temp <= 0.0f) err = E_lakeio_sensor;
  }
  
  return err;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
