/* ---------------------------------------------------------------- */
//
// TempController.m
// FourStar / Henrietta
//
// Created by Christoph Birk on 2014-09-10
// Copyright 2005 Carnegie Observatories. All rights reserved.
//
// Note: "Equalize Channels" removed in v3
// Note: "Re-Open Connection" removed in v3
// Note: "Create PDF" removed in v3
//
/* ---------------------------------------------------------------- */

#define DEBUG           1              // debug level

/* ---------------------------------------------------------------- */

#import  "TempController.h"
#import  "GraphController.h"
#import  "AppDelegate.h"
#import  "Yutils.h"
#import  "Notice.h"
#import  "MTwister.h"
#import  "LakeIO.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

#define TC_N  2
#define TM_N  4

#define M_TITLE        "Temperature Controller"

#define KEY_FRAME      "TempConWindow"
#define KEY_TOOLTIP    "toolTip" 
#define DBE_MONITOR    "tempmon"

#define TIMER_PERIOD   2

#define ST_MIN         2
#define ST_MAX         10
#define LG_MAX         (600/(ST_MAX*TM_N))

#define SETP_MIN       70            // setpoint limits --- todo target
#define SETP_MAX       305

#define RAMP_MAX       1.00          // gradient limits

#define LOG2           0.693147      // log(2)

#define VAL 0.65
static GraphColor myColors[8] = {
  { 0.0, 0.0, 0.0 } ,   // black
  { 0.0, 0.0, VAL } ,   // blue
  { 0.0, VAL, 0.0 } ,   // green
  { VAL, 0.0, 0.0 } ,   // red
  { 0.0, VAL, VAL } ,   // turquoise
  { VAL, 0.0, VAL } ,   // purple
  { VAL, VAL, 0.0 } ,   // yellow
  { VAL, VAL, VAL }     // grey
};
#undef VAL

enum lock_tag { LOCK_TOGGLE=-2,LOCK_UPDATE,LOCK_NO,LOCK_YES};

/* ---------------------------------------------------------------- */

@implementation TempController
{
  NSString      *cName,*mName; // controlName, monitorName TODO
  NSTimer       *myTimer;
  u_long        counter;
  AppDelegate   *app;
  LakeIO        *lakeshore;
  GraphController *controlGraph,*powerGraph,*tempGraph;
  volatile BOOL initialized,reading;
  time_t        next_err_warning;
  int           delta_err_warning;
  float         heater_alarm,alarm_lim;   // TODO both unused
  int           locked;                   // GUI
  time_t        last_thinning;
  // TC
  float         setpoint[TC_N],target[TC_N];
  int           range[TC_N],enabled[TC_N];
  float         ramp[TC_N],power[TC_N],sim_setp[TC_N];
  time_t        last_power[TC_N];
  int           resistor[TC_N];
  double        maxcur[TC_N],maxpow[TC_N];
  double        power_ave[TC_N],sim_power[TC_N];
  int           p_term[TC_N],i_term[TC_N],d_term[TC_N];
  // TM
  volatile int  monitor[TM_N],last_channel;
  double        sim_temp[TM_N];
  float         temperature[TM_N];
  double        gradient[TM_N];
  time_t        last_temp[TM_N];
  int           sample_time;
}

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (void)initialize
{
  if (self != [TempController class]) return;  // prevent sub-classes
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif  

}

/* --- */


+ (NSDictionary*)defaultDictionary
{
  NSString *file = [[NSBundle mainBundle] pathForResource:@"tempcon" ofType:@"xml"];
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:file];
  assert(dict);
  
  return dict;
}

/* --- */

+ (NSString*)defaultHost
{
  NSDictionary *dict  = [TempController defaultDictionary];
  NSString *host = [dict objectForKey:@"host"];
  assert(host);
  
  return host;
}

/* --- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
  default:
    string = [LakeIO errorString:err];
    break;
  }
  if (!string) string = [AppDelegate errorString:err];
  
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (id)initWithWindowNib:(NSString*)name
{
  int      port,i;
  BOOL     online;
  char     host[128],buf[128];
  NSArray  *array;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,[name UTF8String]);
#endif

  if ((self = [super initWithWindowNibName:name]) == nil) return nil;

  myid = 1;                   // FourStar had multiple temp. controllers

  assert([NSThread isMainThread]);
  app = [NSApp delegate];
  assert(app);

  NSString *file = [[NSBundle mainBundle] pathForResource:@"tempcon" ofType:@"xml"];
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:file];
  cName = [[dict objectForKey:@"cName"] copy]; // controller TODO
  mName = [[dict objectForKey:@"mName"] copy]; // monitor
 
  for (i=0; i<TC_N; i++) {
    sprintf(buf,"resistor%d",1+i);
    resistor[i] = [[dict objectForKey:@(buf)] intValue];
    sprintf(buf,"max_cur%d",1+i);
    maxcur[i] = [[dict objectForKey:@(buf)] doubleValue];
    maxpow[i] = resistor[i] * maxcur[i] * maxcur[i];
    enabled[i] =  (resistor[i] && (maxcur[i] > 0)) ? 1 : 0;
    setpoint[i] = target[i] = 295.0f;
    sim_setp[i] = 295.0;
    ramp[i] = 0.0;
    range[i] = 0;
    sim_power[i] = 0.0;
    power_ave[i] = 0.0;
    power[i] = 0.0f;
#if 0 // TODO depends on channel and range
    key = [NSString stringWithFormat:@"p_range%d",range];
    p_term = [[dict objectForKey:key] intValue];
    key = [NSString stringWithFormat:@"i_range%d",range];
    i_term = [[dict objectForKey:key] intValue];
    key = [NSString stringWithFormat:@"d_range%d",range];
    d_term = [[dict objectForKey:key] intValue];
#endif
  }
  
  heater_alarm = [[dict objectForKey:@"heater_alarm"] floatValue];
  alarm_lim = [[dict objectForKey:@"alarm_limit"] floatValue];
  last_thinning = cor_time(0)+2;
  _stop_reading = FALSE;

  // --- Temperature Controller ---
  reading = initialized = FALSE;

  // --- Temperature  Monitor ---
  sample_time = ST_MIN;
  last_channel=TM_N-1;
  
  array = YprefGetArray(DBE_MONITOR,use_shared);
  if (array) YarrayToInts(array,(int*)monitor,TM_N);
  else { for (i=0; i<TM_N; i++) monitor[i]=1; }
  
  for (int i=0; i<TM_N; i++) {
    gradient[i] = 0.0;
    sim_temp[i] = 295.0-i*0.1;
  }

  sprintf(buf,"%s Temperatures",[cName UTF8String]);
  controlGraph = [[GraphController alloc] initWithWindowNib:@"Graph"
                                          context:myid title:buf
                                          unit:"K" N:TC_N
                                          colors:myColors];

  sprintf(buf,"%s Power",[cName UTF8String]);
  powerGraph = [[GraphController alloc] initWithWindowNib:@"Graph"
                                        context:myid title:buf
                                        unit:"%" N:TC_N
                                        colors:myColors];
  
  sprintf(buf,"%s Temperatures",[mName UTF8String]);
  tempGraph = [[GraphController alloc] initWithWindowNib:@"Graph"
                                       context:myid title:buf
                                       unit:"K" N:TM_N
                                       colors:myColors];

  // setup Lakeshore connection
#if defined(SIM_ONLY) && defined(NDEBUG)
  strcpy(host,LOCALHOST);
#else
  NSString *str = [TempController defaultHost];
  YprefGetChar(DBE_LAKE_HOST,[str UTF8String],host,sizeof(host),use_shared);
#endif
  port = 7777;
  online = (BOOL)YprefGetInt(DBE_LAKE_ONLINE,DEF_STATE,use_shared);
  lakeshore = [[LakeIO alloc] initWithHost:host port:port online:online];
  sprintf(buf,"lakeshore: %s:%d",host,port); [main_logger append:buf];

  [[self window] setFrameUsingName:@KEY_FRAME]; // forces lazy window
#ifdef NDEBUG
  [[self window] setStyleMask:([[self window] styleMask] & ~NSWindowStyleMaskClosable)];
#endif

  [self performSelectorInBackground:@selector(run_setup) withObject:nil];

  sprintf(buf,"%s %s up",[mName UTF8String],M_TITLE);
  [app message:@(buf) level:0];
  
  return self;
}

/* ---------------------------------------------------------------- */

- (void)windowDidLoad
{
  int     i,j;
  double  p;
  char    buf[512];
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  assert([NSThread isMainThread]);

  [self.window setTitle:[NSString stringWithFormat:@"%@TempController",cName]];

  // TM --- control arrays ---
  edit_temps = @[edit_temp1,edit_temp2,edit_temp3,edit_temp4]; [edit_temps retain];
  edit_grads = @[edit_grad1,edit_grad2,edit_grad3,edit_grad4]; [edit_grads retain];
  chk_mons = @[chk_mon1,chk_mon2,chk_mon3,chk_mon4]; [chk_mons retain];

  // TC --- control arrays ---
  edit_setps = @[edit_setp1,edit_setp2]; [edit_setps retain];
  edit_targets = @[edit_target1,edit_target2]; [edit_targets retain];
  edit_ramps = @[edit_ramp1,edit_ramp2]; [edit_ramps retain];
  popup_ranges = @[popup_range1,popup_range2]; [popup_ranges retain];
  edit_powers = @[edit_power1,edit_power2]; [edit_powers retain];

  NSString *file = [[NSBundle mainBundle] pathForResource:@"tempcon" ofType:@"xml"];
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:file];

  // labels, colors and tool-tips
  NSArray *wells = [NSArray arrayWithObjects:well1,well2,well3,well4,nil];
  NSArray *labels = [NSArray arrayWithObjects:label1,label2,label3,label4,nil];
  for (i=0; i<TM_N; i++) {
    NSTextField *label = [labels objectAtIndex:i];
    // [label setIntValue:(1+i+(myid-1)*TM_N)];
    // sprintf(buf,"%c",'A'+i); [label setStringValue:@(buf)];
    GraphColor col = myColors[i];
    NSColor *color = [NSColor colorWithCalibratedRed:col.r green:col.g blue:col.b alpha:1];
    NSColorWell *well = [wells objectAtIndex:i];
    [well setColor:color];
    NSString *key = [NSString stringWithFormat:@"%s%d",KEY_TOOLTIP,1+i];
    NSString *string = [dict objectForKey:key];
    [well setToolTip:string];
    NSButton *check = [chk_mons objectAtIndex:i];
    [check setState:((monitor[i]) ? NSControlStateValueOn : NSControlStateValueOff)];
    [tempGraph setChannel:i monitor:((monitor[i]) ? YES : NO)];
    [check setToolTip:string];
    [label setToolTip:string];
    [edit_temps[i] setToolTip:string];
    [edit_grads[i] setToolTip:string];
  }
  
  for (i=0; i<TC_N; i++) {                   // range labels
    NSPopUpButton *popup = popup_ranges[i];
    for (j=0,p=maxpow[i]/100.0; j<3; j++) { // add menu entries
#if 1
      sprintf(buf,"%.2f",p);
      while (buf[strlen(buf)-1] == '0') buf[strlen(buf)-1] = '\0'; // remove trailing '0'
      if (buf[strlen(buf)-1] == '.') buf[strlen(buf)-1] = '\0';    // remove trailing '.'
#else
      switch (i) {
        case 0: strcpy(buf,"lo"); break;
        case 1: strcpy(buf,"med"); break;
        case 2: strcpy(buf,"hi"); break;
        default: assert(0); break;
      }
#endif
      [popup addItemWithTitle:@(buf)];
      p *= 10.0;
    }
  }

  [panel_tune setTitle:[NSString stringWithFormat:@"%@ Controller",cName]];

  [self lockGUI:LOCK_YES];

  assert([NSThread isMainThread]);     // timer on main thread
  myTimer = [NSTimer scheduledTimerWithTimeInterval:TIMER_PERIOD
                     target:self selector:@selector(fireTimer:)
                     userInfo:nil repeats:YES];
}

/* ---------------------------------------------------------------- */

- (void)windowWillClose:(NSNotification*)notification
{
  int  timeout=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,notification);
#endif
  assert([NSThread isMainThread]);

  [myTimer invalidate]; self.stop_reading = TRUE;
  YwindowFrameSave(self.window,KEY_FRAME);

  while (reading) {
    Ysleep(50); timeout++; if (timeout > 12) break;
  }
}

/* ---------------------------------------------------------------- */

#ifdef NDEBUG
- (void)windowDidBecomeKey:(NSNotification*)notification
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,notification);
#endif

  sample_time = ST_MIN; // todo
}
#endif

/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)fireTimer:(NSTimer*)theTimer
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,theTimer);
#endif
  assert([NSThread isMainThread]);
  counter++;

  if (!lakeshore.online) { NSString *string;
    if (counter % 2) string = [NSString stringWithFormat:@"%@TempSimulator",cName];
    else             string = [NSString stringWithFormat:@"%@TempController",cName];
    [self.window setTitle:string];
  }

  if (!initialized) return;

  if (!lakeshore.online) {  // Simulator todo power,setpoint -- need Lakeshore verification
    @synchronized(self) { double rate,dt=TIMER_PERIOD;
      const double drift_target[TM_N]={100.0,120.0,290.0,285.0};
      for (int i=0; i<TM_N; i++) {
        if (i < TC_N) {                // temp. controller
          if (ramp[i]) {               // ramping on
            double sign = (target[i] < sim_setp[i]) ? -1 : ((target[i] > sim_setp[i]) ? +1 : 0);
            // printf("%f %f %.0f\n",target[i],sim_setp[i],sign);
            sim_setp[i] += sign*dt*ramp[i]/60.0;
            if (sign * (target[i]-sim_setp[i]) < 0) {
              sim_setp[i] = target[i]; printf("ramp done\n"); } //xxx
          }
          if (range[i] && (sim_temp[i] < setpoint[i])) {  // heater active
            rate = (setpoint[i] - sim_temp[i])/(300.0/range[i]);
            rate = fmin(rate,setpoint[i]-sim_temp[i]/dt);
            sim_power[i] = fmax(0.0,fmin(100.0,2.0*fabs(sim_temp[i]-target[i])/range[i]));
          } else {                     // drifing
            rate = (target[i] - sim_temp[i])/300.0; // TODO refine by power
            sim_power[i] = 0;
          }
        } else {                       // temp.monitor
          rate = (drift_target[i] - sim_temp[i])/300.0;
        }
        sim_temp[i] += rate*dt;
      } // endfor(i)
    } // synchronized(self)
  } // endif(!online)

  
  if (!reading) {                      // start sampling thread
    if (!self.stop_reading) {
      [self performSelectorInBackground:@selector(run_read) withObject:nil];
    }
  }

  if (cor_time(0) >= last_thinning+app.thinTime) { // remove redundant graph points
    dispatch_async(GLOBAL_QUEUE,^(void) {
      [controlGraph thinData:10.0 after:24];
      [powerGraph thinData:10.0 after:24];
      [tempGraph thinData:20.0 after:24];
    });
    last_thinning = cor_time(0);
  }
}

/* ---------------------------------------------------------------- */

- (void)lockGUI:(int)lock 
{
  assert(LOCK_NO == 0);
  assert(LOCK_YES == 1);
  switch (lock) {
    case LOCK_UPDATE: lock = locked; break;            // unchanged
    case LOCK_TOGGLE: lock = locked = !locked; break;  // toggle
    default: locked = lock; break;            // set
  }
  dispatch_async(MAIN_QUEUE,^(void) { BOOL e; int i;
    for (i=0; i<TC_N; i++) {
      e = enabled[i] && !lock && (ramp[i] == 0);
      [edit_setps[i] setEnabled:e];
      e = enabled[i] && !lock;
      [edit_targets[i] setEnabled:e];
      [edit_ramps[i] setEnabled:e];
      [popup_ranges[i] setEnabled:e];
    }
    for (i=0; i<TM_N; i++) {
      [chk_mons[i] setEnabled:!locked];
    }
  });
}

/* ---------------------------------------------------------------- */

- (void)updateGUI
{
  dispatch_async(MAIN_QUEUE,^(void) { char buf[128];
    for (int i=0; i<TC_N; i++) {
      if (monitor[i] && enabled[i]) {
        [self updateSetpoint:i fromHW:NO];
        [self updatePower:i fromHW:NO];
        sprintf(buf,"%.1f",target[i]); YeditSet(edit_targets[i],buf);
        sprintf(buf,"%.1f",ramp[i]); YeditSet(edit_ramps[i],buf);
        sprintf(buf,"%+.2f",gradient[i]); YeditSet(edit_grads[i],buf);
        assert(range[i] <= 3);
        [popup_ranges[i] selectItemAtIndex:range[i]];
      }
    }
    for (int i=0; i<TM_N; i++) {
      if (monitor[i]) {
        [self updateTemperature:i fromHW:NO];
        sprintf(buf,"%+.2f",gradient[i]);
        YeditSet(edit_grads[i],buf);
      }
    }
  });
}

/* ---------------------------------------------------------------- */

- (int)updateSetpoint:(int)c fromHW:(int)refresh
{
  int err=0;
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,c,refresh);
#endif

  if (refresh) { float f=sim_setp[c];
    err = [lakeshore getSetpoint:&f channel:c];
    if (!err) setpoint[c] = f;
  }
  sprintf(buf,"%.1f",my_round(setpoint[c],1));
  YeditSet(edit_setps[c],buf);

  return err;
}

/* ---------------------------------------------------------------- */

- (int)updateTemperature:(int)c fromHW:(int)refresh
{
  int err=0;

  if (refresh) { float t=sim_temp[c];
    time_t ut = cor_time(0);
    double dt = fmax(sample_time,ut-last_temp[c]); // safety net
    err = [lakeshore getTemp:&t channel:c];
    if (!err) { char buf[128];
      double f = exp(-dt/(60.0/LOG2));      // and smooth gradient
      gradient[c] = f*gradient[c] + (1.0-f)*(t-temperature[c])/(dt/60.0);
      temperature[c] = t;
      last_temp[c] = ut;
      sprintf(buf,"temp%c: %d %.2f %.2f",'A'+c,1+c,temperature[c],gradient[c]);
      [main_logger append:buf];
    }
  }
  dispatch_async(MAIN_QUEUE,^(void) { char buf[128];
    sprintf(buf,"%.1f",temperature[c]);
    YeditSet(edit_temps[c],buf);
    sprintf(buf,"%.2f",gradient[c]);
    YeditSet(edit_grads[c],buf);
  });

  return err;
}

/* ---------------------------------------------------------------- */

- (int)updatePower:(int)c fromHW:(int)refresh
{
  int err=0;
  char buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d)\n",PREFUN,c,refresh);
#endif

  if (refresh) { float p=sim_power[c];
    time_t ut = cor_time(0);
    double dt = (double)(ut-last_power[c]);
    if (dt <= 0) dt = sample_time;          // safety net (clock change)
    err = [lakeshore getPower:&p channel:c];
    if (!err) {
      power[c] = p;
      double f = exp(-dt/(15.0/LOG2));      // smooth average
      power_ave[c] = f*power[c] + (1.0-f)*power_ave[c];
      last_power[c] = ut;
    }
  }
  sprintf(buf,"%.1f",power_ave[c]);
  YeditSet(edit_powers[c],buf);

  return err;
}

/* ---------------------------------------------------------------- */
#pragma mark "Action"
/* ---------------------------------------------------------------- */

- (void)action_menu:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p) %s\n",PREFUN,sender,[[sender title] UTF8String]);
#endif
  assert([NSThread isMainThread]);
  if (!initialized) return;

  if (strstr([[sender title] UTF8String],"TempGraph")) {
    [tempGraph.window orderFront:self];
  } else
  if (strstr([[sender title] UTF8String],"PowerGraph")) { 
    [powerGraph.window orderFront:self];
  } else
  if (strstr([[sender title] UTF8String],"All Graphs")) {
    [controlGraph.window orderFront:self];
    [powerGraph.window orderFront:self];
    [tempGraph.window orderFront:self];
  } else
  if (strstr([[sender title] UTF8String],"GraphPars")) {
    if (controlGraph.window == [NSApp keyWindow]) {
      [controlGraph action_gpars:sender];
    } else
    if (powerGraph.window == [NSApp keyWindow]) {
      [powerGraph action_gpars:sender];
    }
    if (tempGraph.window == [NSApp keyWindow]) {
      [tempGraph action_gpars:sender];
    }
  } else
  if (strstr([[sender title] UTF8String],"TunePars")) {
    [self action_tune:sender];
  } else
  if (strstr([[sender title] UTF8String],"Lock")) {
    [self lockGUI:LOCK_TOGGLE];
  } else {
    assert(0);
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_range:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (!initialized) return;

  NSInteger i = [popup_ranges indexOfObject:sender];
  assert(i != NSNotFound);

  range[i] = (int)[sender indexOfSelectedItem];
  
  dispatch_async(GLOBAL_QUEUE,^(void) {
    int err = [self changeRange:(int)i];
    if (err) { NSString *s;
      s = [NSString stringWithFormat:@"%@TempController: changing 'range' failed, err=%d\r%@",
           cName,err,[[self class] errorString:err]];
      [app message:s level:2];
    }
    [self updateGUI];
  });
}

/* ---------------------------------------------------------------- */

- (IBAction)action_setp:(id)sender
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (!initialized) return;

  NSInteger i = [edit_setps indexOfObject:sender];
  assert(i != NSNotFound);

  float f = (float)fmax(SETP_MIN,fmin(SETP_MAX,[sender doubleValue]));
  int err = [lakeshore setSetpoint:f channel:(int)i];
  if (err) { NSString *string;
    string = [NSString stringWithFormat:@"%@TempController: change of 'setpoint' failed, err=%d\r%@",
      cName,err,[[self class] errorString:err]];
    [app message:string level:2];
  } else {
    setpoint[i] = f;
    sprintf(buf,"%.1f",setpoint[i]); YeditSet(edit_setps[i],buf);
  }
}

/* --- */

- (IBAction)action_target:(id)sender
{
 char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (!initialized) return;
  
  NSInteger i = [edit_targets indexOfObject:sender];
  assert(i != NSNotFound);

  target[i] = (float)fmax(SETP_MIN,fmin(SETP_MAX,[sender doubleValue]));
  sprintf(buf,"%.1f",target[i]); YeditSet(edit_targets[i],buf);
  int err = [lakeshore setSetpoint:target[i] channel:(int)i];
  if (err) { NSString *s;
    s = [NSString stringWithFormat:@"updating setpoint failed, err=%d\r%@",
         err,[[self class] errorString:err]];
    [Notice openNotice:[s UTF8String] title:p_exec time:0 level:2];
  }
  sample_time = ST_MIN;
}

/* --- */

- (IBAction)action_ramp:(id)sender
{
  char buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (!initialized) return;

  NSInteger i = [edit_ramps indexOfObject:sender];
  assert(i != NSNotFound);

  ramp[i] = fabs([sender floatValue]);
  sprintf(buf,"%.1f",ramp[i]); YeditSet(sender,buf);

  int err = [lakeshore setRamp:ramp[i] channel:(int)i];
  if (ramp[i] == 0) {          // adjust target
    target[i] = setpoint[i];
    sprintf(buf,"%.1f",target[i]); YeditSet(edit_targets[i],buf);
  }
  if (err) { NSString *s;
    s = [NSString stringWithFormat:@"updating ramp failed, err=%d\r%@",
         err,[[self class] errorString:err]];
    [Notice openNotice:[s UTF8String] title:p_exec time:0 level:2];
  }
  
  [self lockGUI:LOCK_UPDATE];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_tune:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (!initialized) return;

  NSInteger i = 0; // TODO depends on channel -- yyyzzz menu needs 'A' 'B' submenu
  assert(i != NSNotFound);

  if (strstr([[sender title] UTF8String],"Tune")) {
    [edit_p setIntValue:p_term[i]];
    [edit_i setIntValue:i_term[i]];
    [edit_d setIntValue:d_term[i]];
    [panel_tune makeKeyAndOrderFront:self];
  } else
  if (strstr([[sender title] UTF8String],"Ok")) {
    p_term[i] = imin(999,imax(0,[edit_p intValue]));
    i_term[i] = imin(999,imax(0,[edit_i intValue]));
    d_term[i] = imin(200,imax(0,[edit_d intValue]));
    [panel_tune orderOut:self];
    int err = [lakeshore setP:p_term[i] I:i_term[i] D:d_term[i] channel:(int)i];
    if (err) { NSString *s;
      s = [NSString stringWithFormat:@"%@ %s: setting PID values failed, err=%d\r%@",
        cName,M_TITLE,err,[[self class] errorString:err]];
      [Notice openNotice:[s UTF8String] title:p_exec time:0 level:2];
    }
  } else {
    assert(0);
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_check:(id)sender
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  NSInteger i = [chk_mons indexOfObject:sender];
  assert(i != NSNotFound);

  @synchronized(self) {
    NSButton *but = (NSButton*)sender;
    monitor[i] = ([but state] == NSControlStateValueOn) ? 1 : 0;
  }
  if (monitor[i]) {
    [self updateChannel:(int)i];
  } else {
    [edit_temps[i] setStringValue:@""];
    [edit_grads[i] setStringValue:@""];
  }
  [tempGraph setChannel:(int)i monitor:((monitor[i]) ? YES : NO)];
  
  [self lockGUI:LOCK_UPDATE];
  
  YprefPutArray(DBE_MONITOR,YarrayFromInts((int*)monitor,TM_N),use_shared);
}

/* ---------------------------------------------------------------- */

- (int)changeRange:(int)c
{
  int err;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",PREFUN,c);
#endif
  
  err = [lakeshore open]; if (err) return err;
  
  err = [lakeshore setRange:range[c] channel:c];
  if (!err) {
    if (range[c] > 0) { NSString *key; NSNumber *val;  // update PID values
      NSDictionary *dict  = [TempController defaultDictionary];
      key = [NSString stringWithFormat:@"p_range%d_%d",range[c],1+c];
      val = [dict objectForKey:key];
      if (val) p_term[c] = [val intValue];
      key = [NSString stringWithFormat:@"i_range%d_%d",range[c],1+c];
      val = [dict objectForKey:key];
      if (val) i_term[c] = [val intValue];
      key = [NSString stringWithFormat:@"d_range%d_%d",range[c],1+c];
      val = [dict objectForKey:key];
      if (val) d_term[c] = [val intValue];
      err = [lakeshore setP:p_term[c] I:i_term[c] D:d_term[c] channel:c];
    }
  }
  [lakeshore close];

  return err;
}

/* ---------------------------------------------------------------- */
#pragma mark "Run"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)updateChannel:(int)c
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,c);
#endif

  int err = [lakeshore open];  if (err) return err;

  time_t ut = cor_time(0);
  if (!err) err = [self updateTemperature:c fromHW:YES];
  if (!err) [tempGraph addData:c time:ut value:temperature[c]];
  if (c < TC_N) {
    if (!err) [controlGraph addData:c time:ut value:temperature[c]];
    if (!err) err = [self updateSetpoint:c fromHW:YES]; // todo if ramp && target != setp
    if (!err) err = [self updatePower:c fromHW:YES];
    if (!err) [powerGraph addData:c time:ut value:power_ave[c]];
  }
  // pw = maxpow*power_ave*pow(10.0,range-5.0);
  // TODO update database
  [lakeshore close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (void)run_read                       // thread runs for 1 minute
{
  int c=0,i;

  if (!initialized) return;
  if (reading) return;                 // weak MT-safe
#if (DEBUG > 1)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif
  reading = TRUE;

  time_t last = cor_time(0);

  int err = [lakeshore open];          // keep open
  if (!err) {
    NSDate *date = [NSDate date];
    while (YsecSince(date) < 60.0) {   // run for 1 minute -- TODO test disconnect / re-connect cable TEST@SBS
      Ysleep(350); if (_stop_reading) break;
      if (cor_time(0) >= last+sample_time) {
        last = cor_time(0);
        @synchronized(self) {          // protect 'monitor'
          for (i=1; i<=TM_N; i++) {
            c = (last_channel+i) % TM_N;
            if (monitor[c]) {
              if (c <= last_channel) {
                sample_time = imin(sample_time+1,ST_MAX);
                printf("new round: %d\n",sample_time); //xxx
              }
              last_channel = c;
              break;
            }
          }
        } // end-sync
        if (i <= TM_N) {
          err = [self updateChannel:c]; if (err) break;
        }
      } // endif(time)
      if (err || _stop_reading) break;
    } // endfor(i)
    [lakeshore close];
  } // endif(!err)

  if (err) { NSString *string;
    string = [NSString stringWithFormat:@"%@/%@ TempController: I/O failed, err=%d\r%@",
      cName,mName,err,[[self class] errorString:err]];
    if (cor_time(0) >= next_err_warning) {
      delta_err_warning = imin(600,delta_err_warning+30);
      next_err_warning = cor_time(0)+delta_err_warning;
      [app message:string level:1 time:delta_err_warning];
    } else {
      [app message:string level:0];
    }
  } else delta_err_warning = 0;

#if (DEBUG > 1)
  fprintf(stderr,"%s ... done (err=%d,%d)\n",PREFUN,err,delta_err_warning);
#endif
  if (err) {
    for (int i=0; i<delta_err_warning; i++) {  // back-off
      Ysleep(350); if (_stop_reading) break;
    }
  }
  reading = FALSE;
}

/* ---------------------------------------------------------------- */

- (void)run_setup
{
  int    err=0,i;
  float  temp;
  char   buf[128];
  time_t ut;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,self);
#endif

  initialized = FALSE;

  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  sprintf(buf,"connecting to LakeShore temperature controller #%d",myid);
  Notice *note = [[Notice openNotice:buf title:p_exec] retain];
  
  ut = cor_time(0);
  
  err = [lakeshore open];
  if (!err) {
    err = [lakeshore getVersion:buf];
    if (!err) printf("'%s'\n",buf); //xxx
    for (i=0; i<TC_N; i++) {                // setup all channels -- ignore monitor[]
      if (!err) err = [lakeshore setupCurrent:maxcur[i] resistor:resistor[i] channel:i];
      ramp[i] = 0.0;
      if (!enabled[i]) continue;            // no resistor
      if (!err) err = [lakeshore getP:&p_term[i] I:&i_term[i] D:&d_term[i] channel:i];
      if (!err) err = [lakeshore getRange:&range[i] channel:i];
      power[i] = sim_power[i];
      if (!err) err = [lakeshore getPower:&power[i] channel:i];
      power_ave[i] = (double)power[i];
      last_power[i] = ut;
      setpoint[i] = sim_setp[i];
      if (!err) err = [lakeshore getSetpoint:&setpoint[i] channel:i];
    }
    for (i=0; i<TM_N; i++) {                // all sensors
      temp = sim_temp[i];
      if (!err) err = [lakeshore getTemp:&temp channel:i];
      temperature[i] = temp;
      if (i < TC_N) target[i] = (float)fmin(SETP_MAX,my_round(temp,1));
      last_temp[i] = ut;
    }
    for (i=0; i<TC_N; i++) {                // controllers
      if (!enabled[i]) continue;
      if (temperature[i] > setpoint[i]) {   // carefully raise setpoint
        float r = fmax(30.0,(temperature[i]-setpoint[i])/10.0);
        float t = (temperature[i]-setpoint[i])/r;
        printf("r=%f, t=%f\n",r,t); //xxx
        if (!err) err = [lakeshore setRamp:r channel:i];
        if (!err) err = [lakeshore setSetpoint:setpoint[i] channel:i];
        [NSThread sleepForTimeInterval:(1.0+t)];
      }
      // if (!err) err = [lakeshore setRamp:RAMP_MAX channel:i]; // todo updateRamp/setRamp
    }
    [lakeshore close];
  }
  [note close]; [note release];
  
  if (!err) {                          // ok
    [self updateGUI];
    for (i=0; i<TC_N; i++) {
      [controlGraph addData:i time:ut value:temperature[i]];
      if (enabled[i]) {
        [powerGraph addData:i time:ut value:power[i]];
      }
    }
    for (i=0; i<TM_N; i++) {
      if (monitor[i]) {
        [tempGraph addData:i time:ut value:temperature[i]];
      }
    }
    if (!lakeshore.online) [self lockGUI:LOCK_NO];
    Ysleep(1200);
    initialized = TRUE;
  } else { NSString *string;
    string = [NSString stringWithFormat:@"%@/%@ %s setup failed (err=%d)\r%@",
      cName,mName,M_TITLE,err,[[self class] errorString:err]];
    [app message:string level:2];
  }

#ifdef NDEBUG
  if (!err) {
    dispatch_async(MAIN_QUEUE,^(void) {
      NSMenuItem *item = [[NSMenuItem alloc] initWithTitle:@"All Graphs"
                                             action:NULL keyEquivalent:@""];
      [self action_menu:item];
      [item release];
    });
  }
#endif

  [pool drain];
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
