/* ----------------------------------------------------------------
 *
 * plcsim.c 
 * 
 * Project: PLC Simulator  (OCIW, Pasadena, CA)
 *
 * 2013-03-29  Christoph C. Birk (birk@obs.carnegiescience.edu)
 *
 * ---------------------------------------------------------------- */

#define IS_PLCSIM_C

/* DEFINEs -------------------------------------------------------- */

#define DEBUG           1
#define TIME_TEST

#define P_TITLE         "PLC-Simulator"
#define P_VERSION       "2014-06-20"
#define COOKIE_ADDRESS  0

/* INCLUDEs ------------------------------------------------------- */

#define _REENTRANT

#include <stdlib.h>                    /* exit() */
#include <stdio.h>                     /* sprintf() */ 
#include <string.h>                    /* strcpy() */
#include <unistd.h>                    /* sleep(),nice(),getpid() */
#include <assert.h>

#ifdef TIME_TEST
#include <limits.h>
#include <sys/times.h>
#endif

#include "utils.h"                     /* generic utilities */
#include "tcpip.h"                     /* TCPIP library */
#include "ptlib.h"                     /* POSIX thread lib */

#define PREFUN        __PRETTY_FUNCTION__

/* GLOBALs -------------------------------------------------------- */


/* STATICs -------------------------------------------------------- */

static void*  tcpip_server     (void*);
static void*  tcpip_command    (void*);
static char*  tcpip_handler    (const char*,int);

static int     done=0;
static u_char  coil[0x10000];
static u_short value[0x10000];
static float   air_pressure=96.0f;
static float   plc_temp=20.0f;

/* function prototype(s) ------------------------------------------ */

int main(int argc,char **argv)
{
  int        i;
  long       server_port=0xffff-502;   /* '502' is privileged */
  pthread_t  server_tid;
  char       buffer[1024];

  /* here comes the code ------------------------------------------ */ 

#if (DEBUG > 0)
  sprintf(buffer,"pid=%d",(int)getpid());
  tdebug(P_TITLE,buffer);
#endif
  value[COOKIE_ADDRESS] = value[COOKIE_ADDRESS+1] = 0;

  { extern char *optarg;               /* parse command line */
    extern int opterr,optopt,optind; opterr=0;
    while ((i=getopt(argc,argv,"a:c:p:t:")) != EOF) {
      switch (i) {
      case 'a':                        /* air pressure */
        air_pressure = (float)atof(optarg);
        break;
      case 'c':                        /* cookie */
        value[COOKIE_ADDRESS] = atoi(optarg) & 0xffff;
        value[COOKIE_ADDRESS+1] = (atoi(optarg) >> 16) & 0xffff;
        break;
      case 'p':                        /* TCPIP command port */
        server_port = atol(optarg);
        break;
      case 't':                        /* PLC temperature */
        plc_temp = (float)atof(optarg);
        break;
      case '?':
        fprintf(stderr,"%s: option '-%c' unknown or parameter missing\n",
                P_TITLE,(char)optopt);
        exit(1);
      }
    }
  }

  srand48((long)time((time_t*)NULL));  /* init random numbers */

  /* create TCP/IP server thread ---------------------------------- */
 
  i = thread_create(tcpip_server,(void*)server_port,&server_tid);
  if (i) {
    fprintf(stderr,"%s: creating TCPIP-server thread failed (err=%d)\n",
            P_TITLE,i);
    exit(1);
  }
#if (DEBUG > 0)
  fprintf(stdout,"%s listens on port %ld\n",P_TITLE,server_port);
#endif

  /* this loop drives the program --------------------------------- */

  do {
    sleep(1);
  } while (!done);

  /* free up all the resources and exit --------------------------- */ 

#ifdef TIME_TEST                       /* print cpu-time */
  { struct tms tmsbuf;
    (void)times(&tmsbuf);
    sprintf(buffer,"%s(%d): total-cpu: %5.2f seconds",P_TITLE,(int)getpid(),
      (float)(tmsbuf.tms_stime)/(float)sysconf(_SC_CLK_TCK) +
      (float)(tmsbuf.tms_utime)/(float)sysconf(_SC_CLK_TCK));
    adebug(P_TITLE,buffer);
  }
#endif

#if (DEBUG > 0)
  sprintf(buffer,"good-bye (pid=%d)",(int)getpid());
  tdebug(P_TITLE,buffer);
#endif
 
  return 0;
} 

/* ---------------------------------------------------------------- */

static void* tcpip_server(void* param)
{
  long       port=(long)param,msgsock;
  int        err,sock;
  IP_Address req_address;

#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,param);
#endif

  sock = TCPIP_CreateServerSocket((int)port,&err);  /* open socket */
  if (sock == -1) { char buffer[1024];         /* failed */
    sprintf(buffer,"TCPIP_CreateServerSocket()=%d, err=%d",sock,err);
    tdebug(P_TITLE,buffer);
#if (DEBUG > 0)
    fprintf(stderr,"%s(): '%s'\n",PREFUN,buffer);
#endif
    return((void*)(long)err);
  }

  while (!done) {                        /* main loop  */
    msgsock = TCPIP_ServerAccept(sock,&req_address);
    (void)thread_detach(tcpip_command,(void*)msgsock);
  } /* while(!done) */

  (void)close(sock);

  return((void*)0);                    /* terminate thread */
}

/* ---------------------------------------------------------------- */

static void* tcpip_command(void* param)
{
  int  rval,sock=(int)(long)param;
  char command[256],*answer;

#if (DEBUG > 1)
  fprintf(stderr,"%s(%d)\n",PREFUN,(int)param);
#endif
  do {
    (void)memset(command,0,sizeof(command));       /* clear buffer */
    rval = recv(sock,command,sizeof(command)-1,0); /* read command */
    if (rval > 0) {
      if (!strcmp(command,TCPIP_FINITO_CMD)) {     /* 'terminate' */
        done = TRUE;
        answer = (char*)malloc(256*sizeof(char));
        strcpy(answer,command);
      } else {
        answer = tcpip_handler(command,sock);
      }
      if (answer) {
        if (*answer) {
          send(sock,answer,strlen(answer),0);  /* send answer */
        }
        free((void*)answer);
      }
    }
    if (done) break;
  } while (rval > 0);                  /* loop while there's something */

  close(sock);

  return NULL;
}

/* ---------------------------------------------------------------- */

char* tcpip_handler(const char* command,int sock) 
{
  int     i,nbytes=0,adr;
  u_char  funcode,*cmd;
  u_short address,data;
  char    answer[128];

#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%d)\n",PREFUN,command,sock);
#endif
  cmd = (u_char*)command;

  /* parse -------------------------------------------------------- */
 
  // todo PFS specific stuff, in particular "shutter" !!!
 
  // for (i=0; i<12; i++) printf("0x%02x ",cmd[i]);
  // printf("\n");

  funcode = cmd[7];  
  address = 0x100*(u_short)cmd[8]  + cmd[9];
  data    = 0x100*(u_short)cmd[10] + cmd[11];
  // printf("%02x %04x %04x\n",funcode,address,data);

  switch (funcode) {                   /* function code */
  case 0x05:                           /* set coil or output */
    coil[address] = cmd[10];
    strcpy(answer,"ok\n"); 
    nbytes = strlen(answer);
    if (address >= 3072) {             /* PLC specific coils */
      adr = address-3072;
      printf("set coil[%d]=%u\n",adr,cmd[10]);
      coil[adr+2048] = cmd[10];        /* set output/input */
      switch (adr) {
      case 48:
        coil[3072+129] = (cmd[10]) ? 1 : 0;   // C201
        break; 
      }
    }
    break;
  case 0x01:                           /* get coil */
  case 0x02:                           /* get input */
    memset(answer,0,10); answer[9] = (coil[address]) ? 0x01 : 0x00;
    adr = address-3072; 
    printf("get coil[%d]=%u\n",adr,answer[9]);
    nbytes = 10;
    break;
  case 0x06:                           /* set value */
    if (address <= COOKIE_ADDRESS+1) { /* 16-bit data */
      value[address] = data;
    } else { char buf[32];             /* BCD */
      sprintf(buf,"%2x%02x",cmd[10],cmd[11]);
      value[address] = atoi(buf);
    }
    strcpy(answer,"ok\n"); 
    nbytes = strlen(answer);
    break;
  case 0x04:                           /* get value */
    switch (address) {
    case 32:                           /* V40: air pressure */
      value[address] = (int)(air_pressure*10) + (lrand48()%11) - 5;
      // printf("value(%d)= %d\n",address,value[address]);
      break;
    case 33:                           /* V41: temperature */
      value[address] = (int)(plc_temp*10) + (lrand48()%11) - 5;
      // printf("value(%d)= %d\n",address,value[address]);
      break;
    }
    memset(answer,0,11); 
    if (address <= COOKIE_ADDRESS+1) { /* 16-bit data */
      i = value[address];
    } else { char buf[32];             /* BCD */
      sprintf(buf,"%04d",value[address]);
      i =   0x1000 * (int)(buf[0]-'0')
          + 0x0100 * (int)(buf[1]-'0')
          + 0x0010 * (int)(buf[2]-'0')
          + 0x0001 * (int)(buf[3]-'0');
    }
    answer[9]  = i / 0x100;
    answer[10] = i % 0x100;
    nbytes = 11;
    break;
  default:
    strcpy(answer,"error\n");
    nbytes = strlen(answer);
    break;
  }
  assert(nbytes); 
  send(sock,answer,nbytes,0);     // todo MSG_NOSIGNAL ?

  return NULL;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

