/* ----------------------------------------------------------------
 * 
 * stasim.c
 * 
 * Project: PFS/STA (OCIW, Pasadena, CA)
 *
 * Archon controller simulator
 *
 * 2015-07-09  Christoph C. Birk (birk@obs.carnegiescience.edu)
 * 2016-06-03  TCP/IP interface
 * 2019-10-17  data simulator (FRAME,FETCH)
 * 2019-12-12  effedUp fetch
 *
 * ---------------------------------------------------------------- */

#define IS_CCDSERVER_C
#define IS_MAIN

/* DEFINEs -------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG           1              /* debug level */
#endif

#define TIME_TEST       1 

#define P_TITLE         "Archon-Simulator" 
#define P_VERSION       "2019-12-12"

#define PREFUN          __func__

/* INCLUDEs ------------------------------------------------------- */

#define _REENTRANT                     /* multi-threaded code */

#include <stdlib.h>                    /* atoi(),exit() */
#include <stdio.h>                     /* sprintf() */
#include <string.h>                    /* strcasecmp() */
#include <unistd.h>                    /* write(),getpid(),getopt() */
#include <assert.h>

#if (TIME_TEST > 0)
#include <limits.h> 
#include <sys/times.h>
#endif

#include "utils.h"                     /* generic utilities */
#include "random.h"                    /* Mersenne Twister */
#include "ptlib.h"                     /* POSIX thread lib */
#include "tcpip.h"                     /* TCPIP stuff */

#ifdef MACOSX
#define MSG_NOSIGNAL  0 
#endif

/* EXTERNs -------------------------------------------------------- */

/* function prototype(s) */

#ifdef LINUX
extern void         swab(const void*,void*,size_t);
#endif

/* STATICs -------------------------------------------------------- */

static double startup_time=0;
static u_long timer=0;
static int    done=0;
static int    power=2;
static int    binx=1,biny=2;
static int    preSkip=0,postSkip=0;
static int    nover=80,nbias=40,lines=10560,pixels=1320;
static int    effedUp=0;

/* function prototype(s) ------------------------------------------ */

static void*  tcpip_server     (void*);

/* --- M A I N ---------------------------------------------------- */

int main(int argc,char **argv)
{
  int  i,server_port=4242;
  char buffer[1024];

  /* here comes the code ------------------------------------------ */ 

#if (DEBUG > 1)
  sprintf(buffer,"%s: pid=%d, niceval=%d",P_TITLE,(int)getpid(),nice(0));
  tdebug(argv[0],buffer);
#endif

  InitRandom(0,0,0);                   /* initialize random# */

  { extern char *optarg;               /* parse command line */
    extern int opterr,optopt,optind; opterr=0;
    while ((i=getopt(argc,argv,"p:")) != EOF) {
      switch (i) {
      case 'p':                        /* port */
        server_port = (u_short)atoi(optarg);
        break;
      case '?':
        fprintf(stderr,"%s: option '-%c' unknown or parameter missing\n",
                P_TITLE,(char)optopt);
        exit(1);
      }
    } /* endwhile(getopt) */
  }

  /* check for multiple instances --------------------------------- */

#if (DEBUG == 1)
  sprintf(buffer,"%s: pid=%d, niceval=%d",P_TITLE,(int)getpid(),nice(0));
  tdebug(argv[0],buffer);
#endif

  startup_time = walltime(0);

  tcpip_server((void*)(long)server_port);

  /* free up all the resources and exit --------------------------- */

#if (TIME_TEST > 0)                    /* print cpu-time */
  { struct tms    tmsbuf;
    (void)times(&tmsbuf);
    sprintf(buffer,"(%d): total-cpu: %5.2f seconds",(int)getpid(),
      (float)(tmsbuf.tms_stime)/(float)sysconf(_SC_CLK_TCK) +
      (float)(tmsbuf.tms_utime)/(float)sysconf(_SC_CLK_TCK));
    tdebug(P_TITLE,buffer);
  }
#endif
 
#if (DEBUG > 0)
  sprintf(buffer,"good-bye (pid=%d)",(int)getpid());
  tdebug(P_TITLE,buffer);
#endif

  return(0);
}

/* ---------------------------------------------------------------- */

static void cmd_status(char* answer)
{
  char buf[128];
  const int TARGET_TEMP=-117;

  float temp_ccd = (float)GRandom(TARGET_TEMP,0.1);

  sprintf(answer,"VALID=1 POWER=%d",power);
  sprintf(buf," MOD2/TEMPA=%.1f",temp_ccd); strcat(answer,buf);
  sprintf(buf," MOD2/TEMPB=%.1f",temp_ccd); strcat(answer,buf);
  sprintf(buf," MOD2/TEMPC=-170"); strcat(answer,buf);
  sprintf(buf," MOD2/HEATERATARGET=%d",TARGET_TEMP); strcat(answer,buf);
  sprintf(buf," MOD2/HEATERAOUTPUT=4.5"); strcat(answer,buf);
}

/*----------------------------------------------------------------- */

static void cmd_config(const char* cmd)
{
  char *p;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,cmd);
#endif

  if ((p=strstr(cmd,"HorizontalBinning=")) != NULL) {
    p = strchr(p,'='); if (p) binx = atoi(p+1); 
  } else
  if ((p=strstr(cmd,"VerticalBinning=")) != NULL) {
    p = strchr(p,'='); if (p) biny = atoi(p+1); 
  } else
  if ((p=strstr(cmd,"PreSkipLines=")) != NULL) {
    p = strchr(p,'='); if (p) preSkip = atoi(p+1); 
  } else
  if ((p=strstr(cmd,"PostSkipLines=")) != NULL) {
    p = strchr(p,'='); if (p) postSkip = atoi(p+1); 
  } else
  if ((p=strstr(cmd,"OverscanLines=")) != NULL) {
    p = strchr(p,'='); if (p) nbias = atoi(p+1); 
    // printf("nbias=%d\n",nbias); 
  } else
  if ((p=strstr(cmd,"Lines=")) != NULL) {
    p = strchr(p,'='); if (p) lines = atoi(p+1); 
    // printf("lines=%d\n",lines);
  } else
  if ((p=strstr(cmd,"Pixels")) != NULL) {
    p = strchr(p,'='); if (p) pixels = atoi(p+1); 
    // printf("pixels=%d\n",pixels);
  } 
}

/*----------------------------------------------------------------- */

static void cmd_frame(char* answer)
{
  int           i;
  char          key[32],buf[128];
  static int    wbuf=0,read_lines=0,complete=1,frame=0;
  static u_long timestamp=0;
  static double read_start=0;
  const  int    ntaps=8;
  
  int max_lines = ((10560-preSkip-postSkip)/biny)+nbias;
  assert(lines+nbias == max_lines);
  assert(pixels == (1320/binx)+nover);

  if (complete == 1) {               
    wbuf++; if (wbuf > 3) wbuf=1;
    frame++; complete = 0; read_lines = 0; timestamp = timer;
    read_start = walltime(0);
  }
    
  read_lines = (int)((walltime(0)-read_start)*(10560.0/60.0));
  if (read_lines >= max_lines) {
    read_lines = max_lines; complete = 1;
  }

  sprintf(answer,"TIMER=%016lx RBUF=0 WBUF=%d",timer,wbuf);
  for (i=1; i<=3; i++) { 
    sprintf(key,"BUF%dBASE",i);
    sprintf(buf," %s=%d",key,0); strcat(answer,buf);
    sprintf(key,"BUF%dFRAME",i);
    sprintf(buf," %s=%d",key,frame); strcat(answer,buf);
    sprintf(key,"BUF%dWIDTH",i);
    sprintf(buf," %s=%d",key,ntaps*pixels); strcat(answer,buf);
    sprintf(key,"BUF%dHEIGHT",i);
    sprintf(buf," %s=%d",key,max_lines); strcat(answer,buf);
    sprintf(key,"BUF%dLINES",i);
    sprintf(buf," %s=%d",key,(i==wbuf) ? read_lines : max_lines); strcat(answer,buf);
    sprintf(key,"BUF%dCOMPLETE",i);
    sprintf(buf," %s=%d",key,(i==wbuf) ? complete : 1); strcat(answer,buf);
    sprintf(key,"BUF%dTIMESTAMP",i);
    sprintf(buf," %s=%016lx",key,(i==wbuf) ? timestamp : 0); strcat(answer,buf);
  }
}

/* ---------------------------------------------------------------- */

static int cmd_fetch(const char* cmd,int sock)
{
  u_int i,addr,nblocks;
  char  data[1028];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,cmd,sock);
#endif

  if (effedUp) return 1;               /* 2019-12-12 */

  sscanf(cmd,"FETCH%08X%08X",&addr,&nblocks);
  // printf("%08X %d\n",addr,nblocks);

  u_short *p = (u_short*)data;
  for (i=0; i<1028/2; i++) *p++ = 1000+IRandom(10);
  for (i=0; i<nblocks; i++) { 
    send(sock,data,1028,MSG_NOSIGNAL);
  }
  return 1;
}

/* ---------------------------------------------------------------- */

char* tcpip_handler(const char* command,int sock)
{
  int flag=0;
  char buffer[1024],*h,cmd[128]="",par[128]="",*q,*answer;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d)]\n",PREFUN,command,sock);
#endif

  answer = (char*)malloc(2048*sizeof(char));
  if (strlen(command) < 4) {
    sprintf(answer,"ERROR=%d (invalid commmand '%s')'",-1,cmd);
    return answer;
  }
  if (strlen(command) > 120) {
    sprintf(answer,"ERROR=%d (invalid commmand)'",-2);
    return answer;
  }

  /* parse command ------------------------------------------------------ */
 
  strcpy(buffer,command+3);
  h = strtok_r(buffer," \t\r\n",&q); if (h) strcpy(cmd,h);
  if ((h=strtok_r(NULL," \t\r\n",&q)) != NULL) strcpy(par,h);
  while ((h=strtok_r(NULL," \t\r\n",&q)) != NULL) {
    strcat(par," "); strcat(par,h);
  }
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%s)\n",PREFUN,cmd,par);
#endif
  
  timer = (u_long)((walltime(0)-startup_time)/10.0e-9);

  strcpy(answer,"");
       if ( !strcmp(cmd,"version")) sprintf(answer,"version= %s",P_VERSION);
  else if ( !strcmp(cmd,"SYSTEM"))  sprintf(answer,"BACKPLANE_TYPE=1");
  else if ( !strcmp(cmd,"TIMER"))  sprintf(answer,"%016lx",timer); 
  else if ( !strcmp(cmd,"FRAME"))  cmd_frame(answer);
  else if ( !strcmp(cmd,"STATUS"))  cmd_status(answer);
  else if ( !strcmp(cmd,"POWERON"))  power = 4;
  else if ( !strcmp(cmd,"POWEROFF"))  power = 2;
  else if ( !strcmp(cmd,"FETCHLOG")) sprintf(answer,"(null)");
  else if ( !strcmp(cmd,"CLEARCONFIG")) strcpy(answer,"");
  else if ( !strcmp(cmd,"LOADPARAM")) strcpy(answer,"");
  else if (!strncmp(cmd,"WCONFIG",7)) cmd_config(cmd);
  else if (!strncmp(cmd,"APPLY",5)) strcpy(answer,"");
  else if (!strncmp(cmd,"FETCH",5)) flag = cmd_fetch(cmd,sock);
  else if (!strncmp(cmd,"POLL",4)) strcpy(answer,"");
  else if (!strncmp(cmd,"LOCK",4)) strcpy(answer,"");
  else if (!strncmp(cmd,"fck",3))  effedUp = atoi(par);
  else sprintf(answer,"ERROR=%d (unknown commmand '%s')'",-3,cmd);

  if (flag) { 
    free((void*)answer); answer=NULL;
  } else {
#if (DEBUG > 1)
    fprintf(stdout,"%s: %s\n",P_TITLE,answer);
#endif
    assert(strlen(answer)+8 < sizeof(buffer));
    sprintf(buffer,"%c%c%c%s",'<',command[1],command[2],answer);
    sprintf(answer,"%s\n",buffer);        /* append <LF> */
  }

  return answer;
}

/* ---------------------------------------------------------------------- */

static void* tcpip_command(void* param)
{
  int  rval,sock=(int)(long)param;
  char command[256],*answer;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%ld)\n",PREFUN,(long)param);
#endif

  do { 
    (void)memset(command,0,sizeof(command));       /* clear buffer */
    rval = recv(sock,command,sizeof(command)-1,0); /* read command */
    if (rval > 0) { // TODO read until <CR>
      if (!strcmp(command,TCPIP_FINITO_CMD)) {     /* 'terminate' */
        done = TRUE;
        answer = (char*)malloc(256*sizeof(char));
        strcpy(answer,command);
      } else {
        answer = tcpip_handler(command,sock);
      }
      if (answer) { 
        if (*answer) {
          send(sock,answer,strlen(answer),0);  /* send answer */
        }
        free((void*)answer);
      }
    }
    if (done) break;
    if (effedUp > 1) break;            /* 2019-12-12 */
  } while (rval > 0);                  /* loop while there's something */

  close(sock);

  return NULL;
}

/* ---------------------------------------------------------------- */

static void* tcpip_server(void* param) 
{
  int        port=(int)(long)param;
  int        err,sock,msgsock;
  IP_Address req_address;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,param);
#endif

  sock = TCPIP_CreateServerSocket(port,&err);  /* open socket */
  if (sock == -1) {                            /* failed */
    fprintf(stderr,"TCPIP_CreateServerSocket()=%d, err=%d\n",sock,err);
    return((void*)(long)err);
  }

  while (!done) {                      /* main loop  */
    msgsock = TCPIP_ServerAccept(sock,&req_address);
    (void)thread_detach(tcpip_command,(void*)(long)msgsock);
  } /* while(!done) */

  (void)close(sock);

  return((void*)0);                    /* terminate thread */
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

